#include <sstream>
#include <thread>
#include "FPGAOperation.hpp"
#include "FPGADAPI.hpp"

FPGAOperation::FPGAOperation(OPERATION_TYPE oprationType, string helpContent) : OperationBaseClass(oprationType, helpContent)
{

}

FPGAOperation::~FPGAOperation()
{
  
}

int
FPGAOperation::executeOperation(StringParser in_parser)
{
    string  strParameter;

    in_parser.getParameter(&strParameter, 0);

    if (strParameter == HELP)
    {
        showOperationHelp();
        return 0;
    }

    switch (getOperationType())
    {
        // Operations of Show
        case OPERATION_ALL:
            return executeShowAll(in_parser);
        case OPERATION_DESIGN_ID:
            return executeShowDesignID(in_parser);
        case OPERATION_FW:
            return executeShowFW(in_parser);
        // Operations of SPAD
        case OPERATION_RANGE:
            return executeSPADRange(in_parser);
        case OPERATION_READ:
            return executeSPADRead(in_parser);
        case OPERATION_WRITE:
            return executeSPADWrite(in_parser);
        // Operations of Utility
        case OPERATION_GET:
            return executeUtilityGet(in_parser);
        case OPERATION_SET:
            return executeUtilitySet(in_parser);
        default:
            cout << UNKNOWN_OPERATION_MSG << endl;
            break;
    }

    return -1;
}

int
FPGAOperation::executeShowAll(StringParser in_parser)
{
    if (executeShowDesignID(in_parser) != 0)
        return -1;
    if (executeShowFW(in_parser) != 0)
        return -1;
    return 0;
}

int
FPGAOperation::executeShowDesignID(StringParser in_parser)
{
    int         count;
    FPGADAPI    fpgaDAPI("FPGADAPI");
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    string      strP1, strValue;

    // Check syntax
    in_parser.getParameterNumber(&count);
    if (count != 1)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    in_parser.getParameter(&strP1, 0);

    if ((pErrorMsg = fpgaDAPI.getErrorMsg()))
    {
        printError(pErrorMsg);
        return -1;
    }

    pErrorMsg = fpgaDAPI.readReg(strP1, 0x2C, &strValue);
    if (pErrorMsg)
    {
        printError(pErrorMsg);
        return -1;
    }

    cout << "Design ID:\t" << strValue << endl;
    return 0;
}

int
FPGAOperation::executeShowFW(StringParser in_parser)
{
    int         count;
    FPGADAPI    fpgaDAPI("FPGADAPI");
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    string      strP1, strMajor, strMinor;

    // Check syntax
    in_parser.getParameterNumber(&count);
    if (count != 1)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    in_parser.getParameter(&strP1, 0);

    if ((pErrorMsg = fpgaDAPI.getErrorMsg()))
    {
        printError(pErrorMsg);
        return -1;
    }

    pErrorMsg = fpgaDAPI.readReg(strP1, 0x2A, &strMajor);
    if (pErrorMsg)
    {
        printError(pErrorMsg);
        return -1;
    }

    pErrorMsg = fpgaDAPI.readReg(strP1, 0x2B, &strMinor);
    if (pErrorMsg)
    {
        printError(pErrorMsg);
        return -1;
    }

    cout << "FW Version:\t" << strMajor << "." << strMinor << endl;
    return 0;
}

int
FPGAOperation::executeSPADRange(StringParser in_parser)
{
    int         count;
    string      strP1;
    FPGADAPI    fpgaDAPI("FPGADAPI");
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    UINT32      u32Start = 0, u32End = 0, u32Size = 0;

    // Check syntax
    in_parser.getParameterNumber(&count);
    if (count != 1)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    in_parser.getParameter(&strP1, 0);

    if ((pErrorMsg = fpgaDAPI.getErrorMsg()))
    {
        printError(pErrorMsg);
        return -1;
    }

    pErrorMsg = fpgaDAPI.getSPADInfo(strP1, &u32Start, &u32Size);

    if (pErrorMsg)
    {
        printError(pErrorMsg);
        return -1;
    }

    u32End = u32Start + u32Size - 1;

    cout << "0x" << Int2HexStr(u32Start) << " - " << "0x" << Int2HexStr(u32End) << endl;

    return 0;
}

int
FPGAOperation::executeSPADRead(StringParser in_parser)
{
    int         count;
    string      strP1, strP2, strP3;
    UINT32      u32Offset;
    UINT32      u32Count;

    // Check syntax
    in_parser.getParameterNumber(&count);
    if (count != 3)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    in_parser.getParameter(&strP1, 0);
    in_parser.getParameter(&strP2, 1);
    in_parser.getParameter(&strP3, 2);

    if (in_parser.isHEXString(strP1))
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    u32Offset = std::stoul(strP1, nullptr, 0);
    u32Count = std::stoul(strP2, nullptr, 0);

    return showSPADContent(strP3, u32Offset, u32Count);
}

int
FPGAOperation::executeSPADWrite(StringParser in_parser)
{
    int         count;
    string      strP1, strP2, strP3, strTemp;
    FPGADAPI    fpgaDAPI("FPGADAPI");
    ErrorMsg*   pErrorMsg = NO_ERROR_MSG;
    INT32       s32CharOffset, s32RegCharUnit = 2;
    INT32       s32HandledRegCount;
    UINT32      u32Offset, u32CurrentOffset;

    // Check syntax
    in_parser.getParameterNumber(&count);
    if (count != 3)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    in_parser.getParameter(&strP1, 0);
    in_parser.getParameter(&strP2, 1);
    in_parser.getParameter(&strP3, 2);

    if (in_parser.isHEXString(strP1) ||
        in_parser.isHEXString(strP2))
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    if ((pErrorMsg = fpgaDAPI.getErrorMsg()))
    {
        printError(pErrorMsg);
        return -1;
    }

    u32Offset = std::stoul(strP1, nullptr, 0);
    u32CurrentOffset = u32Offset;

    // Remove 0x from input data
    strP2.erase(0, 2);
    s32CharOffset = strP2.size();

    s32HandledRegCount = 0;

    while (s32RegCharUnit <= s32CharOffset)
    {
        // Get current data from input string
        s32CharOffset -= s32RegCharUnit;
        strTemp.clear();
        strTemp = "0x" + strP2.substr(s32CharOffset, s32RegCharUnit);

        pErrorMsg = fpgaDAPI.writeSPAD(strP3, u32CurrentOffset, strTemp);

        if (pErrorMsg)
        {
            showSPADContent(strP3, u32Offset, (UINT32)s32HandledRegCount);
            printError(pErrorMsg);
            return -1;
        }

        s32HandledRegCount++;
        u32CurrentOffset += s32RegCharUnit >> 1;
    }

    return showSPADContent(strP3, u32Offset, (UINT32)s32HandledRegCount);
}

int
FPGAOperation::executeUtilityGet(StringParser in_parser)
{
    int         count;
    string      strP1, strP2, strValue;
    FPGADAPI    fpgaDAPI("FPGADAPI");
    ErrorMsg*   pErrorMsg = NO_ERROR_MSG;
    UINT32      u32Offset;

    // Check syntax
    in_parser.getParameterNumber(&count);
    if (count != 2)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    in_parser.getParameter(&strP1, 0);
    in_parser.getParameter(&strP2, 1);

    if (in_parser.isHEXString(strP1))
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    if ((pErrorMsg = fpgaDAPI.getErrorMsg()))
    {
        printError(pErrorMsg);
        return -1;
    }

    u32Offset = std::stoul(strP1, nullptr, 0);

    pErrorMsg = fpgaDAPI.readReg(strP2, u32Offset, &strValue);
    if (pErrorMsg)
    {
        printError(pErrorMsg);
        return -1;
    }

    cout << "0x" << strValue << endl;
    return 0;
}

int
FPGAOperation::executeUtilitySet(StringParser in_parser)
{
    int         count;
    string      strP1, strP2, strP3, strValue;
    FPGADAPI    fpgaDAPI("FPGADAPI");
    ErrorMsg*   pErrorMsg = NO_ERROR_MSG;
    UINT32      u32Offset;

    // Check syntax
    in_parser.getParameterNumber(&count);
    if (count != 3)
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    in_parser.getParameter(&strP1, 0);
    in_parser.getParameter(&strP2, 1);
    in_parser.getParameter(&strP3, 2);

    if (in_parser.isHEXString(strP1) ||
        in_parser.isHEXString(strP2))
    {
        cout << SYNTAX_ERROR_MSG << endl;
        return -1;
    }

    u32Offset = std::stoul(strP1, nullptr, 0);

    pErrorMsg = fpgaDAPI.writeReg(strP3, u32Offset, strP2);
    if (pErrorMsg)
    {
        printError(pErrorMsg);
        return -1;
    }

    return 0;
}

void
FPGAOperation::showOperationHelp()
{
    string strOpName, strHelp;

    getName(&strOpName);
    getHelp(&strHelp);

    cout << strHelp << endl;
    cout << "Syntax " << strOpName << " ";

    switch (getOperationType())
    {
        // Operations of Show
        case OPERATION_ALL:
        case OPERATION_DESIGN_ID:
        case OPERATION_FW:
        // Operations of SPAD
        case OPERATION_RANGE:
            cout << "P1" << endl;
            cout << "P1: Device name" << endl;
            break;
        case OPERATION_READ:
            cout << "P1 P2 P3" << endl;
            cout << "P1: Register offset (Hex)" << endl;
            cout << "P2: Count" << endl;
            cout << "P3: Device name" << endl;
            break;
        case OPERATION_WRITE:
            cout << "P1 P2 P3" << endl;
            cout << "P1: Register offset (Hex)" << endl;
            cout << "P2: Data" << endl;
            cout << "P3: Device name" << endl;
            break;
        // Operations of Utility
        case OPERATION_GET:
            cout << "P1 P2" << endl;
            cout << "P1: Register offset (Hex)" << endl;
            cout << "P2: Device name" << endl;
            break;
        case OPERATION_SET:
            cout << "P1 P2 P3" << endl;
            cout << "P1: Register offset (Hex)" << endl;
            cout << "P2: Value (Hex)" << endl;
            cout << "P3: Device name" << endl;
            break;
        default:
            break;
    }
}

int
FPGAOperation::showSPADContent(string in_strDeviceName,
                               UINT32 in_u32RegOffset,
                               UINT32 in_u32RegCount)
{
    FPGADAPI    fpgaDAPI("FPGADAPI");
    ErrorMsg*   pErrorMsg;
    UINT32      u32Offset = in_u32RegOffset;
    INT32       s32RegSizeInByte = 1;
    INT32       s32RegCountPerLine = 16;
    string      strValue;
    INT32       s32TempRegCount;
    std::stringstream stream;

    if ((pErrorMsg = fpgaDAPI.getErrorMsg()))
    {
        printError(pErrorMsg);
        return -1;
    }

    while (in_u32RegCount > 0)
    {
        // Show start address
        cout << "0x" << Int2HexStr(u32Offset) << ":";

        // Show data
        if (in_u32RegCount / s32RegCountPerLine)
        {
            // This line shows two scratch pad registers
            s32TempRegCount = s32RegCountPerLine;
            in_u32RegCount -= s32RegCountPerLine;
        }
        else
        {
            s32TempRegCount = in_u32RegCount;
            in_u32RegCount = 0;
        }

        // Handle one line
        stream.str("");
        stream.clear();
        while (s32TempRegCount--)
        {
            strValue.clear();
            pErrorMsg = fpgaDAPI.readSPAD(in_strDeviceName, u32Offset, &strValue);
            u32Offset += s32RegSizeInByte;
            if (pErrorMsg)
                break;

            stream << " ";
            stream << strValue;
        }

        cout << stream.str() << endl;

        if (pErrorMsg)
        {
            printError(pErrorMsg);
            return -1;
        }
    }

    return 0;
}
