/// @file       ErrorMsg.cpp
/// @brief      ErrorMsg class
/// @details    Provide base class for error message handling
/// @version    1.0.1
/// @date       2014.01.21
/// @author     KenHu
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.

#include "ErrorMsg.hpp"
#include <stdio.h>

/// @fn         ErrorMsg::ErrorMsg(UINT32 in_u32ErrorCode, const CHAR* in_szMessage)
/// @brief      ErrorMsg constructor
/// @param      in_u32ErrorCode is a 32bits error code
/// @param      in_szMessage is a pointer to user's specified char string
/// @exception  Nothing
/// @return     Nothing
ErrorMsg::ErrorMsg(UINT32       in_u32ErrorCode,
                   const CHAR*  in_szMessage)
{
    m_u32ErrorCode = in_u32ErrorCode;

    if (in_szMessage != NULL)
    {
        m_strMessage.append(in_szMessage);
    }

    m_pPrevious = this;
    m_pNext = NULL;
}

/// @overload   ErrorMsg::ErrorMsg(UINT32 in_u32ErrorCode, const string in_strMessage)
/// @param      in_u32ErrorCode is a 32bits error code
/// @param      in_strMessage is a string for user's specified message
/// @exception  Nothing
/// @return     Nothing
ErrorMsg::ErrorMsg(UINT32       in_u32ErrorCode,
                   const string in_strMessage)
{       
    m_u32ErrorCode = in_u32ErrorCode;

    if (!in_strMessage.empty())
    {
        m_strMessage.append(in_strMessage);
    }
    
    m_pPrevious = this;
    m_pNext = NULL;
}

/// @fn         ErrorMsg::~ErrorMsg(void)
/// @brief      ErrorMsg destructor
/// @exception  Nothing
/// @return     Nothing
ErrorMsg::~ErrorMsg(void)
{
    if (isListHead())
    {
        if (m_pNext != NULL)
        {
            m_pNext->m_pPrevious = m_pNext;
        }
    }
    else
    {
        m_pPrevious->m_pNext = m_pNext;

        if (m_pNext != NULL)
        {
            m_pNext->m_pPrevious = m_pPrevious;
        }
    }
}

/// @fn         ErrorMsg::appendToList(ErrorMsg* in_pNextErrorMsg)
/// @brief      Append ErrorMsg to the tail of the list
/// @param      in_pNextErrorMsg is a pointer to next ErrorMsg for appending to the list
/// @exception  Nothing
/// @return     Nothing
void
ErrorMsg::appendToList(ErrorMsg* in_pNextErrorMsg)
{
    ErrorMsg* pTempErrorMsg = this;

    if (in_pNextErrorMsg == NULL)
    {
        return;
    }

    // Find the last ErrorMsg
    while (pTempErrorMsg->m_pNext)
    {
        pTempErrorMsg = pTempErrorMsg->m_pNext;
    }

    pTempErrorMsg->m_pNext = in_pNextErrorMsg;
    in_pNextErrorMsg->m_pPrevious = pTempErrorMsg;
}

/// @fn         ErrorMsg::getErrorCode(void)
/// @brief      Get the error code
/// @exception  Nothing
/// @return     A 32bits error code
UINT32
ErrorMsg::getErrorCode(void)
{
    return m_u32ErrorCode;
}

/// @fn         ErrorMsg::getDetailMessage(void)
/// @brief      Get the specified message for detail error description
/// @exception  Nothing
/// @return     A string for detail error message
string
ErrorMsg::getDetailMessage(void)
{
    return m_strMessage;
}

/// @fn         ErrorMsg::getPreviousErrorMsg(void)
/// @brief      Get previous ErrorMsg for current ErrorMsg
/// @exception  Nothing
/// @return     A pointer to previous ErrorMsg
ErrorMsg*
ErrorMsg::getPreviousErrorMsg(void)
{
    return m_pPrevious;
}

/// @fn         ErrorMsg::getNextErrorMsg(void)
/// @brief      Get next ErrorMsg for current ErrorMsg
/// @exception  Nothing
/// @return     A pointer to next ErrorMsg
ErrorMsg*
ErrorMsg::getNextErrorMsg(void)
{
    return m_pNext;
}

/// @fn         ErrorMsg::isListHead(void)
/// @brief      Check if current ErrorMsg is the head of the list
/// @exception  Nothing
/// @return     A boolean to indicate if current ErrorMsg is the head of the list
BOOL
ErrorMsg::isListHead(void)
{
    if (m_pPrevious == this)
    {
        return TRUE;
    }
    else
    {
        return FALSE;
    }
}
