/// @file       MemoryDAPI_MarchCTest.cpp
/// @brief      Provide MarchC test for MemoryDAPI
/// @details    Verify memory by March C algorithm
/// @author     KenHu
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.

#include "MemoryDAPI.hpp"

/// @fn         MemoryDAPI::marchCTest(PHYSICAL_ADDRESS_T in_startAddress,
///                                    PHYSICAL_ADDRESS_T in_endAddress,
///                                    INT32              in_s32ElementSize,
///                                    UINT64             in_u64UserTestPattern)
/// @brief      Verify memory by March C algorithm
/// @param      in_startAddress is a physcial address for start address
/// @param      in_endAddress is a physcial address for end address
/// @param      in_s32ElementSize is 32bits value for element size (ex: 8/16/32/64)
/// @param      in_u64UserTestPattern is a 64bits test pattern
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg* 
MemoryDAPI::marchCTest(PHYSICAL_ADDRESS_T in_startAddress,
                       PHYSICAL_ADDRESS_T in_endAddress,
                       INT32              in_s32ElementSize,
                       UINT64             in_u64UserTestPattern)
{
    ErrorMsg*           pErrorMsg = NULL;
    PHYSICAL_ADDRESS_T  startAddress = 0;
    PHYSICAL_ADDRESS_T  endAddress = 0;
    PHYSICAL_ADDRESS_T  currentAddress = 0;
    PHYSICAL_ADDRESS_T  memoryBlockSize = 0;    
    PHYSICAL_ADDRESS_T  step = (PHYSICAL_ADDRESS_T)(in_s32ElementSize >> 3);
    UINT64              u64InverseTestPattern = (UINT64)(~in_u64UserTestPattern);
    BOOL                bIsFailed = FALSE;
    UINT32              u32FinishedStep = 0;
    UINT32              u32MapSize = m_pMemoryDDAPI->getMemoryMapSize();

    switch (in_s32ElementSize)
    {
        case 8:
        case 16:
        case 32:
        case 64:
            // Round up start address for alignment
            startAddress = alignMemoryAddress(in_startAddress, in_s32ElementSize);
            
            // Modify end address to point to the last accessible address
            if (in_endAddress & (PHYSICAL_ADDRESS_T)(step - 1))
            {
                endAddress = in_endAddress & ~((PHYSICAL_ADDRESS_T)(step - 1));
                endAddress -= step;
            }
            else
            {
                endAddress = in_endAddress - step;
            }            

            memoryBlockSize = endAddress - startAddress;
            break;
        default:
            return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_MARCHC_TEST | ERROR_BAD_PARAMETER_THREE), 
                                  "Unknown element size",
                                  NULL);
            break;

    }

    initProgress(memoryBlockSize * 6);
    
    currentAddress = startAddress;

    // Write inverse pattern with forward addressing
    while (currentAddress <= endAddress)
    {
        switch (in_s32ElementSize)
        {
            case 8:
                pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT8)u64InverseTestPattern);
                break;                
            case 16:
                pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT16)u64InverseTestPattern);
                break;                
            case 32:
                pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT32)u64InverseTestPattern);
                break;                
            case 64:
                pErrorMsg = m_pMemoryDDAPI->write(currentAddress, u64InverseTestPattern);
                break;
        }
        
        if (pErrorMsg != NO_ERROR_MSG)
        {
            initProgress(0);
            
            return returnErrorMsg(ERROR_MEMORY_DAPI_MARCHC_TEST, NULL, pErrorMsg);
        }

        currentAddress += step;
        
        u32FinishedStep += step;

        if ((u32FinishedStep % u32MapSize) == 0)
        {
            updateProgress(u32FinishedStep);

            u32FinishedStep = 0;
        }
    }

    if (u32FinishedStep)
    {
        updateProgress(u32FinishedStep);

        u32FinishedStep = 0;        
    }
    
    // 1st March operation for reading inverse pattern and writing pattern with forward addressing
    if (pErrorMsg == NO_ERROR_MSG)
    {
        pErrorMsg = verfiyMemoryBlockByMarchC(startAddress, endAddress, in_s32ElementSize, u64InverseTestPattern, in_u64UserTestPattern);
    }

    // 2nd March operation for reading pattern and writing inverse pattern with forward addressing
    if (pErrorMsg == NO_ERROR_MSG)
    {
        pErrorMsg = verfiyMemoryBlockByMarchC(startAddress, endAddress, in_s32ElementSize, in_u64UserTestPattern, u64InverseTestPattern);
    }

    // 3rd March operation for reading inverse pattern and writing pattern with backward addressing
    if (pErrorMsg == NO_ERROR_MSG)
    {
        pErrorMsg = verfiyMemoryBlockByMarchC(endAddress, startAddress, in_s32ElementSize, u64InverseTestPattern, in_u64UserTestPattern);
    }
    
    // 4th March operation for reading writing pattern and inverse pattern with backward addressing
    if (pErrorMsg == NO_ERROR_MSG)
    {
        pErrorMsg = verfiyMemoryBlockByMarchC(endAddress, startAddress, in_s32ElementSize, in_u64UserTestPattern, u64InverseTestPattern);
    } 
    
    // Read inverse pattern with forward addressing
    if (pErrorMsg == NO_ERROR_MSG)
    {
        currentAddress = startAddress;
        while (currentAddress <= endAddress)
        {
            switch (in_s32ElementSize)
            {
                case 8:
                    if ((UINT8)u64InverseTestPattern != m_pMemoryDDAPI->read8(currentAddress))
                    {
                        bIsFailed = TRUE;
                    }
                    break;                
                case 16:
                    if ((UINT16)u64InverseTestPattern != m_pMemoryDDAPI->read16(currentAddress))
                    {
                        bIsFailed = TRUE;
                    }
                    break;                
                case 32:
                    if ((UINT32)u64InverseTestPattern != m_pMemoryDDAPI->read32(currentAddress))
                    {
                        bIsFailed = TRUE;
                    }
                    break;                
                case 64:
                    if (u64InverseTestPattern != m_pMemoryDDAPI->read64(currentAddress))
                    {
                        bIsFailed = TRUE;
                    }
                    break;
            }            
        
            if (bIsFailed)
            {
                pErrorMsg = m_pMemoryDDAPI->getErrorMsg();
                if (pErrorMsg == NO_ERROR_MSG)
                {
                    initProgress(0);
                    
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_MARCHC_TEST | ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL), NULL, NULL);
                }
                else
                {
                    initProgress(0);
                    
                    return returnErrorMsg(ERROR_MEMORY_DAPI_MARCHC_TEST, NULL, pErrorMsg);
                }
            }

            currentAddress += step;

            u32FinishedStep += step;
            
            if ((u32FinishedStep % u32MapSize) == 0)
            {
                updateProgress(u32FinishedStep);

                u32FinishedStep = 0;
            }
        }

        if (u32FinishedStep)
        {
            updateProgress(u32FinishedStep);

            u32FinishedStep = 0;        
        }        
    }

    if (pErrorMsg != NO_ERROR_MSG)
    {
        initProgress(0);
        
        return returnErrorMsg(ERROR_MEMORY_DAPI_MARCHC_TEST, NULL, pErrorMsg);
    }
  
    return NO_ERROR_MSG;
}

/// @fn         MemoryDAPI::executeMarchCTest(PHYSICAL_ADDRESS_T in_startAddress,
///                                           PHYSICAL_ADDRESS_T in_endAddress,
///                                           INT32              in_s32ElementSize,
///                                           UINT64             in_u64UserTestPattern)
/// @brief      Execute March C- test
/// @param      in_startAddress is a physcial address for start address
/// @param      in_endAddress is a physcial address for end address
/// @param      in_s32ElementSize is 32bits value for element size (ex: 8/16/32/64)
/// @param      in_u64VerifyTestPattern is a 64bits test pattern for verification
/// @param      in_u64WriteTestPattern is a 64bits test pattern for writing
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg* 
MemoryDAPI::verfiyMemoryBlockByMarchC(PHYSICAL_ADDRESS_T in_startAddress,
                                      PHYSICAL_ADDRESS_T in_endAddress,
                                      INT32              in_s32ElementSize,
                                      UINT64             in_u64VerifyTestPattern,
                                      UINT64             in_u64WriteTestPattern)
{
    ErrorMsg*           pErrorMsg = NULL;
    PHYSICAL_ADDRESS_T  currentAddress = in_startAddress;
    PHYSICAL_ADDRESS_T  step = (PHYSICAL_ADDRESS_T)(in_s32ElementSize >> 3);
    BOOL                bIsFailed = FALSE;
    UINT32              u32FinishedStep = 0;
    UINT32              u32MapSize = m_pMemoryDDAPI->getMemoryMapSize();    
    
    if (in_startAddress < in_endAddress)
    {
        while (currentAddress <= in_endAddress)
        {
            switch (in_s32ElementSize)
            {
                case 8:
                    if ((UINT8)in_u64VerifyTestPattern != m_pMemoryDDAPI->read8(currentAddress))
                    {
                        bIsFailed = TRUE;
                    }                    
                    break;                      
                case 16:
                    if ((UINT16)in_u64VerifyTestPattern != m_pMemoryDDAPI->read16(currentAddress))
                    {
                        bIsFailed = TRUE;
                    }
                    break;                      
                case 32:
                    if ((UINT32)in_u64VerifyTestPattern != m_pMemoryDDAPI->read32(currentAddress))
                    {
                        bIsFailed = TRUE;
                    }
                    break;                    
                case 64:
                    if (in_u64VerifyTestPattern != m_pMemoryDDAPI->read64(currentAddress))
                    {
                        bIsFailed = TRUE;
                    }
                    break;
            }           

            if (bIsFailed)
            {
                pErrorMsg = m_pMemoryDDAPI->getErrorMsg();
                if (pErrorMsg == NO_ERROR_MSG)
                {
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK_BY_MARCHC | ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL),
                                           NULL,
                                           NULL);
                }
                else
                {
                    return returnErrorMsg(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK_BY_MARCHC, NULL, pErrorMsg);
                }
            }

            switch (in_s32ElementSize)
            {
                case 8:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT8)in_u64WriteTestPattern);                    
                    break;                      
                case 16:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT16)in_u64WriteTestPattern);
                    break;                      
                case 32:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT32)in_u64WriteTestPattern);
                    break;                    
                case 64:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, in_u64WriteTestPattern);
                    break;
            } 

            if (pErrorMsg != NO_ERROR_MSG)
            {
                return returnErrorMsg(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK_BY_MARCHC, NULL, pErrorMsg);
            } 

            currentAddress += step;

            u32FinishedStep += step;
            
            if ((u32FinishedStep % u32MapSize) == 0)
            {
                updateProgress(u32FinishedStep);

                u32FinishedStep = 0;
            }
        }

        if (u32FinishedStep)
        {
            updateProgress(u32FinishedStep);

            u32FinishedStep = 0;        
        }
    }
    else
    {
        while (currentAddress >= in_endAddress)
        {
            switch (in_s32ElementSize)
            {
                case 8:
                    if ((UINT8)in_u64VerifyTestPattern != m_pMemoryDDAPI->read8(currentAddress))
                    {
                        bIsFailed = TRUE;
                    }                    
                    break;                      
                case 16:
                    if ((UINT16)in_u64VerifyTestPattern != m_pMemoryDDAPI->read16(currentAddress))
                    {
                        bIsFailed = TRUE;
                    }
                    break;                      
                case 32:
                    if ((UINT32)in_u64VerifyTestPattern != m_pMemoryDDAPI->read32(currentAddress))
                    {
                        bIsFailed = TRUE;
                    }
                    break;                    
                case 64:
                    if (in_u64VerifyTestPattern != m_pMemoryDDAPI->read64(currentAddress))
                    {
                        bIsFailed = TRUE;
                    }
                    break;
            }           

            if (bIsFailed)
            {
                pErrorMsg = m_pMemoryDDAPI->getErrorMsg();
                if (pErrorMsg == NO_ERROR_MSG)
                {
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK_BY_MARCHC | ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL),
                                           NULL,
                                           NULL);
                }
                else
                {
                    return returnErrorMsg(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK_BY_MARCHC, NULL, pErrorMsg);
                }
            }
            
            switch (in_s32ElementSize)
            {
                case 8:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT8)in_u64WriteTestPattern);                    
                    break;                      
                case 16:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT16)in_u64WriteTestPattern);
                    break;                      
                case 32:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, (UINT32)in_u64WriteTestPattern);
                    break;                    
                case 64:
                    pErrorMsg = m_pMemoryDDAPI->write(currentAddress, in_u64WriteTestPattern);
                    break;
            } 

            if (pErrorMsg != NO_ERROR_MSG)
            {
                return returnErrorMsg(ERROR_MEMORY_DAPI_VERIFY_MEMORY_BLOCK_BY_MARCHC, NULL, pErrorMsg);
            } 

            currentAddress -= step;

            u32FinishedStep += step;
            
            if ((u32FinishedStep % u32MapSize) == 0)
            {
                updateProgress(u32FinishedStep);

                u32FinishedStep = 0;
            }
        }

        if (u32FinishedStep)
        {
            updateProgress(u32FinishedStep);

            u32FinishedStep = 0;        
        } 
    }

    return NO_ERROR_MSG;    
}

