/// @file       MemoryDAPI_PatternTest.cpp
/// @brief      Provide pattern test for MemoryDAPI
/// @details    Verify memory by write and read test pattern
/// @author     KenHu
/// @copyright  Copyright 2015-2016 Hon Hai Technology Group. All Rights Reserved.

#include "MemoryDAPI.hpp"
#include <unistd.h>
#include <sys/mman.h>
#include <fcntl.h>

/// @fn         MemoryDAPI::patternTest(PHYSICAL_ADDRESS_T in_startAddress,
///                                     PHYSICAL_ADDRESS_T in_endAddress,
///                                     INT32              in_s32ElementSize
///                                     UINT64             in_u64UserTestPattern)
/// @brief      Verify memory by a test pattern
/// @param      in_startAddress is a physical address for start address
/// @param      in_endAddress is a physical address for end address
/// @param      in_s32ElementSize is 32bits value for element size (ex: 8/16/32/64)
/// @param      in_u64UserTestPattern is a 64bits test pattern
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg* 
MemoryDAPI::patternTest(PHYSICAL_ADDRESS_T in_startAddress,
                        PHYSICAL_ADDRESS_T in_endAddress,
                        INT32              in_s32ElementSize,
                        UINT64             in_u64UserTestPattern)
{
    ErrorMsg*           pErrorMsg = NULL;
    PHYSICAL_ADDRESS_T  testBlockSize = (PHYSICAL_ADDRESS_T)m_pMemoryDDAPI->getMemoryMapSize();
    PHYSICAL_ADDRESS_T  memoryBlockSize = 0;
    PHYSICAL_ADDRESS_T  verifiedMemorySize = 0;
    PHYSICAL_ADDRESS_T  startAddress = in_startAddress;
    PHYSICAL_ADDRESS_T  endAddress = in_endAddress;
    PHYSICAL_ADDRESS_T  currentAddress = in_startAddress;

    switch (in_s32ElementSize)
    {
        case 8:
        case 16:
        case 32:
        case 64:
            // Round up memory address for alignment
            startAddress = alignMemoryAddress(startAddress, in_s32ElementSize);

            memoryBlockSize = endAddress - startAddress;
            break;
        default:
            return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_PATTERN_TEST | ERROR_BAD_PARAMETER_THREE),
                                  "invalid data line amount",
                                  NULL);
            break;

    }
    
    if (memoryBlockSize == 0)
    {
        return NO_ERROR_MSG;
    }

    initProgress(memoryBlockSize);
    
    while (verifiedMemorySize < memoryBlockSize)
    {
        // The last test memory block size may be smaller than memory map size 
        if ((memoryBlockSize - verifiedMemorySize) < testBlockSize)
        {
            testBlockSize = memoryBlockSize - verifiedMemorySize;
        }
        
        // Write, read and verify test pattern for test block size
        switch (in_s32ElementSize)
        {
            case 8:
                pErrorMsg = verifyMemoryBlock(currentAddress, 
                                              (currentAddress + testBlockSize),
                                              (UINT8)in_u64UserTestPattern);
                break;
            case 16:
                pErrorMsg = verifyMemoryBlock(currentAddress, 
                                              (currentAddress + testBlockSize),
                                              (UINT16)in_u64UserTestPattern);
                break;
            case 32:
                pErrorMsg = verifyMemoryBlock(currentAddress, 
                                              (currentAddress + testBlockSize),
                                              (UINT32)in_u64UserTestPattern);
                break;
            case 64:
                pErrorMsg = verifyMemoryBlock(currentAddress, 
                                              (currentAddress + testBlockSize),
                                              (UINT64)in_u64UserTestPattern);
                break;
        }

        if (pErrorMsg != NO_ERROR_MSG)
        {
            initProgress(0);
            return returnErrorMsg(ERROR_MEMORY_DAPI_PATTERN_TEST, NULL, pErrorMsg);
        }

        currentAddress += testBlockSize;
        verifiedMemorySize += testBlockSize;

        updateProgress(testBlockSize);
    }

    return NO_ERROR_MSG;
}

/// @fn         MemoryDAPI::patternTest(PHYSICAL_ADDRESS_T in_startAddress,
///                                     PHYSICAL_ADDRESS_T in_endAddress)
/// @brief      Verify memory by default test pattern
/// @param      in_startAddress is a physical address for start address
/// @param      in_endAddress is a physical address for end address
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDAPI::patternTest(PHYSICAL_ADDRESS_T in_startAddress,
                        PHYSICAL_ADDRESS_T in_endAddress)
{
    PHYSICAL_ADDRESS_T  testBlockSize = 0, memoryBlockSize = 0, verifiedMemorySize = 0;
    PHYSICAL_ADDRESS_T  startAddress, currentAddress;
    SYSTEM_ADDRESS_T    testAddress, endAddress, mapAddress;
    string              strErrorMsg;
    INT32               s32FD;
    PHYSICAL_ADDRESS_T  pageSize = getpagesize();

    // Round up memory address for alignment
    startAddress = alignMemoryAddress(in_startAddress, 64);

    // Align memory address for page size
    if (startAddress & (pageSize - 1))
    {
        startAddress &= ~(pageSize - 1) + pageSize;
    }

    currentAddress = startAddress;

    memoryBlockSize = in_endAddress - startAddress;

    if (memoryBlockSize == 0)
    {
        return NO_ERROR_MSG;
    }

    initProgress(memoryBlockSize);

    s32FD = open("/dev/mem", O_RDWR | O_SYNC);

    testBlockSize = getTestBlockSize();
    while (verifiedMemorySize < memoryBlockSize)
    {
        if ((memoryBlockSize - verifiedMemorySize) < testBlockSize)
        {
            testBlockSize = memoryBlockSize - verifiedMemorySize;
        }

        mapAddress = (SYSTEM_ADDRESS_T)mmap64(0, testBlockSize, (PROT_READ | PROT_WRITE), MAP_SHARED, s32FD, currentAddress);
        testAddress = mapAddress;
        endAddress = mapAddress + testBlockSize;
        while (testAddress < endAddress)
        {
            // Verify by all 0's pattern
            *((UINT64 *)testAddress) = 0x0ULL;
            if ((0x0ULL) != *((UINT64 *)testAddress))
            {
                munmap((void *)mapAddress, testBlockSize);
                close(s32FD);
                initProgress(0);
                strErrorMsg = "failed address is " + UINT64ToHexStr(currentAddress + (testAddress - mapAddress));
                return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_PATTERN_TEST | ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL),
                                                  strErrorMsg.c_str(),
                                                  NULL);
            }

            // Verify by all 1's pattern
            *((UINT64 *)testAddress) = 0xFFFFFFFFFFFFFFFFULL;
            if ((0xFFFFFFFFFFFFFFFFULL) != *((UINT64 *)testAddress))
            {
                munmap((void *)mapAddress, testBlockSize);
                close(s32FD);
                initProgress(0);
                strErrorMsg = "failed address is " + UINT64ToHexStr(currentAddress + (testAddress - mapAddress));
                return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DAPI_PATTERN_TEST | ERROR_MEMORY_DAPI_TEST_PATTERN_COMPARE_FAIL),
                                                  strErrorMsg.c_str(),
                                                  NULL);
            }

            testAddress += 8;
            updateProgress(8);
        }

        munmap((void *)mapAddress, testBlockSize);
        currentAddress += testBlockSize;
        verifiedMemorySize += testBlockSize;
    }

    close(s32FD);

    return NO_ERROR_MSG;
}
