/// @file       ControllerDDAPI.hpp
/// @brief      ControllerDDAPI class
/// @details    Provide low layer diagnostic and utility APIs for Controller module
/// @author     KenHu
/// @copyright  Copyright 2015-2016 Hon Hai Technology Group. All Rights Reserved.

#include "ControllerDDAPI.hpp"
#include "GPIODDAPI.hpp"
#include "FPGADDAPI.hpp"

/// @fn         ControllerDDAPI::ControllerDDAPI() : BaseClassDDAPI("ControllerDDAPI")
/// @brief      ControllerDDAPI constructor
/// @exception  Nothing
/// @return     Nothing
ControllerDDAPI::ControllerDDAPI() : BaseClassDDAPI("ControllerDDAPI")
{

}

/// @fn         ControllerDDAPI::~ControllerDDAPI(void)
/// @brief      ControllerDDAPI destructor
/// @exception  Nothing
/// @return     Nothing
ControllerDDAPI::~ControllerDDAPI(void)
{

}

/// @fn         PSUDDAPI::getLEDList(list<string> *out_pList)
///
/// @brief      Get LED name list
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
ControllerDDAPI::getLEDList(list<string> *out_pLEDList)
{
    string strModuleName, strTemp;

    if (out_pLEDList == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_CONTROLLER_DDAPI_GET_LED_LIST | ERROR_BAD_PARAMETER),
                              "NULL pointer");
    }

    strModuleName = CONTROLLER_MODULE_NAME;
    strModuleName += "_";
    strTemp = strModuleName + CONTROLLER_ACTIVITY_LED;
    out_pLEDList->push_back(strTemp);
    strTemp = strModuleName + CONTROLLER_FAULT_LED;
    out_pLEDList->push_back(strTemp);

    return NO_ERROR_MSG;
}

/// @fn         ControllerDDAPI::getLEDStatus(string in_strLEDName,
///                                           string *out_strpStatus)
///
/// @brief      Get status of the given LED
/// @param      in_strLEDName is a string for the LED name
/// @param      out_strpStatus is a pointer to a string for the status
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
ControllerDDAPI::getLEDStatus(string in_strLEDName,
                              string *out_strpStatus)
{
    string      strModuleName = CONTROLLER_MODULE_NAME;
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    FPGADDAPI   *pFPGADDAPI = NULL;
    UINT8       u8Value = 0;

    strModuleName += "_";
    in_strLEDName.erase(0, strModuleName.size());

    if ((pFPGADDAPI = new FPGADDAPI()) == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_CONTROLLER_DDAPI_GET_LED_STATUS | ERROR_CREATE_OBJECT_FAIL),
                              "fail to new FPGADDAPI");
    }
    else if ((pErrorMsg = pFPGADDAPI->getErrorMsg()) != NO_ERROR_MSG)
    {
        delete pFPGADDAPI;
        return returnErrorMsg(ERROR_CONTROLLER_DDAPI_GET_LED_STATUS, NULL, pErrorMsg);
    }

    // Read LED Status Register
    pErrorMsg = pFPGADDAPI->read(FPGA_DEVICE_NAME, 0x12, &u8Value);
    delete pFPGADDAPI;

    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_CONTROLLER_DDAPI_GET_LED_STATUS, NULL, pErrorMsg);
    }

    if (in_strLEDName == CONTROLLER_ACTIVITY_LED)
    {
        if (u8Value & 0x1) // Bit 0 is Controller Active LED
        {
            *out_strpStatus = "on";
        }
        else
        {
            *out_strpStatus = "off";
        }
    }
    else if (in_strLEDName == CONTROLLER_FAULT_LED)
    {
        if (u8Value & 0x2) // Bit 1 is Controller Attention LED
        {
            *out_strpStatus = "on";
        }
        else
        {
            *out_strpStatus = "off";
        }
    }
    else
    {
        return returnErrorMsg((ERRORCODE)(ERROR_CONTROLLER_DDAPI_GET_LED_STATUS | ERROR_NO_DEVICE_FOUND),
                              in_strLEDName.c_str());
    }

    return NO_ERROR_MSG;
}

/// @fn         ControllerDDAPI::getSlotID(UINT32 *out_u32pSlotID)
///
/// @brief      Get slot index of the controller
/// @param      out_u32pSlotID is a pointer to a UINT32 for the slot index
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
ControllerDDAPI::getSlotID(UINT32 *out_u32pSlotID)
{
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    GPIODDAPI   gpioDDAPI("GPIODDAPI");
    BOOL        bGPIOValue = FALSE;

    if ((pErrorMsg = gpioDDAPI.getErrorMsg()) != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_CONTROLLER_DDAPI_GET_SLOT_ID, NULL, pErrorMsg);
    }

    pErrorMsg = gpioDDAPI.readBit(SLOT_GPIO_NAME, &bGPIOValue);
    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_CONTROLLER_DDAPI_GET_SLOT_ID, NULL, pErrorMsg);
    }

    *out_u32pSlotID = (UINT32)bGPIOValue;
    return NO_ERROR_MSG;
}

/// @fn         ControllerDDAPI::setLEDStatus(string in_strDeviceName,
///                                           string in_strStatus)
///
/// @brief      Set status to the given LED
/// @param      in_strLEDName is a string for the LED
/// @param      in_strStatus is a string for the status
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
ControllerDDAPI::setLEDStatus(string in_strLEDName,
                              string in_strStatus)
{
    string      strModuleName = CONTROLLER_MODULE_NAME;
    ErrorMsg    *pErrorMsg = NO_ERROR_MSG;
    FPGADDAPI   *pFPGADDAPI = NULL;
    UINT8       u8ControlBit, u8Value = 0, u8Temp = 0;
    
    strModuleName += "_";
    in_strLEDName.erase(0, strModuleName.size());

    if (in_strLEDName == CONTROLLER_ACTIVITY_LED)
    {
        u8ControlBit = 0x10;

        if (in_strStatus == "on")
        {
            u8Value = 0x10;
        }
        else if (in_strStatus == "off")
        {
            u8Value = 0x00;
        }
        else
        {
            return returnErrorMsg((ERRORCODE)(ERROR_CONTROLLER_DDAPI_SET_LED_STATUS | ERROR_UNEXPECTED_VALUE),
                                  in_strStatus.c_str());
        }
    }
    else if (in_strLEDName == CONTROLLER_FAULT_LED)
    {
        u8ControlBit = 0x20;

        if (in_strStatus == "on")
        {
            u8Value = 0x20;
        }
        else if (in_strStatus == "off")
        {
            u8Value = 0x00;
        }
        else
        {
            return returnErrorMsg((ERRORCODE)(ERROR_CONTROLLER_DDAPI_SET_LED_STATUS | ERROR_UNEXPECTED_VALUE),
                                  in_strStatus.c_str());
        }
    }
    else
    {
        return returnErrorMsg((ERRORCODE)(ERROR_CONTROLLER_DDAPI_SET_LED_STATUS | ERROR_NO_DEVICE_FOUND),
                              in_strLEDName.c_str());
    }

    u8Value |= 0x3; // Add mask bits for both of LEDs

    if ((pFPGADDAPI = new FPGADDAPI()) == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_CONTROLLER_DDAPI_SET_LED_STATUS | ERROR_CREATE_OBJECT_FAIL),
                              "fail to new FPGADDAPI");
    }
    else if ((pErrorMsg = pFPGADDAPI->getErrorMsg()) != NO_ERROR_MSG)
    {
        delete pFPGADDAPI;
        return returnErrorMsg(ERROR_CONTROLLER_DDAPI_SET_LED_STATUS, NULL, pErrorMsg);
    }

    // Read LED Status Register
    if ((pErrorMsg = pFPGADDAPI->read(FPGA_DEVICE_NAME, 0x12, &u8Temp)) == NO_ERROR_MSG)
    {   
        // Merge original LED setting
        u8Value |= ((u8Temp << 4) & ~u8ControlBit);
        
        // Write LED Mask and Set Register
        pErrorMsg = pFPGADDAPI->write(FPGA_DEVICE_NAME, 0x10, u8Value);
    }

    delete pFPGADDAPI;

    if (pErrorMsg != NO_ERROR_MSG)
    {
        return returnErrorMsg(ERROR_CONTROLLER_DDAPI_SET_LED_STATUS, NULL, pErrorMsg);
    }

    return NO_ERROR_MSG;
}
