/// @file       MemoryDDAPI.cpp
/// @brief      MemoryDDAPI class
/// @details    Provide low layer diagnostic and utility APIs for memory module
/// @version    1.0.0
/// @date       2014.01.29
/// @author     KenHu
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.

#include <unistd.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <string>
#include <sstream>
#include <fstream>
#include <dirent.h>

#include "MemoryDDAPI.hpp"
#include "CommonFunctions.hpp"

#define DIMM_SPD_BYTE_AMOUNT    256

CHAR    g_szMCDrvIOCTLResult[FXN_MC_DRV_RESULT_STRING_LEN];

/// @fn         MemoryDDAPI::MemoryDDAPI(STRING in_szClassName) : BaseClassDDAPI(in_szClassName)
/// @brief      MemoryDDAPI constructor
/// @param      in_szClassName is a char string for class name
/// @exception  Nothing
/// @return     Nothing
MemoryDDAPI::MemoryDDAPI(STRING in_szClassName) : BaseClassDDAPI(in_szClassName)
{
    m_memoryMapInformation.s32FileDescriptor = -1;
    m_memoryMapInformation.phyStartAddress = (PHYSICAL_ADDRESS_T)NULL;
    m_memoryMapInformation.mapSize = 0;
    m_memoryMapInformation.mapAddress = (SYSTEM_ADDRESS_T)NULL;
    m_s32FXNMCDrvFD = -1;
}

/// @fn         MemoryDDAPI::~MemoryDDAPI(void)
/// @brief      MemoryDDAPI destructor
/// @exception  Nothing
/// @return     Nothing
MemoryDDAPI::~MemoryDDAPI(void)
{
    list<ECCDIMMInformation *>::iterator  it;

    if (m_memoryMapInformation.mapAddress != 0)
    {
        // unmap previous mapped page
        munmap((void*)m_memoryMapInformation.mapAddress, m_memoryMapInformation.mapSize);
        m_memoryMapInformation.mapAddress = (SYSTEM_ADDRESS_T)NULL;
    }

    if (m_memoryMapInformation.s32FileDescriptor >= 0)
    {
        close(m_memoryMapInformation.s32FileDescriptor);
    }
    
    if (m_s32FXNMCDrvFD >= 0)
    {
        close(m_s32FXNMCDrvFD);
    }
}

/// @fn         MemoryDDAPI::clearECCReport(void)
/// @brief      Clear ECC report of memory controller
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDDAPI::clearECCReport(void)
{
    DIR*            pDir;
    struct dirent*  pDirEntry;
    string          strTemp;

    if ((pDir = opendir(EDAC_SYS_FILE_PATH)) == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_CLEAR_ECC_REPORT | ERROR_OPEN_DIRECTORY),
                              EDAC_SYS_FILE_PATH,
                              NULL);
    }

    while ((pDirEntry = readdir(pDir)) != NULL)
    {
        if (!(pDirEntry->d_type & DT_DIR))
        {
            continue;
        }

        if (strstr(pDirEntry->d_name, "mc") == NULL)
        {
            continue;
        }

        strTemp = "echo 1 > ";
        strTemp += EDAC_SYS_FILE_PATH;
        strTemp += pDirEntry->d_name;
        strTemp += "/reset_counters";

        system(strTemp.c_str());
    }

    if (strTemp.empty())
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_CLEAR_ECC_REPORT | ERROR_MEMORY_DDAPI_EDAC_DEVICE_NOT_FOUND));
    }

    return NO_ERROR_MSG;
}

/// @fn         MemoryDDAPI::getControllerChannelCount(INT32* out_ps32Count)
/// @brief      Get the amount of slots of each channel
/// @param      out_ps32Count is a pointer to a INT32 value for the amount of channels
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDDAPI::getControllerChannelCount(INT32* out_s32pCount)
{
    string      strFile;
    string      strTemp;
    FILE*       pFile = NULL;
    CHAR        szTemp[3];

    if (out_s32pCount == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_CONTROLLER_CHANNEL_COUNT | ERROR_BAD_PARAMETER),
                              "NULL pointer");
    }

    strFile = LEMUR_CONFIG_DIR;
    strFile += MEMORY_CONF_FILE;

    strTemp = "cat " + strFile;
    strTemp += "|awk -F: '/CHANNEL_NUM/'|grep -iv \"#\"|tr -d ' '|awk 'BEGIN {FS=\"=\"}{print $2}'";

    if ((pFile = popen(strTemp.c_str(), "r")) == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_CONTROLLER_CHANNEL_COUNT | ERROR_OPEN_FILE),
                              strTemp.c_str());
    }

    if (fgets(szTemp, sizeof(szTemp), pFile) == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_CONTROLLER_CHANNEL_COUNT | ERROR_CONFIGURATION_NOT_FOUND));
    }

    pclose(pFile);

    *out_s32pCount = atoi(szTemp);

    if (*out_s32pCount > 0)
    {
        return NO_ERROR_MSG;
    }
    else
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_CONTROLLER_CHANNEL_COUNT | ERROR_UNEXPECTED_VALUE),
                              "amount of channels must be larger than 0");
    }
}

/// @fn         MemoryDDAPI::getControllerSlotCount(INT32* out_ps32Count)
/// @brief      Get the amount of slots of each channel
/// @param      out_ps32Count is a pointer to a INT32 value for the amount of slots
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDDAPI::getControllerSlotCount(INT32* out_s32pCount)
{
    string      strFile;
    string      strTemp;
    FILE*       pFile = NULL;
    CHAR        szTemp[3];

    if (out_s32pCount == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_CONTROLLER_SLOT_COUNT | ERROR_BAD_PARAMETER),
                              "NULL pointer");
    }

    strFile = LEMUR_CONFIG_DIR;
    strFile += MEMORY_CONF_FILE;

    strTemp = "cat " + strFile;
    strTemp += "|awk -F: '/CHANNEL_SLOT_NUM/'|grep -iv \"#\"|tr -d ' '|awk 'BEGIN {FS=\"=\"}{print $2}'";

    if ((pFile = popen(strTemp.c_str(), "r")) == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_CONTROLLER_SLOT_COUNT | ERROR_OPEN_FILE),
                              strTemp.c_str());
    }

    if (fgets(szTemp, sizeof(szTemp), pFile) == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_CONTROLLER_SLOT_COUNT | ERROR_CONFIGURATION_NOT_FOUND));
    }

    pclose(pFile);

    *out_s32pCount = atoi(szTemp);

    if (*out_s32pCount > 0)
    {
        return NO_ERROR_MSG;
    }
    else
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_CONTROLLER_SLOT_COUNT | ERROR_UNEXPECTED_VALUE),
                              "amount of slots must be larger than 0");
    }
}

/// @fn         MemoryDDAPI::getECCReport(string* out_strpECCReport)
/// @brief      Get ECC report from memory controller
/// @param      out_pErrorList is a pointer to a list of ECCInformation
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDDAPI::getECCReport(list<ECCInformation>* out_pErrorList)
{
    ErrorMsg*       pErrorMsg = NO_ERROR_MSG;
    INT32           s32Channels = 1, s32Slots = 1;
    FILE*           pFile = NULL;
    CHAR            szTemp[128];
    DIR*            pDir;
    struct dirent*  pDirEntry;
    INT32           s32ControllerID = 0, s32ChannelID = 0, s32SlotID = 0, s32ErrorCount;
    ECCInformation  errorInformation;

    if (out_pErrorList == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_ECC_REPORT | ERROR_BAD_PARAMETER),
                              "NULL pointer");
    }

    out_pErrorList->clear();

    // Get information about controller
    pErrorMsg = getControllerChannelCount(&s32Channels);
    if (pErrorMsg)
    {
        return returnErrorMsg(ERROR_MEMORY_DDAPI_GET_ECC_REPORT, NULL, pErrorMsg);
    }

    pErrorMsg = getControllerSlotCount(&s32Slots);
    if (pErrorMsg)
    {
        return returnErrorMsg(ERROR_MEMORY_DDAPI_GET_ECC_REPORT, NULL, pErrorMsg);
    }

    // Get error counts from edac sys files
    if ((pDir = opendir(EDAC_SYS_FILE_PATH)) == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_ECC_REPORT | ERROR_OPEN_DIRECTORY),
                              EDAC_SYS_FILE_PATH,
                              NULL);
    }

    while ((pDirEntry = readdir(pDir)) != NULL)
    {
        if (!(pDirEntry->d_type & DT_DIR))
        {
            continue;
        }

        if (strstr(pDirEntry->d_name, "mc") == NULL)
        {
            continue;
        }

        sscanf(pDirEntry->d_name, "mc%d", &s32ControllerID);

        memset(szTemp, 0, sizeof(szTemp));
        sprintf(szTemp, "cat %s%s/ce_count", EDAC_SYS_FILE_PATH, pDirEntry->d_name);

        if ((pFile = popen(szTemp, "r")) == NULL)
        {
            pclose(pFile);
            return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_ECC_REPORT | ERROR_OPEN_FILE),
                                  szTemp);
        }

        fgets(szTemp, sizeof(szTemp), pFile);
        pclose(pFile);

        s32ErrorCount = atoi(szTemp);
        if (s32ErrorCount == 0)
        {
            continue;
        }

        errorInformation.m_s32ControllerID = s32ControllerID;

        for (s32SlotID = 0; s32SlotID < s32Slots; s32SlotID++)
        {
            for (s32ChannelID = 0; s32ChannelID < s32Channels; s32ChannelID++)
            {
                sprintf(szTemp, "cat %s%s/csrow%d/ch%d_ce_count",
                                EDAC_SYS_FILE_PATH,
                                pDirEntry->d_name,
                                s32SlotID,
                                // For Aurora, channel2 and channel3 map to edac's channel4 and channel5
                                s32ChannelID < 2? s32ChannelID : s32ChannelID + 2);

                if ((pFile = popen(szTemp, "r")) == NULL)
                {
                    continue;
                }

                fgets(szTemp, 128, pFile);
                pclose(pFile);

                s32ErrorCount = atoi(szTemp);

                if (s32ErrorCount > 0)
                {
                    errorInformation.m_s32ChannelID = s32ChannelID;
                    errorInformation.m_s32SlotID = s32SlotID;
                    errorInformation.m_s32ErrorCount = s32ErrorCount;
                    out_pErrorList->push_back(errorInformation);
                }
            }
        }
    }

    return NO_ERROR_MSG;
}

/// @fn         MemoryDDAPI::getMemoryBuffer(UINT32             in_u32RequestedSize,
///                                          SYSTEM_ADDRESS_T*  out_pBufferAddress)
/// @brief      Get memory buffer for specified memory size
/// @param      in_u32RequestedSize is a 32bits value for requested memory size
/// @param      out_pBufferAddress is a pointer to system memory address of allocated memory buffer
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDDAPI::getMemoryBuffer(UINT32             in_u32RequestedSize,
                             SYSTEM_ADDRESS_T*  out_pBufferAddress)
{
    UINT8* u8pBufferAddress = NULL;

    if (in_u32RequestedSize == 0)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_MEMORY_BUFFER | ERROR_BAD_PARAMETER_ONE),
                              "Memory size can not be zero",
                              NULL);
    }

    if (out_pBufferAddress == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_MEMORY_BUFFER | ERROR_BAD_PARAMETER_TWO),
                              "NULL pointer",
                              NULL);
    }

    u8pBufferAddress = (UINT8 *)malloc(in_u32RequestedSize);

    if (u8pBufferAddress == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_MEMORY_BUFFER | ERROR_ALLOCATE_BUFFER_FAIL),
                              NULL,
                              NULL);
    }
    else
    {
        *out_pBufferAddress = (SYSTEM_ADDRESS_T)u8pBufferAddress;
        return NO_ERROR_MSG;
    }
}

/// @fn         MemoryDDAPI::initializeMemoryLayoutFile(void)
/// @brief      Initialize memory layout file
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDDAPI::initializeMemoryLayoutFile(void)
{
    CHAR                szTemp[256], *chpTemp;
    FILE                *pFile = NULL;
    PHYSICAL_ADDRESS_T  currentAddress = 0ULL, startAddress, endAddress;

    sprintf(szTemp, "mkdir -p %s", MEMORY_TEMP_FILE_DIR);
    system(szTemp);

    sprintf(szTemp, "cat %smc0/%s > %s", EDAC_SYS_FILE_PATH, EDAC_FILE_LAYOUT, TEMP_FILE_MEMORY_LAYOUT);
    if (system(szTemp) != 0)
    {
        sprintf(szTemp, "rm %s", TEMP_FILE_MEMORY_LAYOUT);
        system(szTemp);
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_INITIALIZE_MEMORY_LAYOUT_FILE | ERROR_OPEN_FILE),
                              TEMP_FILE_MEMORY_LAYOUT);
    }

    sprintf(szTemp, "cat %s", TEMP_FILE_MEMORY_LAYOUT);
    if ((pFile = popen(szTemp, "r")) == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_INITIALIZE_MEMORY_LAYOUT_FILE | ERROR_OPEN_FILE),
                              TEMP_FILE_MEMORY_LAYOUT);
    }

    while (fgets(szTemp, sizeof(szTemp), pFile) != NULL)
    {
        if (strstr(szTemp, "-") == NULL)
            continue;
        if ((chpTemp = strstr(szTemp, ":")) == NULL)
            continue;

        sscanf(++chpTemp, "0x%llx-0x%llx", &startAddress, &endAddress);

        if ((startAddress >= endAddress) || (currentAddress > startAddress))
        {
            pclose(pFile);
            sprintf(szTemp, "rm %s", TEMP_FILE_MEMORY_LAYOUT);
            system(szTemp);
            return returnErrorMsg(ERRORCODE(ERROR_MEMORY_DDAPI_INITIALIZE_MEMORY_LAYOUT_FILE | ERROR_MEMORY_DDAPI_INCORRECT_LAYOUT_INFO));
        }

        currentAddress = endAddress;
    }

    pclose(pFile);

    return NO_ERROR_MSG;
}

/// @fn         MemoryDDAPI::getMemoryMapSize(void)
/// @brief      Get memory map size
/// @exception  Nothing
/// @return     An INT32 value for memory map size
INT32
MemoryDDAPI::getMemoryMapSize(void)
{
    return sysconf(_SC_PAGESIZE);
}

/// @fn         MemoryDDAPI::getPhysicalAddress(PHYSICAL_ADDRESS_T   in_rankOffset,
///                                             INT32                in_s32Controller,
///                                             INT32                in_s32Channel,
///                                             INT32                in_s32Rank,
///                                             PHYSICAL_ADDRESS_T   *out_pPhysicalAddress)
/// @brief      Get the system address for the given rank and its rank offset
/// @param      in_rankOffset is a PHYSICAL_ADDRESS_T value for the rank offset
/// @param      in_s32Controller is an INT32 value for the controller index
/// @param      in_s32Channel is an INT32 value for the channel index
/// @param      in_s32Rank is an INT32 value for the rank index
/// @param      out_pPhysicalAddress is a pointer to a PHYSICAL_ADDRESS_T value for the system address
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDDAPI::getPhysicalAddress(PHYSICAL_ADDRESS_T   in_rankOffset,
                                INT32                in_s32Controller,
                                INT32                in_s32Channel,
                                INT32                in_s32Rank,
                                PHYSICAL_ADDRESS_T   *out_pPhysicalAddress)
{
    CHAR    szTemp[256];
    FILE    *pFile;
    UINT64  u64PhysicalAddress = 0ULL;

    // Do not have to check out_pPhysicalAddress because this api is only used by memory module

    m_resourceLock.lock(ERROR_MEMORY_DDAPI_GET_PHYSICAL_ADDRESS);

    sprintf(szTemp, "echo \"%s %d %d %d\" > %smc0/%s",
            UINT64ToHexStr(in_rankOffset).c_str(),
            in_s32Controller,
            in_s32Channel,
            in_s32Rank,
            EDAC_SYS_FILE_PATH,
            EDAC_FILE_RANK_TO_ADDR);
    system(szTemp);

    sprintf(szTemp, "cat %smc0/%s", EDAC_SYS_FILE_PATH, EDAC_FILE_RANK_TO_ADDR);
    if ((pFile = popen(szTemp, "r")) == NULL)
    {
        m_resourceLock.release(ERROR_MEMORY_DDAPI_GET_PHYSICAL_ADDRESS);
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_PHYSICAL_ADDRESS | ERROR_OPEN_FILE));
    }

    fgets(szTemp, sizeof(szTemp), pFile);
    sscanf(szTemp, "0x%llx", &u64PhysicalAddress);
    pclose(pFile);

    m_resourceLock.release(ERROR_MEMORY_DDAPI_GET_PHYSICAL_ADDRESS);

    if (u64PhysicalAddress == 0xFFFFFFFFFFFFFFFF)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_PHYSICAL_ADDRESS | ERROR_MEMORY_DDAPI_ADDRESS_TRANSLATION_FAIL));
    }
    *out_pPhysicalAddress = u64PhysicalAddress;

    return NO_ERROR_MSG;
}

/// @fn         MemoryDDAPI::getRankOffset(PHYSICAL_ADDRESS_T    in_physicalAddress,
///                                        PHYSICAL_ADDRESS_T    *out_pRankOffset
/// @brief      Get the rank offset for the given system address
/// @param      in_physicalAddress is a PHYSICAL_ADDRESS_T value for the system address
/// @param      out_pRankOffset is a pointer to a PHYSICAL_ADDRESS_T value for the rank offset
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDDAPI::getRankOffset(PHYSICAL_ADDRESS_T    in_physicalAddress,
                           PHYSICAL_ADDRESS_T    *out_pRankOffset)
{
    CHAR    szTemp[256];
    FILE    *pFile;
    UINT64  u64RankOffset = 0ULL;

    // Do not have to check out_pRankOffset because this api is only used by memory module

    m_resourceLock.lock(ERROR_MEMORY_DDAPI_GET_RANK_OFFSET);

    sprintf(szTemp, "echo \"%s\" > %smc0/%s", UINT64ToHexStr(in_physicalAddress).c_str(), EDAC_SYS_FILE_PATH, EDAC_FILE_ADDR_TO_RANK);
    system(szTemp);

    sprintf(szTemp, "cat %smc0/%s", EDAC_SYS_FILE_PATH, EDAC_FILE_ADDR_TO_RANK);
    if ((pFile = popen(szTemp, "r")) == NULL)
    {
        m_resourceLock.release(ERROR_MEMORY_DDAPI_GET_RANK_OFFSET);
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_RANK_OFFSET | ERROR_OPEN_FILE),
                              szTemp);
    }

    fgets(szTemp, sizeof(szTemp), pFile);
    sscanf(szTemp, "0x%llx", &u64RankOffset);
    pclose(pFile);

    m_resourceLock.release(ERROR_MEMORY_DDAPI_GET_RANK_OFFSET);

    if (u64RankOffset == 0xFFFFFFFFFFFFFFFF)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_RANK_OFFSET | ERROR_MEMORY_DDAPI_ADDRESS_TRANSLATION_FAIL));
    }
    *out_pRankOffset = u64RankOffset;


    return NO_ERROR_MSG;
}

/// @fn         MemoryDDAPI::getX86HighRamRange(MEMORY_RANGE *out_pRange)
/// @brief      Get x86 high RAM memory range
/// @param      out_pRange is a pointer to MemoryRange
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDDAPI::getX86HighRamRange(MemoryRange *out_pRange)
{
    DIR*            pDir;
    struct dirent*  pDirEntry;
    string          strFilePath;
    FILE*           pFile;
    PHYSICAL_ADDRESS_T  startAddr;

    if (out_pRange == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_HIGH_RAM_RANGE | ERROR_BAD_PARAMETER),
                              "NULL pointer",
                              NULL);
    }

    if ((pDir = opendir(E820_MEMMAP_PATH)) == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_HIGH_RAM_RANGE | ERROR_OPEN_DIRECTORY),
                              E820_MEMMAP_PATH,
                              NULL);
    }

    while ((pDirEntry = readdir(pDir)) != NULL)
    {
        if (!strcmp(".", pDirEntry->d_name))
        {
            continue;
        }

        if (!strcmp("..", pDirEntry->d_name))
        {
            continue;
        }

        if (pDirEntry->d_type & DT_DIR)
        {
            strFilePath = E820_MEMMAP_PATH;
            strFilePath += pDirEntry->d_name;
            strFilePath += "/start";

            if ((pFile = fopen(strFilePath.c_str(), "r")) == NULL)
            {
                closedir(pDir);
                return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_HIGH_RAM_RANGE | ERROR_OPEN_FILE),
                                      strFilePath.c_str(),
                                      NULL);
            }

            fscanf(pFile, "%16llx\n", &startAddr);
            fclose(pFile);

            if (startAddr == X86_HIGH_MEMORY_BASE_ADDRESS)
            {
                strFilePath = E820_MEMMAP_PATH;
                strFilePath += pDirEntry->d_name;
                strFilePath += "/end";

                if ((pFile = fopen(strFilePath.c_str(), "r")) == NULL)
                {
                    closedir(pDir);
                    return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_GET_HIGH_RAM_RANGE | ERROR_OPEN_FILE),
                                          strFilePath.c_str(),
                                          NULL);
                }

                fscanf(pFile, "%16llx\n", &out_pRange->m_endAddress);

                out_pRange->m_startAddress = X86_HIGH_MEMORY_BASE_ADDRESS;

                fclose(pFile);
                break;
            }
        }
    }

    closedir(pDir);

    return NO_ERROR_MSG;
}

/// @fn         MemoryDDAPI::injectECC(PHYSICAL_ADDRESS_T    in_physicalAddress,
///                                    UINT64                in_u64Pattern,
///                                    INT32                 in_s32ErrorBit)
/// @brief      Inject an ECC error to given physical address
/// @param      in_physicalAddress is a physical address
/// @param      in_u64Pattern is a UINT64 value for the pattern
/// @param      in_s32ErrorBit is a INT32 value to indicate the error bit
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDDAPI::injectECC(PHYSICAL_ADDRESS_T    in_physicalAddress,
                       UINT64                in_u64Pattern,
                       INT32                 in_s32ErrorBit)
{
    DIR*            pDir;
    struct dirent*  pDirEntry;
    string          strPath;
    CHAR            szTemp[128];
    FILE*           pFile;
    INT32           s32ErrorCode;

    if (in_s32ErrorBit > 63)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_INJECT_ECC | ERROR_BAD_PARAMETER_THREE),
                              "error bit should be 0 ~ 63");
    }

    if ((pDir = opendir(EDAC_SYS_FILE_PATH)) == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_INJECT_ECC | ERROR_OPEN_DIRECTORY),
                              EDAC_SYS_FILE_PATH,
                              NULL);
    }

    while ((pDirEntry = readdir(pDir)) != NULL)
    {
        if (!(pDirEntry->d_type & DT_DIR))
        {
            continue;
        }

        if (strstr(pDirEntry->d_name, "mc0") != NULL)
        {
            strPath = EDAC_SYS_FILE_PATH;
            strPath += pDirEntry->d_name;
            strPath += "/";
            break;
        }
    }

    closedir(pDir);

    if (strPath.empty())
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_INJECT_ECC | ERROR_MEMORY_DDAPI_EDAC_DEVICE_NOT_FOUND));
    }

    sprintf(szTemp, "%s%s", strPath.c_str(), EDAC_FILE_INJECT_STATUS);
    if ((pFile = fopen(szTemp, "r")) == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_INJECT_ECC | ERROR_FUNCTION_NOT_SUPPORTED));
    }
    fclose(pFile);

    // Inject error (format: addr pattern error_bit)
    sprintf(szTemp, "echo \"0x%llx 0x%llx %d\" > %s%s",
                    in_physicalAddress,
                    in_u64Pattern,
                    in_s32ErrorBit,
                    strPath.c_str(),
                    EDAC_FILE_INJECT_ERROR);
    system(szTemp);

    // Check status
    sprintf(szTemp , "cat %s%s", strPath.c_str(), EDAC_FILE_INJECT_STATUS);
    if ((pFile = popen(szTemp, "r")) == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_INJECT_ECC | ERROR_OPEN_FILE),
                              szTemp);
    }

    if (fgets(szTemp, sizeof(szTemp), pFile) == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_INJECT_ECC | ERROR_UNEXPECTED_VALUE));
    }

    s32ErrorCode = atoi(szTemp);

    pclose(pFile);

    if (s32ErrorCode != 0)
    {
        switch (s32ErrorCode)
        {
            case ECC_INJECT_LOCK:
                addErrorMsg(ERROR_MEMORY_DRIVER_ECC_INJECT_LOCK, NULL);
                break;
            case ADDR_TRANSLATION_FAIL:
                addErrorMsg(ERROR_MEMORY_DRIVER_ADDR_TRANSLATION_FAIL, NULL);
                break;
            case IOREMAP_FAIL:
                addErrorMsg(ERROR_MEMORY_DRIVER_IOREMAP_FAIL, NULL);
                break;
            case ECC_INJECT_FAIL:
                addErrorMsg(ERROR_MEMORY_DRIVER_ECC_INJECT_FAIL, NULL);
                break;
            case COMPARE_DATA_FAIL:
                addErrorMsg(ERROR_MEMORY_DRIVER_COMPARE_DATA_FAIL, NULL);
                break;
            case ENABLE_ECC_INJECT_FAIL:
                addErrorMsg(ERROR_MEMORY_DRIVER_ENABLE_ECC_INJECT_FAIL, NULL);
                break;
            case INCORRECT_ECC_COUNT:
                addErrorMsg(ERROR_MEMORY_DRIVER_INCORRECT_ECC_COUNT, NULL);
                break;
            case NO_ECC_COUNT:
                addErrorMsg(ERROR_MEMORY_DRIVER_NO_ECC_COUNT, NULL);
                break;
            default:
                return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_INJECT_ECC | ERROR_UNEXPECTED_VALUE));
                break;
        }

        return returnErrorMsg(ERROR_MEMORY_DDAPI_INJECT_ECC, NULL, getErrorMsg());
    }
    return NO_ERROR_MSG;
}

/// @fn         MemoryDDAPI::read8(PHYSICAL_ADDRESS_T  in_physicalAddress)
/// @brief      Read 8bits data from specified physical address 
/// @param      in_physicalAddress is a physical address
/// @exception  Nothing
/// @return     A 8bits data
UINT8
MemoryDDAPI::read8(PHYSICAL_ADDRESS_T  in_physicalAddress)
{
	ERRORCODE			eErrorCode = NO_ERROR;
    SYSTEM_ADDRESS_T 	mapAddress = (SYSTEM_ADDRESS_T) NULL;
    string              strErrorMsg;

	// Remap physical memory address to system address
    eErrorCode = remapIOAddress(in_physicalAddress, 1, &mapAddress);
    if (eErrorCode != NO_ERROR)
    {
        strErrorMsg = "failed address: " + UINT64ToHexStr(in_physicalAddress);
        addErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_READ_MEMORY | eErrorCode), strErrorMsg.c_str());
        return 0xde;
    }

    return *((UINT8 *)mapAddress);
}

/// @fn         MemoryDDAPI::read16(PHYSICAL_ADDRESS_T in_physicalAddress)
/// @brief      Read 16bits data from specified physical address 
/// @param      in_physicalAddress is a physical address
/// @exception  Nothing
/// @return     A 16bits data
UINT16
MemoryDDAPI::read16(PHYSICAL_ADDRESS_T in_physicalAddress)
{
	ERRORCODE			eErrorCode = NO_ERROR;
    SYSTEM_ADDRESS_T 	mapAddress = (SYSTEM_ADDRESS_T) NULL;
    string              strErrorMsg;

	// Remap physcial memory address to system address
    eErrorCode = remapIOAddress(in_physicalAddress, 2, &mapAddress);
    if (eErrorCode != NO_ERROR)
    {
        strErrorMsg = "failed address: " + UINT64ToHexStr(in_physicalAddress);
        addErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_READ_MEMORY | eErrorCode), strErrorMsg.c_str());
        return 0xde;
    }

    return *((UINT16 *)mapAddress);
}

/// @fn         MemoryDDAPI::read32(PHYSICAL_ADDRESS_T in_physicalAddress)
/// @brief      Read 32bits data from specified physical address 
/// @param      in_physicalAddress is a physical address
/// @exception  Nothing
/// @return     A 32bits data
UINT32
MemoryDDAPI::read32(PHYSICAL_ADDRESS_T in_physicalAddress)
{
	ERRORCODE			eErrorCode = NO_ERROR;
    SYSTEM_ADDRESS_T 	mapAddress = (SYSTEM_ADDRESS_T) NULL;
    string              strErrorMsg;
    
	// Remap physcial memory address to system address
    eErrorCode = remapIOAddress(in_physicalAddress, 4, &mapAddress);
    if (eErrorCode != NO_ERROR)
    {
        strErrorMsg = "failed address: " + UINT64ToHexStr(in_physicalAddress);
        addErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_READ_MEMORY | eErrorCode), strErrorMsg.c_str());
        return 0xde;
    }

    return *((UINT32 *)mapAddress);
}

/// @fn         MemoryDDAPI::read64(PHYSICAL_ADDRESS_T in_physicalAddress)
/// @brief      Read 64bits data from specified physical address 
/// @param      in_physicalAddress is a physical address
/// @exception  Nothing
/// @return     A 64bits data
UINT64
MemoryDDAPI::read64(PHYSICAL_ADDRESS_T in_physicalAddress)
{
	ERRORCODE			eErrorCode = NO_ERROR;
    SYSTEM_ADDRESS_T 	mapAddress = (SYSTEM_ADDRESS_T) NULL;
    string              strErrorMsg;
    
	// Remap physcial memory address to system address
    eErrorCode = remapIOAddress(in_physicalAddress, 8, &mapAddress);
    if (eErrorCode != NO_ERROR)
    {
        strErrorMsg = "failed address: " + UINT64ToHexStr(in_physicalAddress);
        addErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_READ_MEMORY | eErrorCode), strErrorMsg.c_str());
        return 0xde;
    }

    return *((UINT64 *)mapAddress);
}

/// @fn         MemoryDDAPI::remapIOAddress(PHYSICAL_ADDRESS_T  in_physicalAddress,
///                                         PHYSICAL_ADDRESS_T  in_minMapSize,
///                                         SYSTEM_ADDRESS_T*   out_pMapAddress)
/// @brief      Map physical memory address to system memory address
/// @param      in_physicalAddress is a physical address
/// @param      in_minMapSize is a value of physical address for minimum requested map size in bytes
/// @param      out_pMapAddress is a pointer to system memory address
/// @exception  Nothing
/// @return     A 32bits error code
ERRORCODE
MemoryDDAPI::remapIOAddress(PHYSICAL_ADDRESS_T  in_physicalAddress,
                            PHYSICAL_ADDRESS_T  in_minMapSize,
                            SYSTEM_ADDRESS_T*   out_pMapAddress)
{
    PHYSICAL_ADDRESS_T  phyStartAddress = 0;
    PHYSICAL_ADDRESS_T  mapSize = 0;
    PHYSICAL_ADDRESS_T  mapMask = 0;
    void*               pTempMapAddress = NULL;


    *out_pMapAddress = (SYSTEM_ADDRESS_T)NULL;

    if (m_memoryMapInformation.mapSize != 0)
    {
        mapSize = m_memoryMapInformation.mapSize;
        mapMask = mapSize - 1;

        if (m_memoryMapInformation.phyStartAddress == ((in_physicalAddress + in_minMapSize - 1) & ~mapMask))
        {
            // Current physical address is in the mapped block
            *out_pMapAddress = (m_memoryMapInformation.mapAddress + ((SYSTEM_ADDRESS_T) (in_physicalAddress & mapMask)));
            return NO_ERROR;
        }
        else
        {
            // Unmap previous mapped block
            munmap((void*)m_memoryMapInformation.mapAddress, mapSize);
            m_memoryMapInformation.mapAddress = (SYSTEM_ADDRESS_T) NULL;
            m_memoryMapInformation.mapSize = 0;

            mapSize = getMemoryMapSize();
            mapMask = mapSize - 1;
        }
    }
    else
    {
        m_memoryMapInformation.s32FileDescriptor = open("/dev/mem", O_RDWR | O_SYNC);
        if (m_memoryMapInformation.s32FileDescriptor < 0)
        {
            return ERROR_MEMORY_DDAPI_OPEN_MEM_DEVICE_FAIL;
        }

        mapSize = getMemoryMapSize();
        mapMask = mapSize - 1;
    }

    phyStartAddress = in_physicalAddress & ~mapMask;

    // If current memory operation will access memory address which exceeds
    // the bundary of the mapped block, update memory map parameters
    if (phyStartAddress != ((in_physicalAddress + in_minMapSize - 1) & ~mapMask))
    {
        phyStartAddress = in_physicalAddress;

        mapSize = in_minMapSize;
        mapMask = mapSize - 1;
    }

    // Map physcial address to system address
    pTempMapAddress = mmap64(0,
                             mapSize,
                             (PROT_READ | PROT_WRITE),
                             MAP_SHARED,
                             m_memoryMapInformation.s32FileDescriptor,
                             phyStartAddress);

    if (pTempMapAddress == MAP_FAILED)
    {
        if (mapSize != in_minMapSize)
        {
            // Map for request minimum map size
            phyStartAddress = in_physicalAddress;
            mapSize = in_minMapSize;
            mapMask = mapSize - 1;

            pTempMapAddress = mmap64(0,
                                     mapSize,
                                     (PROT_READ | PROT_WRITE),
                                     MAP_SHARED,
                                     m_memoryMapInformation.s32FileDescriptor,
                                     phyStartAddress);
        }

        if (pTempMapAddress == MAP_FAILED)
        {
            close(m_memoryMapInformation.s32FileDescriptor);
            m_memoryMapInformation.s32FileDescriptor = -1;
            return ERROR_MEMORY_DDAPI_MMAP_FAIL;
        }
    }

    m_memoryMapInformation.mapAddress = (SYSTEM_ADDRESS_T) pTempMapAddress;
    m_memoryMapInformation.phyStartAddress = phyStartAddress;
    m_memoryMapInformation.mapSize = mapSize;

    *out_pMapAddress = (m_memoryMapInformation.mapAddress + ((SYSTEM_ADDRESS_T) (in_physicalAddress & mapMask)));

    return NO_ERROR;
}

/// @fn         MemoryDDAPI::write(PHYSICAL_ADDRESS_T  in_physicalAddress,
///                                UINT8			   in_u8Value)
/// @brief      Write 8bits data to specified physical address 
/// @param      in_physicalAddress is a physical address
/// @param      in_u8Value is a 8bits data to be written to memory
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDDAPI::write(PHYSICAL_ADDRESS_T  in_physicalAddress,
                   UINT8			   in_u8Value)
{
    ERRORCODE			eErrorCode = NO_ERROR;
    SYSTEM_ADDRESS_T 	mapAddress = (SYSTEM_ADDRESS_T) NULL;
    string              strErrorMsg;
    
	// Remap physcial memory address to system address
    eErrorCode = remapIOAddress(in_physicalAddress, 1, &mapAddress);
    if (eErrorCode != NO_ERROR)
    {
        strErrorMsg = "failed address: " + UINT64ToHexStr(in_physicalAddress);
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_WRITE_MEMORY | eErrorCode), strErrorMsg.c_str(), NULL);
    }

    *((UINT8*)mapAddress) = in_u8Value;

    return returnErrorMsg(NO_ERROR, NULL, NULL);
}

/// @overload   MemoryDDAPI::write(PHYSICAL_ADDRESS_T  in_physicalAddress,
///                                UINT16			   in_u16Value)
/// @brief      Write 16bits data to specified physical address 
/// @param      in_physicalAddress is a physical address
/// @param      in_u16Value is a 16bits data to be written to memory
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDDAPI::write(PHYSICAL_ADDRESS_T  in_physicalAddress,
                   UINT16			   in_u16Value)
{
    ERRORCODE			eErrorCode = NO_ERROR;
    SYSTEM_ADDRESS_T 	mapAddress = (SYSTEM_ADDRESS_T) NULL;
    string              strErrorMsg;
    
	// Remap physcial memory address to system address
    eErrorCode = remapIOAddress(in_physicalAddress, 2, &mapAddress);
    if (eErrorCode != NO_ERROR)
    {
        strErrorMsg = "failed address: " + UINT64ToHexStr(in_physicalAddress);
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_WRITE_MEMORY | eErrorCode), strErrorMsg.c_str(), NULL);
    }

    *((UINT16*)mapAddress) = in_u16Value;

    return NO_ERROR_MSG;
}

/// @overload   MemoryDDAPI::write(PHYSICAL_ADDRESS_T  in_physicalAddress,
///                                UINT32			   in_u32Value)
/// @brief      Write 32bits data to specified physical address 
/// @param      in_physicalAddress is a physical address
/// @param      in_u32Value is a 32bits data to be written to memory
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDDAPI::write(PHYSICAL_ADDRESS_T  in_physicalAddress,
                   UINT32			   in_u32Value)
{
    ERRORCODE			eErrorCode = NO_ERROR;
    SYSTEM_ADDRESS_T 	mapAddress = (SYSTEM_ADDRESS_T) NULL;
    string              strErrorMsg;
    
	// Remap physcial memory address to system address
    eErrorCode = remapIOAddress(in_physicalAddress, 4, &mapAddress);
    if (eErrorCode != NO_ERROR)
    {
        strErrorMsg = "failed address: " + UINT64ToHexStr(in_physicalAddress);
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_WRITE_MEMORY | eErrorCode), strErrorMsg.c_str(), NULL);
    }

    *((UINT32*)mapAddress) = in_u32Value;

    return NO_ERROR_MSG;
}

/// @overload   MemoryDDAPI::write(PHYSICAL_ADDRESS_T  in_physicalAddress,
///                                UINT64			   in_u64Value)
/// @brief      Write 64bits data to specified physical address 
/// @param      in_physicalAddress is a physical address
/// @param      in_u64Value is a 64bits data to be written to memory
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
MemoryDDAPI::write(PHYSICAL_ADDRESS_T  in_physicalAddress,
                   UINT64			   in_u64Value)
{
    ERRORCODE			eErrorCode = NO_ERROR;
    SYSTEM_ADDRESS_T 	mapAddress = (SYSTEM_ADDRESS_T) NULL;
    string              strErrorMsg;
    
	// Remap physcial memory address to system address
    eErrorCode = remapIOAddress(in_physicalAddress, 8, &mapAddress);
    if (eErrorCode != NO_ERROR)
    {
        strErrorMsg = "failed address: " + UINT64ToHexStr(in_physicalAddress);
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_WRITE_MEMORY | eErrorCode), strErrorMsg.c_str(), NULL);
    }

    *((UINT64*)mapAddress) = in_u64Value;

    return NO_ERROR_MSG;
}

/// @fn         MemoryDDAPI::openMemoryControllerDriver(void)
/// @brief      Open the device node of the memory controller to use IOCTL commands
/// @exception  Nothing
/// @return     A boolean value to indicate whether the device file is opened successfully or not
BOOL
MemoryDDAPI::openMemoryControllerDriver(void)
{
    string strMCDrvPath;

    if (m_s32FXNMCDrvFD <= 0)
    {
        strMCDrvPath = "/dev/";
        strMCDrvPath.append(FXN_MC_DRV_NAME);

        m_s32FXNMCDrvFD = open(strMCDrvPath.c_str(), O_RDWR);
    }

    return (0 < m_s32FXNMCDrvFD)? TRUE : FALSE;
}

/// @fn         MemoryDDAPI::releaseMemoryBuffer(SYSTEM_ADDRESS_T in_bufferAddress)
/// @brief      Release memory buffer for specified memory buffer address 
/// @param      in_bufferAddress is a memory address for specified memory buffer
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg* 
MemoryDDAPI::releaseMemoryBuffer(SYSTEM_ADDRESS_T in_bufferAddress)
{
    if ((SYSTEM_ADDRESS_T *)in_bufferAddress == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_MEMORY_DDAPI_RELEASE_MEMORY_BUFFER | ERROR_BAD_PARAMETER),
                              "Invalid buffer address",
                              NULL);
    }

    free((void *)in_bufferAddress);

    return NO_ERROR_MSG;
}
