/// @file       MemoryDDAPI.hpp
/// @brief      MemoryDDAPI class
/// @details    Provide low layer diagnostic and utility APIs for memory module
/// @author     KenHu
/// @copyright  Copyright 2014 Hon Hai Technology Group. All Rights Reserved.

#ifndef MEMORY_DDAPI_HPP
#define MEMORY_DDAPI_HPP

#include <list>

#include "BaseClassDDAPI.hpp"
#include "DIMMInformation.hpp"
#include "ECCInformation.hpp"
#include "MemoryRange.hpp"
#include "MemoryTempFile.hpp"
#include "FXNMCDrvIOCTL.h"

#define MEMORY_CONF_FILE                "memory.conf"

#define X86_HIGH_MEMORY_BASE_ADDRESS    0x100000000ULL
#define E820_MEMMAP_PATH                "/sys/firmware/memmap/"

#define EDAC_SYS_FILE_PATH              "/sys/devices/system/edac/mc/"
#define EDAC_FILE_LAYOUT                "layout"
#define EDAC_FILE_INJECT_ERROR          "inject_error"
#define EDAC_FILE_INJECT_STATUS         "inject_status"
#define EDAC_FILE_ADDR_TO_RANK          "addr_to_rank"
#define EDAC_FILE_RANK_TO_ADDR          "rank_to_addr"

using::std::list;

enum InjectFailType {
    ECC_INJECT_LOCK = 1,
    ADDR_TRANSLATION_FAIL,
    IOREMAP_FAIL,
    ECC_INJECT_FAIL,
    COMPARE_DATA_FAIL,
    ENABLE_ECC_INJECT_FAIL,
    INCORRECT_ECC_COUNT,
    NO_ECC_COUNT,
};

/// @class      MemoryDDAPI MemoryDDAPI.hpp "MemoryDDAPI.hpp"
/// @brief      Provide low layer diagnostic and utility APIs for memory module
/// @param      m_memoryMapInformation is a MemoryMapInformation structure for memory mapping
/// @param      m_s32FXNMCDrvFD is a file descriptor for memory controller device node
/// @param      m_pI2CDAPI is a pointer to I2C DAPI
class MemoryDDAPI : public BaseClassDDAPI
{
    public:
                                MemoryDDAPI(STRING in_szClassName);
        virtual                 ~MemoryDDAPI(void);

        ErrorMsg*               clearECCReport(void);

        ErrorMsg*               getControllerChannelCount(INT32* out_ps32Count);

        ErrorMsg*               getControllerSlotCount(INT32* out_ps32Count);

        ErrorMsg*               getECCReport(list<ECCInformation>* out_pErrorList);

        ErrorMsg*               getX86HighRamRange(MemoryRange *out_pRange);

        ErrorMsg*               getMemoryBuffer(UINT32              in_u32RequestedSize,
                                                SYSTEM_ADDRESS_T*   out_pBufferAddress);

        INT32                   getMemoryMapSize(void);

        ErrorMsg*               getPhysicalAddress(PHYSICAL_ADDRESS_T   in_rankOffset,
                                                   INT32                in_s32Controller,
                                                   INT32                in_s32Channel,
                                                   INT32                in_s32Rank,
                                                   PHYSICAL_ADDRESS_T   *out_pPhysicalAddress);

        ErrorMsg*               getRankOffset(PHYSICAL_ADDRESS_T    in_physicalAddress,
                                              PHYSICAL_ADDRESS_T    *out_pRankOffset);

        ErrorMsg*               initializeMemoryLayoutFile(void);

        ErrorMsg*               injectECC(PHYSICAL_ADDRESS_T    in_physicalAddress,
                                          UINT64                in_u64Pattern,
                                          INT32                 in_s32ErrorBit);

        ErrorMsg*               write(PHYSICAL_ADDRESS_T    in_physicalAddress,
                                      UINT8                 in_u8Value);

        ErrorMsg*               write(PHYSICAL_ADDRESS_T    in_physicalAddress,
                                      UINT16                in_u16Value);

        ErrorMsg*               write(PHYSICAL_ADDRESS_T    in_physicalAddress,
                                      UINT32                in_32uValue);

        ErrorMsg*               write(PHYSICAL_ADDRESS_T    in_physicalAddress,
                                      UINT64                in_u64Value);

        UINT8                   read8(PHYSICAL_ADDRESS_T    in_physicalAddress);
            
        UINT16                  read16(PHYSICAL_ADDRESS_T   in_physicalAddress);
            
        UINT32                  read32(PHYSICAL_ADDRESS_T   in_physicalAddress);
            
        UINT64                  read64(PHYSICAL_ADDRESS_T   in_physicalAddress);

        ErrorMsg*               releaseMemoryBuffer(SYSTEM_ADDRESS_T in_bufferAddress);

    protected:

    private:
        typedef struct {
            INT32                   s32FileDescriptor;
            PHYSICAL_ADDRESS_T      phyStartAddress;
            PHYSICAL_ADDRESS_T      mapSize;
            SYSTEM_ADDRESS_T        mapAddress;
        } MemoryMapInformation;

        BOOL                    openMemoryControllerDriver(void);

        ERRORCODE               remapIOAddress(PHYSICAL_ADDRESS_T   in_physicalAddress,
                                               PHYSICAL_ADDRESS_T   in_minMapSize,
                                               SYSTEM_ADDRESS_T*    out_pMapAddress);

        MemoryMapInformation    m_memoryMapInformation; 

        INT32                   m_s32FXNMCDrvFD;   
};
#endif

