/// @file       SPDDDAPI.cpp
/// @brief      SPDDDAPI class
/// @details    Provide low layer utility APIs for DIMM's SPD
/// @author     KenHu
/// @copyright  Copyright 2015-2016 Hon Hai Technology Group. All Rights Reserved.

#include "SPDDDAPI.hpp"

/// @fn         SPDDDAPI::SPDDDAPI(void) : BaseClassDDAPI()
/// @brief      SPDDDAPI constructor
/// @exception  Nothing
/// @return     Nothing
SPDDDAPI::SPDDDAPI(void) : BaseClassDDAPI("SPDDDAPI")
{
    m_pI2CDAPI = NULL;

    if ((m_pI2CDAPI = new I2CDAPI("I2CDAPI")) == NULL)
    {
        addErrorMsg((ERRORCODE)(ERROR_SPD_DDAPI_CONSTRUCTOR | ERROR_DIAGNOSTIC_DRIVER_API_CONSTRUCTOR1_ERROR),
                    "create I2CDAPI fail",
                    NULL);
    }
    else if (m_pI2CDAPI->getErrorMsg() != NO_ERROR_MSG)
    {
        addErrorMsg(ERROR_SPD_DDAPI_CONSTRUCTOR, NULL, m_pI2CDAPI->getErrorMsg());
    }
}


/// @fn         SPDDDAPI::~SPDDDAPI(void)
/// @brief      SPDDDAPI destructor
/// @exception  Nothing
/// @return     Nothing
SPDDDAPI::~SPDDDAPI(void)
{
    if (m_pI2CDAPI)
    {
        delete m_pI2CDAPI;
        m_pI2CDAPI = NULL;
    }
}


/// @fn         SPDDDAPI::getDeviceList(list<string>    *out_pDeviceList,
///                                     BOOL            in_bDetectedOnly)
///
/// @brief      Get SPD device list
/// @param      out_pDeviceList is a pointer to a list of strings
/// @param      in_bDetectedOnly is a bool value: TRUE - list detected devices FALSE - list supported devices
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SPDDDAPI::getDeviceList(list<string>    *out_pDeviceList,
                        BOOL            in_bDetectedOnly)
{
    ErrorMsg*               pErrorMsg;
    list<string>            i2cDevList;
    list<string>::iterator  itDev;
    DDRType                 eDDRType;
    std::size_t             position;

    if (out_pDeviceList == NULL)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SPD_DDAPI_GET_DEVICE_LIST | ERROR_BAD_PARAMETER),
                              "NULL pointer",
                              NULL);
    }

    out_pDeviceList->clear();

    if ((pErrorMsg = m_pI2CDAPI->DisplayI2cDevicesList(&i2cDevList)))
    {
        return returnErrorMsg(ERROR_SPD_DDAPI_GET_DEVICE_LIST, NULL, pErrorMsg);
    }

    itDev = i2cDevList.begin();
    while (itDev != i2cDevList.end())
    {
        if (itDev->find("spd") != std::string::npos)
        {
            // Remove un-useful characters from i2c device name
            position = itDev->find(" ");
            if (position != std::string::npos)
                itDev->erase(position);

            if (in_bDetectedOnly && (pErrorMsg = getDeviceType(*itDev, &eDDRType)))
            {
                deleteErrorMsg(pErrorMsg);
                itDev = i2cDevList.erase(itDev);
            }
            else
            {
                itDev++;
            }
        }
        else
        {
            itDev = i2cDevList.erase(itDev);
        }
    }

    if (i2cDevList.empty())
    {
        return returnErrorMsg(ERRORCODE(ERROR_SPD_DDAPI_GET_DEVICE_LIST | ERROR_SPD_DDAPI_NO_SPD_DEVICE));
    }

    *out_pDeviceList = i2cDevList;

    return NO_ERROR_MSG;
}

/// @fn         SPDDDAPI::getDeviceType(string  in_strDIMM,
///                                     DDRType *out_epType)
///
/// @brief      Get device type based on DIMM's SPD
/// @param      in_strDIMM is a string for SPD device name
/// @param      out_epType is a pointer to DDRType
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SPDDDAPI::getDeviceType(string  in_strDIMM,
                        DDRType *out_epType)
{
    ErrorMsg*   pErrorMsg;
    UINT8       u8Data = 0x0;
    INT32       s32Count = 0;

    if (out_epType == NULL)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SPD_DDAPI_GET_DEVICE_TYPE | ERROR_BAD_PARAMETER_TWO),
                              "NULL pointer",
                              NULL);
    }

    *out_epType = DDR_TYPE_UNKNOWN;

    pErrorMsg = m_pI2CDAPI->Read(in_strDIMM, SPD_OFFSET_DEVICE_TYPE, SPD_OFFSET_DEVICE_TYPE, &u8Data, &s32Count);
    if (pErrorMsg)
    {
        return returnErrorMsg(ERROR_SPD_DDAPI_GET_DEVICE_TYPE, NULL, pErrorMsg);
    }

    switch (u8Data)
    {
        case DDR3:
        case DDR4:
            *out_epType = static_cast<DDRType>(u8Data);
            break;
        default:
            return returnErrorMsg((ERRORCODE)(ERROR_SPD_DDAPI_GET_DEVICE_TYPE | ERROR_UNEXPECTED_VALUE),
                                  "unknown device");
    }

    return NO_ERROR_MSG;
}

/// @fn         SPDDDAPI::getDRAMVendor(string in_strDIMM,
///                                     string *out_strpName)
///
/// @brief      Get DRAM vendor name based on DIMM's SPD
/// @param      in_strDIMM is a string for SPD device name
/// @param      out_strpName is a pointer to a string for the vendor name
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SPDDDAPI::getDRAMVendor(string in_strDIMM,
                        string *out_strpName)
{
#include "JEDEC_ManufactureID.hpp"

    ErrorMsg*   pErrorMsg = NO_ERROR_MSG;
    DDRType     eDDRType = DDR_TYPE_UNKNOWN;
    INT32       s32StartAddr = 0;
    INT32       s32EndAddr = 0;
    UINT8       u8Data[SPD_LENGTH_DRAM_MANUFACTURE_ID];
    INT32       s32Count = 0;
    string      strError;

    if (out_strpName == NULL)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SPD_DDAPI_GET_MODULE_VENDOR | ERROR_BAD_PARAMETER_TWO),
                              "NULL pointer",
                              NULL);
    }

    out_strpName->clear();

    if ((pErrorMsg = getDeviceType(in_strDIMM, &eDDRType)))
    {
        return returnErrorMsg(ERROR_SPD_DDAPI_GET_DRAM_VENDOR, NULL, pErrorMsg);
    }

    switch (eDDRType)
    {
    case DDR3:
        s32StartAddr = SPD_DDR3_OFFSET_DRAM_MANUFACTURE_ID;
        s32EndAddr = s32StartAddr + SPD_LENGTH_DRAM_MANUFACTURE_ID - 1;
        break;
    case DDR4:
        s32StartAddr = SPD_DDR4_OFFSET_DRAM_MANUFACTURE_ID;
        s32EndAddr = s32StartAddr + SPD_LENGTH_DRAM_MANUFACTURE_ID - 1;
        break;
    default:
        // Do not have to consider this error because getDeviceType() will return error when the device type is unknown
        break;
    }

    memset(u8Data, 0, sizeof(u8Data));

    if ((pErrorMsg = m_pI2CDAPI->Read(in_strDIMM, s32StartAddr, s32EndAddr, u8Data, &s32Count)))
    {
        return returnErrorMsg(ERROR_SPD_DDAPI_GET_DRAM_VENDOR, NULL, pErrorMsg);
    }

    // Don't care parity bit
    // Byte one: number of continuation codes, Byte two: manufacturer’s ID code
    u8Data[0] &= 0x7F;
    u8Data[1] &= 0x7F;

    if (MAX_NUMBER_OF_MANUFACTURE_ID_BANK <= u8Data[0])
    {
        strError = "unsupported bank index(0x" + Int2HexStr(u8Data[0]) + ")";
        return returnErrorMsg((ERRORCODE)(ERROR_SPD_DDAPI_GET_DRAM_VENDOR | ERROR_UNEXPECTED_VALUE),
                              strError.c_str());
    }

    if (u8Data[1] == 0x0)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_SPD_DDAPI_GET_DRAM_VENDOR | ERROR_UNEXPECTED_VALUE),
                              "manufacture id should not be 0");
    }

    if (manufactureIDTable[u8Data[0]][(u8Data[1] - 1)] == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_SPD_DDAPI_GET_DRAM_VENDOR | ERROR_SPD_DDAPI_UNKNOWN_VENDOR));
    }

    *out_strpName = manufactureIDTable[u8Data[0]][(u8Data[1] - 1)];

    return NO_ERROR_MSG;
}

/// @fn         SPDDDAPI::getModuleType(string in_strDIMM,
///                                     string *out_strpType)
///
/// @brief      Get module type based on DIMM's SPD
/// @param      in_strDIMM is a string for SPD device name
/// @param      out_strpType is a pointer to a string for module type
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SPDDDAPI::getModuleType(string in_strDIMM,
                        string *out_strpType)
{
    ErrorMsg*   pErrorMsg;
    UINT8       u8Data = 0x0;
    INT32       s32Count = 0;

    if (out_strpType == NULL)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SPD_DDAPI_GET_MODULE_TYPE | ERROR_BAD_PARAMETER_TWO),
                              "NULL pointer",
                              NULL);
    }

    out_strpType->clear();

    pErrorMsg = m_pI2CDAPI->Read(in_strDIMM, SPD_OFFSET_MODULE_TYPE, SPD_OFFSET_MODULE_TYPE, &u8Data, &s32Count);
    if (pErrorMsg)
    {
        return returnErrorMsg(ERROR_SPD_DDAPI_GET_MODULE_TYPE, NULL, pErrorMsg);
    }

    out_strpType->clear();

    switch (u8Data)
    {
        case RDIMM:
            *out_strpType = "Registered DIMM";
            break;
        case UDIMM:
            *out_strpType = "Unbuffered DIMM";
            break;
        case SO_DIMM:
            *out_strpType = "Unbuffered SO-DIMM (64b)";
            break;
        case LRDIMM:
            *out_strpType = "Load Reduced DIMM";
            break;
        case MINI_RDIMM:
            *out_strpType = "Mini Registered DIMM";
            break;
        case MINI_UDIMM:
            *out_strpType = "Mini Unbuffered DIMM";
            break;
        case SO_RDIMM_72B:
            *out_strpType = "Registered SO-DIMM";
            break;
        case SO_UDIMM_72B:
            *out_strpType = "Unbuffered SO-DIMM (72b)";
            break;
        case SO_DIMM_16B:
            *out_strpType = "Unbuffered SO-DIMM (16b)";
            break;
        case SO_DIMM_32B:
            *out_strpType = "Unbuffered SO-DIMM (32b)";
            break;
        default:
            return returnErrorMsg((ERRORCODE)(ERROR_SPD_DDAPI_GET_MODULE_TYPE | ERROR_UNEXPECTED_VALUE),
                                  "unknown module");
    }

    return NO_ERROR_MSG;
}

/// @fn         SPDDDAPI::getModuleVendor(string in_strDIMM,
///                                       string *out_strpName)
///
/// @brief      Get module vendor name based on DIMM's SPD
/// @param      in_strDIMM is a string for SPD device name
/// @param      out_strpName is a pointer to a string for the vendor name
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SPDDDAPI::getModuleVendor(string in_strDIMM,
                          string *out_strpName)
{
#include "JEDEC_ManufactureID.hpp"

    ErrorMsg*   pErrorMsg = NO_ERROR_MSG;
    DDRType     eDDRType = DDR_TYPE_UNKNOWN;
    INT32       s32StartAddr = 0;
    INT32       s32EndAddr = 0;
    UINT8       u8Data[SPD_LENGTH_MODULE_MANUFACTURE_ID];
    INT32       s32Count = 0;
    string      strError;

    if (out_strpName == NULL)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SPD_DDAPI_GET_MODULE_VENDOR | ERROR_BAD_PARAMETER_TWO),
                              "NULL pointer",
                              NULL);
    }

    out_strpName->clear();

    if ((pErrorMsg = getDeviceType(in_strDIMM, &eDDRType)))
    {
        return returnErrorMsg(ERROR_SPD_DDAPI_GET_MODULE_VENDOR, NULL, pErrorMsg);
    }

    switch (eDDRType)
    {
    case DDR3:
        s32StartAddr = SPD_DDR3_OFFSET_MODULE_MANUFACTURE_ID;
        s32EndAddr = s32StartAddr + SPD_LENGTH_MODULE_MANUFACTURE_ID - 1;
        break;
    case DDR4:
        s32StartAddr = SPD_DDR4_OFFSET_MODULE_MANUFACTURE_ID;
        s32EndAddr = s32StartAddr + SPD_LENGTH_MODULE_MANUFACTURE_ID - 1;
        break;
    default:
        // Do not have to consider this error because getDeviceType() will return error when the device type is unknown
        break;
    }

    memset(u8Data, 0, sizeof(u8Data));

    if ((pErrorMsg = m_pI2CDAPI->Read(in_strDIMM, s32StartAddr, s32EndAddr, u8Data, &s32Count)))
    {
        return returnErrorMsg(ERROR_SPD_DDAPI_GET_MODULE_VENDOR, NULL, pErrorMsg);
    }

    // Don't care parity bit
    // Byte one: number of continuation codes, Byte two: manufacturer’s ID code
    u8Data[0] &= 0x7F;
    u8Data[1] &= 0x7F;

    if (MAX_NUMBER_OF_MANUFACTURE_ID_BANK <= u8Data[0])
    {
        strError = "unsupported bank index(0x" + Int2HexStr(u8Data[0]) + ")";
        return returnErrorMsg((ERRORCODE)(ERROR_SPD_DDAPI_GET_MODULE_VENDOR | ERROR_UNEXPECTED_VALUE),
                              strError.c_str());
    }

    if (u8Data[1] == 0x0)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_SPD_DDAPI_GET_MODULE_VENDOR | ERROR_UNEXPECTED_VALUE),
                              "manufacture id should not be 0");
    }

    if (manufactureIDTable[u8Data[0]][(u8Data[1] - 1)] == NULL)
    {
        return returnErrorMsg((ERRORCODE)(ERROR_SPD_DDAPI_GET_MODULE_VENDOR | ERROR_SPD_DDAPI_UNKNOWN_VENDOR));
    }

    *out_strpName = manufactureIDTable[u8Data[0]][(u8Data[1] - 1)];

    return NO_ERROR_MSG;
}

/// @fn         SPDDDAPI::getPartNumber(string in_strDIMM,
///                                     string *out_strpNumber)
///
/// @brief      Get the part number based on DIMM's SPD
/// @param      in_strDIMM is a string for SPD device name
/// @param      out_strpNumber is a pointer to a string for the part number
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SPDDDAPI::getPartNumber(string in_strDIMM,
                        string *out_strpNumber)
{
    ErrorMsg*   pErrorMsg = NO_ERROR_MSG;
    DDRType     eDDRType = DDR_TYPE_UNKNOWN;
    INT32       s32StartAddr = 0;
    INT32       s32EndAddr = 0;
    UINT8       u8Data[SPD_DDR4_OFFSET_MODULE_PART_NUMBER + 1]; // Use max length as buffer size
    INT32       s32Count = 0;

    if (out_strpNumber == NULL)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SPD_DDAPI_GET_PART_NUMBER | ERROR_BAD_PARAMETER_TWO),
                              "NULL pointer",
                              NULL);
    }

    out_strpNumber->clear();

    if ((pErrorMsg = getDeviceType(in_strDIMM, &eDDRType)))
    {
        return returnErrorMsg(ERROR_SPD_DDAPI_GET_PART_NUMBER, NULL, pErrorMsg);
    }

    memset(u8Data, '\0', sizeof(u8Data));

    switch (eDDRType)
    {
    case DDR3:
        s32StartAddr = SPD_DDR3_OFFSET_MODULE_PART_NUMBER;
        s32EndAddr = s32StartAddr + SPD_DDR3_LENGTH_MODULE_PART_NUMBER - 1;
        break;
    case DDR4:
        s32StartAddr = SPD_DDR4_OFFSET_MODULE_PART_NUMBER;
        s32EndAddr = s32StartAddr + SPD_DDR4_LENGTH_MODULE_PART_NUMBER - 1;
        break;
    default:
        // Do not have to consider this error because getDeviceType() will return error when the device type is unknown
        break;
    }

    if ((pErrorMsg = m_pI2CDAPI->Read(in_strDIMM, s32StartAddr, s32EndAddr, u8Data, &s32Count)))
    {
        return returnErrorMsg(ERROR_SPD_DDAPI_GET_PART_NUMBER, NULL, pErrorMsg);
    }

    *out_strpNumber = (CHAR *)u8Data;

    return NO_ERROR_MSG;
}

/// @fn         SPDDDAPI::getRankCount(string in_strDIMM,
///                                    UINT32 *out_u32pCount)
///
/// @brief      Get the number of ranks based on DIMM's SPD
/// @param      in_strDIMM is a string for SPD device name
/// @param      out_u32pCount is a pointer to UINT32 for rank count
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SPDDDAPI::getRankCount(string in_strDIMM,
                       UINT32 *out_u32pCount)
{
    ErrorMsg*   pErrorMsg = NO_ERROR_MSG;
    DDRType     eDDRType = DDR_TYPE_UNKNOWN;
    INT32       s32Offset = 0;
    UINT8       u8Data;
    INT32       s32Count = 0;
    string      strError;

    if (out_u32pCount == NULL)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SPD_DDAPI_GET_RANK_COUNT | ERROR_BAD_PARAMETER_TWO),
                              "NULL pointer",
                              NULL);
    }

    *out_u32pCount = 0;

    if ((pErrorMsg = getDeviceType(in_strDIMM, &eDDRType)))
    {
        return returnErrorMsg(ERROR_SPD_DDAPI_GET_RANK_COUNT, NULL, pErrorMsg);
    }

    switch (eDDRType)
    {
        case DDR3:
            s32Offset = SPD_DDR3_OFFSET_MODULE_ORGANIZATION;
            break;
        case DDR4:
            s32Offset = SPD_DDR4_OFFSET_MODULE_ORGANIZATION;
            break;
        default:
            break;
    }

    if ((pErrorMsg = m_pI2CDAPI->Read(in_strDIMM, s32Offset, s32Offset, &u8Data, &s32Count)))
    {
        return returnErrorMsg(ERROR_SPD_DDAPI_GET_RANK_COUNT, NULL, pErrorMsg);
    }

    // Parse for number of ranks
    u8Data = (u8Data & 0x38) >> 3;
    if (u8Data > 0x3)
    {
        strError = "unknown number of ranks(" + Int2HexStr(u8Data) + ")";
        return returnErrorMsg((ERRORCODE)(ERROR_SPD_DDAPI_GET_RANK_COUNT | ERROR_UNEXPECTED_VALUE),
                              strError.c_str());
    }

    *out_u32pCount = 1 + u8Data;

    return NO_ERROR_MSG;
}

/// @fn         SPDDDAPI::getSerialNumber(string in_strDIMM,
///                                       string *out_strpNumber)
///
/// @brief      Get the serial number based on DIMM's SPD
/// @param      in_strDIMM is a string for SPD device name
/// @param      out_strpNumber is a pointer to a string for the serial number
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SPDDDAPI::getSerialNumber(string in_strDIMM,
                          string *out_strpNumber)
{
    ErrorMsg*   pErrorMsg = NO_ERROR_MSG;
    DDRType     eDDRType = DDR_TYPE_UNKNOWN;
    INT32       s32StartAddr = 0;
    INT32       s32EndAddr = 0;
    UINT8       u8Data[SPD_LENGTH_MODULE_SERIAL_NUMBER];
    INT32       s32Count = 0;
    INT32       i;

    if (out_strpNumber == NULL)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SPD_DDAPI_GET_SERIAL_NUMBER | ERROR_BAD_PARAMETER_TWO),
                              "NULL pointer",
                              NULL);
    }

    out_strpNumber->clear();

    if ((pErrorMsg = getDeviceType(in_strDIMM, &eDDRType)))
    {
        return returnErrorMsg(ERROR_SPD_DDAPI_GET_SERIAL_NUMBER, NULL, pErrorMsg);
    }

    memset(u8Data, '\0', sizeof(u8Data));

    switch (eDDRType)
    {
    case DDR3:
        s32StartAddr = SPD_DDR3_OFFSET_MODULE_SERIAL_NUMBER;
        s32EndAddr = s32StartAddr + SPD_LENGTH_MODULE_SERIAL_NUMBER - 1;
        break;
    case DDR4:
        s32StartAddr = SPD_DDR4_OFFSET_MODULE_SERIAL_NUMBER;
        s32EndAddr = s32StartAddr + SPD_LENGTH_MODULE_SERIAL_NUMBER - 1;
        break;
    default:
        // Do not have to consider this error because getDeviceType() will return error when the device type is unknown
        break;
    }

    if ((pErrorMsg = m_pI2CDAPI->Read(in_strDIMM, s32StartAddr, s32EndAddr, u8Data, &s32Count)))
    {
        return returnErrorMsg(ERROR_SPD_DDAPI_GET_SERIAL_NUMBER, NULL, pErrorMsg);
    }

    for (i = 0; i < s32Count; i++)
    {
        out_strpNumber->append(Int2HexStr(u8Data[i]));
    }

    return NO_ERROR_MSG;
}

/// @fn         SPDDDAPI::getSize(string in_strDIMM,
///                               UINT64 *out_u64pSize)
///
/// @brief      Get memory size based on DIMM's SPD
/// @param      in_strDIMM is a string for SPD device name
/// @param      out_u64pSize is a pointer to UINT64 for memory size
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SPDDDAPI::getSize(string in_strDIMM,
                  UINT64 *out_u64pSize)
{
    ErrorMsg*   pErrorMsg;
    INT32       s32Offset;
    UINT8       u8Data;
    INT32       s32Count = 0;
    DDRType     eDDRType;
    string      strError;
    UINT32      u32Capacity;
    UINT32      u32BusWidth;
    UINT32      u32SDRAMWidth;
    UINT32      u32Rank;

    if (out_u64pSize == NULL)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SPD_DDAPI_GET_SIZE | ERROR_BAD_PARAMETER_TWO),
                              "NULL pointer",
                              NULL);
    }

    *out_u64pSize = 0;

    if ((pErrorMsg = getDeviceType(in_strDIMM, &eDDRType)))
    {
        return returnErrorMsg(ERROR_SPD_DDAPI_GET_SIZE, NULL, pErrorMsg);
    }

    // Get density
    s32Offset = SPD_OFFSET_SDRAM_DENSITY_BANKS;

    if ((pErrorMsg = m_pI2CDAPI->Read(in_strDIMM, s32Offset, s32Offset, &u8Data, &s32Count)))
    {
        return returnErrorMsg(ERROR_SPD_DDAPI_GET_SIZE, NULL, pErrorMsg);
    }

    u8Data &= 0xF;
    switch (eDDRType)
    {
        case DDR3:
            if (u8Data > 0x6)
            {
                strError = "unknown SDRAM u32Capacity (0x" + Int2HexStr(u8Data) + ")";
                return returnErrorMsg((ERRORCODE)(ERROR_SPD_DDAPI_GET_SIZE | ERROR_UNEXPECTED_VALUE),
                                      strError.c_str());
            }
            break;
        case DDR4:
            if (u8Data > 0x7)
            {
                strError = "unknown SDRAM u32Capacity (0x" + Int2HexStr(u8Data) + ")";
                return returnErrorMsg((ERRORCODE)(ERROR_SPD_DDAPI_GET_SIZE | ERROR_UNEXPECTED_VALUE),
                                      strError.c_str());
            }
            break;
        default:
            break;
    }

    u32Capacity = (UINT32)256 << u8Data;

    // Get memory primary bus width
    switch (eDDRType)
    {
        case DDR3:
            s32Offset = SPD_DDR3_OFFSET_MODULE_MEM_BUS_WIDTH;
            break;
        case DDR4:
            s32Offset = SPD_DDR4_OFFSET_MODULE_MEM_BUS_WIDTH;
            break;
        default:
            break;
    }

    if ((pErrorMsg = m_pI2CDAPI->Read(in_strDIMM, s32Offset, s32Offset, &u8Data, &s32Count)))
    {
        return returnErrorMsg(ERROR_SPD_DDAPI_GET_SIZE, NULL, pErrorMsg);
    }

    u8Data &= 0x7;

    if (u8Data > 0x3)
    {
        strError = "unknown bus width(0x" + Int2HexStr(u8Data) + ")";
        return returnErrorMsg((ERRORCODE)(ERROR_SPD_DDAPI_GET_SIZE | ERROR_UNEXPECTED_VALUE),
                              strError.c_str());
    }

    u32BusWidth = (UINT32)(8 << u8Data);

    // Get memory rank and SDRAM device width
    switch (eDDRType)
    {
        case DDR3:
            s32Offset = SPD_DDR3_OFFSET_MODULE_ORGANIZATION;
            break;
        case DDR4:
            s32Offset = SPD_DDR4_OFFSET_MODULE_ORGANIZATION;
            break;
        default:
            break;
    }

    if ((pErrorMsg = m_pI2CDAPI->Read(in_strDIMM, s32Offset, s32Offset, &u8Data, &s32Count)))
    {
        return returnErrorMsg(ERROR_SPD_DDAPI_GET_SIZE, NULL, pErrorMsg);
    }

    // Parse for SDRAM device width
    if ((u8Data & 0x7) > 0x3)
    {
        strError = "unknown SDRAM device width(0x" + Int2HexStr(u8Data & 0x7) + ")";
        return returnErrorMsg((ERRORCODE)(ERROR_SPD_DDAPI_GET_SIZE | ERROR_UNEXPECTED_VALUE),
                              strError.c_str());
    }

    u32SDRAMWidth = (UINT32)(4 << (u8Data & 0x7));

    // Parse for number of ranks
    u8Data = (u8Data & 0x38) >> 3;
    if (u8Data > 0x3)
    {
        strError = "unknown number of ranks(0x" + Int2HexStr(u8Data) + ")";
        return returnErrorMsg((ERRORCODE)(ERROR_SPD_DDAPI_GET_SIZE | ERROR_UNEXPECTED_VALUE),
                              strError.c_str());
    }

    u32Rank = 1 + u8Data;

    if (eDDRType == DDR4)
    {
        s32Offset = SPD_DDR4_OFFSET_SDRAM_PACKAGE_TYPE;

        if ((pErrorMsg = m_pI2CDAPI->Read(in_strDIMM, s32Offset, s32Offset, &u8Data, &s32Count)))
        {
            return returnErrorMsg(ERROR_SPD_DDAPI_GET_SIZE, NULL, pErrorMsg);
        }

        // Non-monolithic and single load stack (3DS)
        if (((u8Data & 0x80) == 0x80) && ((u8Data & 0x3) == 2))
        {
            if (u32Rank > 2)
            {
                return returnErrorMsg((ERRORCODE)(ERROR_SPD_DDAPI_GET_SIZE | ERROR_UNEXPECTED_VALUE),
                                      "unexpected number of ranks for 3DS");
            }

            // Multiply by die count
            u32Rank *= (((u8Data & 0x70) >> 4) + 1);
        }

    }

    // Calculate total size
    *out_u64pSize = (UINT64)(u32Capacity / 8 * u32BusWidth / u32SDRAMWidth * u32Rank) << 20;

    return NO_ERROR_MSG;
}

/// @fn         SPDDDAPI::getSpeed(string in_strDIMM,
///                                UINT32 *out_u32pSpeed)
///
/// @brief      Get speed based on DIMM's SPD
/// @param      in_strDIMM is a string for SPD device name
/// @param      out_u32pSpeed is a pointer to UINT32 for speed
/// @exception  Nothing
/// @return     A pointer to ErrorMsg for error message list
ErrorMsg*
SPDDDAPI::getSpeed(string in_strDIMM,
                   UINT32 *out_u32pSpeed)
{
    ErrorMsg*   pErrorMsg;
    UINT8       u8Data;
    INT32       s32Count = 0;
    DDRType     eDDRType;
    string      strError;

    if (out_u32pSpeed == NULL)
    {
        return returnErrorMsg(ERRORCODE(ERROR_SPD_DDAPI_GET_SPEED | ERROR_BAD_PARAMETER_TWO),
                              "NULL pointer",
                              NULL);
    }

    *out_u32pSpeed = 0;

    if ((pErrorMsg = getDeviceType(in_strDIMM, &eDDRType)))
    {
        return returnErrorMsg(ERROR_SPD_DDAPI_GET_SPEED, NULL, pErrorMsg);
    }

    switch (eDDRType)
    {
        case DDR3:
            pErrorMsg = m_pI2CDAPI->Read(in_strDIMM,
                                         SPD_DDR3_OFFSET_MINI_CYCLE_TIME,
                                         SPD_DDR3_OFFSET_MINI_CYCLE_TIME,
                                         &u8Data,
                                         &s32Count);
            if (pErrorMsg)
            {
                return returnErrorMsg(ERROR_SPD_DDAPI_GET_SIZE, NULL, pErrorMsg);
            }

            switch (u8Data)
            {
                case 0x14:
                    *out_u32pSpeed = 800;
                    break;
                case 0xF:
                    *out_u32pSpeed = 1066;
                    break;
                case 0xC:
                    *out_u32pSpeed = 1333;
                    break;
                case 0xA:
                    *out_u32pSpeed = 1600;
                    break;
                case 0x9:
                    *out_u32pSpeed = 1866;
                    break;
                case 0x8:
                    *out_u32pSpeed = 2133;
                    break;
                default:
                    strError = "unknown speed(0x" + Int2HexStr(u8Data) + ")";
                    return returnErrorMsg((ERRORCODE)(ERROR_SPD_DDAPI_GET_SPEED | ERROR_UNEXPECTED_VALUE),
                                          strError.c_str());
            }
            break;
        case DDR4:
            pErrorMsg = m_pI2CDAPI->Read(in_strDIMM,
                                         SPD_DDR4_OFFSET_MINI_CYCLE_TIME,
                                         SPD_DDR4_OFFSET_MINI_CYCLE_TIME,
                                         &u8Data,
                                         &s32Count);
            if (pErrorMsg)
            {
                return returnErrorMsg(ERROR_SPD_DDAPI_GET_SIZE, NULL, pErrorMsg);
            }

            switch (u8Data)
            {
                case 0xA:
                    *out_u32pSpeed = 1600;
                    break;
                case 0x9:
                    *out_u32pSpeed = 1866;
                    break;
                case 0x8:
                    *out_u32pSpeed = 2133;
                    break;
                case 0x7:
                    *out_u32pSpeed = 2400;
                    break;
                case 0x6:
                    *out_u32pSpeed = 2666;
                    break;
                case 0x5:
                    *out_u32pSpeed = 3200;
                    break;
                default:
                    strError = "unknown speed(0x" + Int2HexStr(u8Data) + ")";
                    return returnErrorMsg((ERRORCODE)(ERROR_SPD_DDAPI_GET_SPEED | ERROR_UNEXPECTED_VALUE),
                                          strError.c_str());
            }
            break;
        default:
            // Do not have to consider this error because getDeviceType() will return error when the device type is unknown
            break;
    }

    return NO_ERROR_MSG;
}
