/*
 * PMC-Sierra SPC 8001 SAS/SATA based host adapters driver
 *
 * Copyright (c) 2008-2009 USI Co., Ltd.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification.
 * 2. Redistributions in binary form must reproduce at minimum a disclaimer
 *    substantially similar to the "NO WARRANTY" disclaimer below
 *    ("Disclaimer") and any redistribution must be conditioned upon
 *    including a substantially similar Disclaimer requirement for further
 *    binary redistribution.
 * 3. Neither the names of the above-listed copyright holders nor the names
 *    of any contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * NO WARRANTY
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 */
#include <linux/firmware.h>
#include <linux/slab.h>
#include "pm8001_sas.h"
#include "pm80xx_hwi.h"
int pm8001_major = -1;


#include <linux/poll.h>
#include <linux/mutex.h>
#include <linux/miscdevice.h>
#include "pm8001_ctl.h"

static DEFINE_MUTEX(_ctl_mutex);

static struct fasync_struct *async_queue;

#if 0
/*
 * pm8001_ctl_fasync -
 * @fd -
 * @filep -
 * @mode -
 *
 * Called when application request fasyn callback handler.
 */
static int pm8001_ctl_fasync(int fd, struct file *filep, int mode)
{
	return fasync_helper(fd, filep, mode, &async_queue);
}

/*
 * pm8001_ctl_release -
 * @inode -
 * @filep -
 *
 * Called when application releases the fasyn callback handler.
 */
static int pm8001_ctl_release(struct inode *inode, struct file *filep)
{
	return fasync_helper(-1, filep, 0, &async_queue);
}

#endif

/**
 * pm8001_ctl_verify_adapter - validates chip_number passed from application
 * @ppm8001_ha: per adapter object
 * @chip_number: chip id.
 *
 * Return (-1) means error, else chip_number.
 */
static int
pm8001_ctl_verify_adapter(int chip_number, struct pm8001_hba_info **ppm8001_ha)
{
	struct pm8001_hba_info *pm8001_ha;

	list_for_each_entry(pm8001_ha, &hba_list, list) {
		if (pm8001_ha->id != chip_number)
		{
			printk(KERN_ERR "pm8001_ctl_verify_adapter:id=0x%2x chip_number=0x%2x\n", pm8001_ha->id, chip_number );
			continue;
		}
		*ppm8001_ha = pm8001_ha;
		return chip_number;
	}
	*ppm8001_ha = NULL;
	return -1;
}

/* Give it some long enough timeout. In seconds. */
#define SMP_TIMEOUT 10

static int pm8001_smp_execute_task(struct domain_device *dev, void *req, int req_size,
			    void *resp, int resp_size)
{
	int res, retry;
	struct sas_task *task = NULL;

	for (retry = 0; retry < 3; retry++) {
		task = sas_alloc_slow_task(GFP_KERNEL);
		if (!task)
			return -ENOMEM;

		task->dev = dev;
		task->task_proto = dev->tproto;
		sg_init_one(&task->smp_task.smp_req, req, req_size);
		sg_init_one(&task->smp_task.smp_resp, resp, resp_size);

		task->task_done = pm8001_task_done;

		task->slow_task->timer.data = (unsigned long) task;
		task->slow_task->timer.function = pm8001_tmf_timedout;
		task->slow_task->timer.expires = jiffies + SMP_TIMEOUT*HZ;
		add_timer(&task->slow_task->timer);

		res = pm8001_queue_command(task, 1, GFP_KERNEL);

		if (res) {
			del_timer(&task->slow_task->timer);
			pm8001_printk("Executing SMP task failed:%d\n", res);
			goto ex_err;
		}

		wait_for_completion(&task->slow_task->completion);
		//res = -ETASK;
		res = -1;
		if ((task->task_state_flags & SAS_TASK_STATE_ABORTED)) {
			pm8001_printk("smp task timed out or aborted\n");
			pm8001_abort_task(task);
			if (!(task->task_state_flags & SAS_TASK_STATE_DONE)) {
				pm8001_printk("SMP task aborted and not done\n");
				goto ex_err;
			}
		}
		if (task->task_status.resp == SAS_TASK_COMPLETE &&
		    task->task_status.stat == SAM_STAT_GOOD) {
			res = 0;
			break;
		} if (task->task_status.resp == SAS_TASK_COMPLETE &&
		      task->task_status.stat == SAS_DATA_UNDERRUN) {
			/* no error, but return the number of bytes of
			 * underrun */
			res = task->task_status.residual;
			break;
		} if (task->task_status.resp == SAS_TASK_COMPLETE &&
		      task->task_status.stat == SAS_DATA_OVERRUN) {
			res = -EMSGSIZE;
			break;
		} else {
			pm8001_printk("%s: task to dev %016llx response: 0x%x "
				    "status 0x%x\n", __FUNCTION__,
				    SAS_ADDR(dev->sas_addr),
				    task->task_status.resp,
				    task->task_status.stat);
			sas_free_task(task);
			task = NULL;
		}
	}
ex_err:
	BUG_ON(retry == 3 && task != NULL);
	if (task != NULL) {
		sas_free_task(task);
	}
	return res;
}
struct domain_device *pm8001_find_expander_by_sas_addr(struct pm8001_hba_info *pm8001_ha, u8 *sas_addr)
{
	u32 i;
	for (i = 0; i < PM8001_MAX_DEVICES; i++) {
		if (pm8001_ha->devices[i].sas_device) {
			struct domain_device *dev = pm8001_ha->devices[i].sas_device;
			if (SAS_ADDR(sas_addr)
			    == SAS_ADDR(dev->sas_addr) && DEV_IS_EXPANDER(dev->dev_type))
				return dev;
		}
	}
	PM8001_FAIL_DBG(pm8001_ha,
			pm8001_printk("max support %d devices, ignore ..\n",
				      PM8001_MAX_DEVICES));
	return NULL;
}

/**
 * _ctl_do_mpt_command - main handler for MPT2COMMAND opcode
 * @pm8001_ha: per adapter object
 * @karg - (struct adpt_ioctl_command)
 * @mf - pointer to mf in user space
 */
static long
pm8001_ctl_do_smp_command(struct pm8001_hba_info *pm8001_ha,
    struct adpt_ioctl_command karg, void __user *mf)
{
	struct PM8001_SMP_PASSTHROUGH_REQUEST *smp_request;
	void *data_out = NULL;
	size_t data_out_sz = 0;
	void *data_in = NULL;
	size_t data_in_sz = 0;
	int ret;
	struct domain_device *dev;
	u8	sas_addr[SAS_ADDR_SIZE];
	u64 addr;

	smp_request = kzalloc(karg.dataSgeOffset*4, GFP_KERNEL);
	if (!smp_request) {
			printk(KERN_ERR "%s: failed obtaining a memory for "
		    "mpi_request\n", pm8001_ha->name);
			ret = -ENOMEM;
			goto out;
		}

		/* Check for overflow and wraparound */
		if (karg.dataSgeOffset * 4 > 64 ||
			karg.dataSgeOffset > (UINT_MAX / 4)) {
			ret = -EINVAL;
			goto out;
		}

		/* copy in request message frame from user */
		if (copy_from_user(smp_request, mf, karg.dataSgeOffset*4)) {
			printk(KERN_ERR "failure at %s:%d/%s()!\n", __FILE__, __LINE__,
				__func__);
			ret = -EFAULT;
			goto out;
	}
		addr = cpu_to_be64((u64)
				(*(u64 *)&smp_request->sas_addr));;
		memcpy(sas_addr, &addr, 8);

	dev = pm8001_find_expander_by_sas_addr(pm8001_ha, sas_addr);
	if (!dev) {
		printk(KERN_ERR "dev is null,failure at %s:%d/%s()!\n", __FILE__, __LINE__,
						__func__);
		ret = -ENODEV;
		goto out;
		}
	ret = 0;
	data_out_sz = karg.dataOutSize + 4;
	data_in_sz = karg.dataInSize - 4 ;

	/* obtain dma-able memory for data transfer */
	if (data_out_sz) /* WRITE */ {
		data_out = kzalloc(data_out_sz, GFP_KERNEL);
		if (!data_out) {
			printk(KERN_ERR "failure at %s:%d/%s()!\n", __FILE__,
			    __LINE__, __func__);
			ret = -ENOMEM;;
			goto out;
		}
		if (copy_from_user(data_out, karg.dataOutBufPtr,
			data_out_sz)) {
			printk(KERN_ERR "failure at %s:%d/%s()!\n", __FILE__,
			    __LINE__, __func__);
			ret =  -EFAULT;
			goto out;
		}
	}

	if (data_in_sz) /* READ */ {
		data_in = kzalloc(data_in_sz, GFP_KERNEL);
		if (!data_in) {
			printk(KERN_ERR "failure at %s:%d/%s()!\n", __FILE__,
			    __LINE__, __func__);
			ret = -ENOMEM;
			goto out;
		}
	}

	ret = pm8001_smp_execute_task(dev, data_out, data_out_sz, data_in,
				       data_in_sz);

	if (ret) {
			pm8001_printk("send SMP to ex %016llx failed:0x%x\n",
				    SAS_ADDR(dev->sas_addr), ret);
			goto out;
		}


	/* copy out xdata to user */
	if (data_in_sz) {
		if (copy_to_user(karg.dataInBufPtr, data_in,
		    data_in_sz)) {
			printk(KERN_ERR "failure at %s:%d/%s()!\n", __FILE__,
			    __LINE__, __func__);
			ret = -ENODATA;
			goto out;
		}
	}


 out:

	/* free memory associated with sg buffers */
	if (smp_request)
		kfree(smp_request);
	if (data_in)
		kfree(data_in);


	if (data_out)
		kfree(data_out);
	return ret;
}

/**
 * pm8001_ctl_ioctl_smp - SMP ioctl entry point
 * @file - (struct file)
 * @cmd - ioctl opcode
 * @arg -
 */
static long
pm8001_ctl_ioctl_smp(struct file *file, unsigned int cmd, void __user *arg)
{
		long ret = -EINVAL;
		struct adpt_ioctl_command karg;
		struct adpt_ioctl_command __user *uarg;
		struct pm8001_hba_info *pm8001_ha;
		if (copy_from_user(&karg, arg, sizeof(karg))) {
					printk(KERN_ERR "failure at %s:%d/%s()!\n",
						__FILE__, __LINE__, __func__);
					return -EFAULT;
		}

		if (pm8001_ctl_verify_adapter(karg.hdr.iocnum, &pm8001_ha) == -1 ||
					!pm8001_ha)
				return -ENODEV;

			uarg = arg;
		ret = pm8001_ctl_do_smp_command(pm8001_ha, karg, &uarg->MF);
		PM8001_IOCTL_DBG(pm8001_ha,
			pm8001_printk(" cmd: IOCTL_SMP_CMND_PASSTHROUGH\n"));
		return ret;

}

#if 0
/**
 * pm8001_ctl_ioctl_main - main ioctl entry point
 * @file - (struct file)
 * @cmd - ioctl opcode
 * @arg -
 */
static long
pm8001_ctl_ioctl_main(struct file *file, unsigned int cmd, void __user *arg)
{
	long ret = -EINVAL;

	u16 arg_length;
	struct pm8001_ioctl_payload __user *payload = (void __user *) arg;
	struct pm8001_ioctl_payload	*ioctl_payload ;
	DECLARE_COMPLETION_ONSTACK(completion);
	struct pm8001_hba_info *pm8001_ha = NULL;
	arg_length = sizeof(struct pm8001_ioctl_payload) + payload->length;
	ioctl_payload = kzalloc(arg_length, GFP_KERNEL);
	if (!ioctl_payload) {
				printk(KERN_ERR "pm8001: failed obtaining a memory for "
				"ioctl_payload\n");
			ret = -ENOMEM;
			goto out;
		}

	if (copy_from_user(ioctl_payload, payload, arg_length)) {
		pm8001_printk("copy data from user space @ %p failed\n",
		payload);
		ret = -EFAULT;
		goto out;
	}

	if (pm8001_ctl_verify_adapter(ioctl_payload->id, &pm8001_ha) == -1 ||
		!pm8001_ha) {
		pm8001_printk("find the chip fail\n");
		ret =  -ENODEV;
		goto out;
	}

	pm8001_ha->nvmd_completion = &completion;
	switch (cmd) {
	case IOCTL_GET_EVENT_LOG1:
		PM8001_IOCTL_DBG(pm8001_ha,
			pm8001_printk("cmd:IOCTL_GET_EVENT_LOG1\n"));
		ioctl_payload->status = IOCTL_STATUS_OK;
		if ((4096 + 32) < ioctl_payload->offset) {
			ioctl_payload->status = IOCTL_STATUS_NO_MORE_DATA;
			ioctl_payload->length = 0;
			ret = IOCTL_CALL_FAIL;
		}
		memcpy(ioctl_payload->func_specific,
		pm8001_ha->memoryMap.region[AAP1].virt_ptr +
		ioctl_payload->offset, ioctl_payload->length);
		if (copy_to_user((void *)arg, (void *)ioctl_payload,
			arg_length)) {
			PM8001_IOCTL_DBG(pm8001_ha,
				pm8001_printk("copy to user ERROR\n"));
			ret = -EFAULT;
		}
		goto out;

	case IOCTL_GET_EVENT_LOG2:
		PM8001_IOCTL_DBG(pm8001_ha,
			pm8001_printk(" cmd: IOCTL_GET_EVENT_LOG2\n"));
		ioctl_payload->status = IOCTL_STATUS_OK;
		if ((4096+32) < ioctl_payload->offset) {
			ioctl_payload->status = IOCTL_STATUS_NO_MORE_DATA;
			ioctl_payload->length = 0;
			ret = IOCTL_CALL_FAIL;
		}
		memcpy(ioctl_payload->func_specific,
			(u8 *)pm8001_ha->memoryMap.region[IOP].virt_ptr,
			ioctl_payload->length);
		if (copy_to_user((void *)arg,
			(void *)ioctl_payload, arg_length)) {
			PM8001_IOCTL_DBG(pm8001_ha,
				pm8001_printk("copy to user ERROR\n"));
			ret = -EFAULT;
		}
		goto out;
	case IOCTL_FW_CONTROL:
		PM8001_IOCTL_DBG(pm8001_ha,
			pm8001_printk(" cmd: IOCTL_FW_CONTROL\n"));
		ret = PM8001_CHIP_DISP->fw_flash_update_req(pm8001_ha,
			ioctl_payload);
		break;
	case IOCTL_NVMD_GET:
		PM8001_IOCTL_DBG(pm8001_ha,
			pm8001_printk(" cmd: IOCTL_NVMD_GET\n"));
		ret = PM8001_CHIP_DISP->get_nvmd_req(pm8001_ha,
			ioctl_payload);
		break;
	case IOCTL_NVMD_SET:
		PM8001_IOCTL_DBG(pm8001_ha,
			pm8001_printk(" cmd: IOCTL_NVMD_SET\n"));
		ret = PM8001_CHIP_DISP->set_nvmd_req(pm8001_ha,
			ioctl_payload);
		break;
	case IOCTL_SAS_DIAG_START_END:
		PM8001_IOCTL_DBG(pm8001_ha,
			pm8001_printk("cmd: IOCTL_SAS_DIAG_START_END\n"));
		ret = PM8001_CHIP_DISP->sas_diag_start_end_req(pm8001_ha,
			ioctl_payload);
		break;
	case IOCTL_SAS_DIAG_EXECUTE:
		PM8001_IOCTL_DBG(pm8001_ha,
			pm8001_printk(" cmd: IOCTL_SAS_DIAG_EXECUTE\n"));
		ret = PM8001_CHIP_DISP->sas_diag_execute_req(pm8001_ha,
			ioctl_payload);
		break;

	case IOCTL_SET_PHY_STATE:
		PM8001_IOCTL_DBG(pm8001_ha,
			pm8001_printk(" cmd: IOCTL_SET_PHY_STATE\n"));
		pm8001_ha->flags |= PM8001F_USER_TIME;
		ret = PM8001_CHIP_DISP->set_phy_state_req(pm8001_ha,
			ioctl_payload);
		break;
	default:
		ret = IOCTL_CALL_INVALID_CODE;
		goto out;
	}
	wait_for_completion(&completion);
	if (copy_to_user((void *)arg, (void *)ioctl_payload, arg_length)) {
		PM8001_IOCTL_DBG(pm8001_ha,
			pm8001_printk("copy to user ERROR\n"));
		ret = -EFAULT;
	}

out:
	if (ioctl_payload)
	kfree(ioctl_payload);
	return ret;
}

/**
 * pm8001_ctl_ioctl - main ioctl entry point (unlocked)
 * @file - (struct file)
 * @cmd - ioctl opcode
 * @arg -
 */
static long
pm8001_ctl_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	long ret;

	mutex_lock(&_ctl_mutex);


	if (cmd == IOCTL_SMP_CMND_PASSTHROUGH)
	{
        ret = pm8001_ctl_ioctl_smp(file, cmd, (void __user *)arg);
		printk(KERN_ERR "reached: IOCTL_SMP_CMND_PASSTHROUGH:ret = 0x%2lx\n", ret);
	}
	else
	{
        ret = pm8001_ctl_ioctl_main(file, cmd, (void __user *)arg);
		printk(KERN_ERR "reached: pm8001_ctl_ioctl_main:ret = 0x%2lx\n", ret);
	}

	mutex_unlock(&_ctl_mutex);

	return ret;
}
#endif /* 0 */

/* scsi host attributes */

/**
 * pm8001_ctl_mpi_interface_rev_show - MPI interface revision number
 * @cdev: pointer to embedded class device
 * @buf: the buffer returned
 *
 * A sysfs 'read-only' shost attribute.
 */
static ssize_t pm8001_ctl_mpi_interface_rev_show(struct device *cdev,
	struct device_attribute *attr, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
	struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;

	if (pm8001_ha->chip_id == chip_8001)
	{
		return snprintf(buf, PAGE_SIZE, "%d\n", 
			pm8001_ha->main_cfg_tbl.pm8001_tbl.interface_rev);
	}
	else
	{
		return snprintf(buf, PAGE_SIZE, "%d\n", 
			pm8001_ha->main_cfg_tbl.pm80xx_tbl.interface_rev);
	}	
}
static
DEVICE_ATTR(interface_rev, S_IRUGO, pm8001_ctl_mpi_interface_rev_show, NULL);

/**
 * pm8001_ctl_fw_version_show - firmware version
 * @cdev: pointer to embedded class device
 * @buf: the buffer returned
 *
 * A sysfs 'read-only' shost attribute.
 */
static ssize_t pm8001_ctl_fw_version_show(struct device *cdev,
	struct device_attribute *attr, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
	struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;

	if (pm8001_ha->chip_id == chip_8001) {
		return snprintf(buf, PAGE_SIZE, "%02x.%02x.%02x.%02x\n",
		       (u8)(pm8001_ha->main_cfg_tbl.pm8001_tbl.firmware_rev >> 24),
		       (u8)(pm8001_ha->main_cfg_tbl.pm8001_tbl.firmware_rev >> 16),
		       (u8)(pm8001_ha->main_cfg_tbl.pm8001_tbl.firmware_rev >> 8),
		       (u8)(pm8001_ha->main_cfg_tbl.pm8001_tbl.firmware_rev));
	}
	else {
		return snprintf(buf, PAGE_SIZE, "%02x.%02x.%02x.%02x\n",
		       (u8)(pm8001_ha->main_cfg_tbl.pm80xx_tbl.firmware_rev >> 24),
		       (u8)(pm8001_ha->main_cfg_tbl.pm80xx_tbl.firmware_rev >> 16),
		       (u8)(pm8001_ha->main_cfg_tbl.pm80xx_tbl.firmware_rev >> 8),
		       (u8)(pm8001_ha->main_cfg_tbl.pm80xx_tbl.firmware_rev));
	}	
}
static DEVICE_ATTR(fw_version, S_IRUGO, pm8001_ctl_fw_version_show, NULL);
/**
 * pm8001_ctl_max_out_io_show - max outstanding io supported
 * @cdev: pointer to embedded class device
 * @buf: the buffer returned
 *
 * A sysfs 'read-only' shost attribute.
 */
static ssize_t pm8001_ctl_max_out_io_show(struct device *cdev,
	struct device_attribute *attr, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
	struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;

	if (pm8001_ha->chip_id == chip_8001) {
		return snprintf(buf, PAGE_SIZE, "%d\n",
			pm8001_ha->main_cfg_tbl.pm8001_tbl.max_out_io);
	}
	else {
		return snprintf(buf, PAGE_SIZE, "%d\n",
			pm8001_ha->main_cfg_tbl.pm80xx_tbl.max_out_io);
	}	
}
static DEVICE_ATTR(max_out_io, S_IRUGO, pm8001_ctl_max_out_io_show, NULL);
/**
 * pm8001_ctl_max_devices_show - max devices support
 * @cdev: pointer to embedded class device
 * @buf: the buffer returned
 *
 * A sysfs 'read-only' shost attribute.
 */
static ssize_t pm8001_ctl_max_devices_show(struct device *cdev,
	struct device_attribute *attr, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
	struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;

	if (pm8001_ha->chip_id == chip_8001) {
		return snprintf(buf, PAGE_SIZE, "%04d\n",
			(u16)(pm8001_ha->main_cfg_tbl.pm8001_tbl.max_sgl >> 16));
	}
	else {
		return snprintf(buf, PAGE_SIZE, "%04d\n",
			(u16)(pm8001_ha->main_cfg_tbl.pm80xx_tbl.max_sgl >> 16));
	}
}
static DEVICE_ATTR(max_devices, S_IRUGO, pm8001_ctl_max_devices_show, NULL);
/**
 * pm8001_ctl_max_sg_list_show - max sg list supported iff not 0.0 for no
 * hardware limitation
 * @cdev: pointer to embedded class device
 * @buf: the buffer returned
 *
 * A sysfs 'read-only' shost attribute.
 */
static ssize_t pm8001_ctl_max_sg_list_show(struct device *cdev,
	struct device_attribute *attr, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
	struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;

	if (pm8001_ha->chip_id == chip_8001) {
		return snprintf(buf, PAGE_SIZE, "%04d\n",
			pm8001_ha->main_cfg_tbl.pm8001_tbl.max_sgl & 0x0000FFFF);
	}
	else {
		return snprintf(buf, PAGE_SIZE, "%04d\n",
			pm8001_ha->main_cfg_tbl.pm80xx_tbl.max_sgl & 0x0000FFFF);
	}
}
static DEVICE_ATTR(max_sg_list, S_IRUGO, pm8001_ctl_max_sg_list_show, NULL);

#define SAS_1_0 0x1
#define SAS_1_1 0x2
#define SAS_2_0 0x4

static ssize_t
show_sas_spec_support_status(unsigned int mode, char *buf)
{
	ssize_t len = 0;

	if (mode & SAS_1_1)
		len = sprintf(buf, "%s", "SAS1.1");
	if (mode & SAS_2_0)
		len += sprintf(buf + len, "%s%s", len ? ", " : "", "SAS2.0");
	len += sprintf(buf + len, "\n");

	return len;
}

/**
 * pm8001_ctl_sas_spec_support_show - sas spec supported
 * @cdev: pointer to embedded class device
 * @buf: the buffer returned
 *
 * A sysfs 'read-only' shost attribute.
 */
static ssize_t pm8001_ctl_sas_spec_support_show(struct device *cdev,
	struct device_attribute *attr, char *buf)
{
	unsigned int mode;
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
	struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;
	if (pm8001_ha->chip_id == chip_8001) {
		mode = (pm8001_ha->main_cfg_tbl.pm8001_tbl.ctrl_cap_flag & 0xfe000000)>>25; /* fe000000 means supports SAS2.1 */
	}
	else {
		mode = (pm8001_ha->main_cfg_tbl.pm80xx_tbl.ctrl_cap_flag & 0xfe000000)>>25; /* fe000000 means supports SAS2.1 */
	}
	return show_sas_spec_support_status(mode, buf);
}
static DEVICE_ATTR(sas_spec_support, S_IRUGO,
		   pm8001_ctl_sas_spec_support_show, NULL);

/**
 * pm8001_ctl_sas_address_show - sas address
 * @cdev: pointer to embedded class device
 * @buf: the buffer returned
 *
 * This is the controller sas address
 *
 * A sysfs 'read-only' shost attribute.
 */
static ssize_t pm8001_ctl_host_sas_address_show(struct device *cdev,
	struct device_attribute *attr, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
	struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;
	return snprintf(buf, PAGE_SIZE, "0x%016llx\n",
			be64_to_cpu(*(__be64 *)pm8001_ha->sas_addr));
}
static DEVICE_ATTR(host_sas_address, S_IRUGO,
		   pm8001_ctl_host_sas_address_show, NULL);

/**
 * pm8001_ctl_logging_level_show - logging level
 * @cdev: pointer to embedded class device
 * @buf: the buffer returned
 *
 * A sysfs 'read/write' shost attribute.
 */
static ssize_t pm8001_ctl_logging_level_show(struct device *cdev,
	struct device_attribute *attr, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
	struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;

	return snprintf(buf, PAGE_SIZE, "%08xh\n", pm8001_ha->logging_level);
}
static ssize_t pm8001_ctl_logging_level_store(struct device *cdev,
	struct device_attribute *attr, const char *buf, size_t count)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
	struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;
	int val = 0;

	if (sscanf(buf, "%x", &val) != 1)
		return -EINVAL;

	pm8001_ha->logging_level = val;
	/*  added to server different chips*/
	if(pm8001_ha->chip_id == chip_8001)
	{
	    pm8001_update_main_config_table(pm8001_ha);
	}
	else
	{
		pm80xx_update_main_config_table(pm8001_ha);
	}
	return strlen(buf);
}

static DEVICE_ATTR(logging_level, S_IRUGO | S_IWUSR,
	pm8001_ctl_logging_level_show, pm8001_ctl_logging_level_store);
/**
 * pm8001_ctl_aap_log_show - aap1 event log
 * @cdev: pointer to embedded class device
 * @buf: the buffer returned
 *
 * A sysfs 'read-only' shost attribute.
 */
static ssize_t pm8001_ctl_aap_log_show(struct device *cdev,
	struct device_attribute *attr, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
	struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;
	int i;
#define AAP1_MEMMAP(r, c) \
	(*(u32 *)((u8*)pm8001_ha->memoryMap.region[AAP1].virt_ptr + (r) * 32 \
	+ (c)))

	char *str = buf;
	int max = 2;
	for (i = 0; i < max; i++) {
		str += sprintf(str, "0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x"
			       "0x%08x 0x%08x\n",
			       AAP1_MEMMAP(i, 0),
			       AAP1_MEMMAP(i, 4),
			       AAP1_MEMMAP(i, 8),
			       AAP1_MEMMAP(i, 12),
			       AAP1_MEMMAP(i, 16),
			       AAP1_MEMMAP(i, 20),
			       AAP1_MEMMAP(i, 24),
			       AAP1_MEMMAP(i, 28));
	}

	return str - buf;
}
static DEVICE_ATTR(aap_log, S_IRUGO, pm8001_ctl_aap_log_show, NULL);
/**
 * pm8001_ctl_bios_version_show - Bios version Display
 * @cdev:pointer to embedded class device
 * @buf:the buffer returned
 * A sysfs 'read-only' shost attribute.
 */
static ssize_t pm8001_ctl_bios_version_show(struct device *cdev,
	struct device_attribute *attr,char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
	struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;
	char *str = buf;
	int BiosIndex;
	void *virt_addr;
	DECLARE_COMPLETION_ONSTACK(completion);
	struct pm8001_ioctl_payload payload;
	pm8001_ha->nvmd_completion = &completion;
	payload.minor_function = 7;
	payload.offset = 0; 
	payload.length = 4096;
	payload.func_specific = kzalloc(4096, GFP_KERNEL);
	PM8001_CHIP_DISP->get_nvmd_req(pm8001_ha, &payload);
	wait_for_completion(&completion);
	virt_addr = pm8001_ha->memoryMap.region[NVMD].virt_ptr;
        for(BiosIndex=BIOSOFFSET;BiosIndex<BIOS_OFFSET_LIMIT;BiosIndex++)
        {
		str += sprintf(str,"%c",*((u8 *)(virt_addr+BiosIndex)));
        }
	return str - buf;
}
static DEVICE_ATTR(bios_version,S_IRUGO,pm8001_ctl_bios_version_show,NULL);

 /* pm8001_ctl_ctrl_wwn_show - wwn Display
 * @cdev:pointer to embedded class device
 * @buf:the buffer returned
 * A sysfs 'read-only' shost attribute.
 */
static ssize_t pm8001_ctl_ctrl_wwn_show(struct device *cdev,
	struct device_attribute *attr,char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
	struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;
	char *str = buf;
        void *virt_addr;
	int Index;
	DECLARE_COMPLETION_ONSTACK(completion);
	struct pm8001_ioctl_payload payload;

	pm8001_ha->nvmd_completion = &completion;
	payload.minor_function = 0;
	payload.offset = WWNOFFSET; 
	payload.length = 8;
	payload.func_specific = kzalloc(8, GFP_KERNEL);
	PM8001_CHIP_DISP->get_nvmd_req(pm8001_ha, &payload);
	wait_for_completion(&completion);
	virt_addr = pm8001_ha->memoryMap.region[NVMD].virt_ptr;
        for(Index=0;Index<8;Index++)
        {
		str += sprintf(str,"%.2x",*((u8 *)(virt_addr+Index)));
        }
	return str - buf;
}
static DEVICE_ATTR(ctrl_wwn,S_IRUGO,pm8001_ctl_ctrl_wwn_show,NULL);

 /* pm8001_ctl_ctrl_serialno_show - serial no Display
 * @cdev:pointer to embedded class device
 * @buf:the buffer returned
 * A sysfs 'read-only' shost attribute.
 */
static ssize_t pm8001_ctl_ctrl_serialno_show(struct device *cdev,
	struct device_attribute *attr,char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
	struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;
	char *str = buf;
	void *virt_addr;
	int Index;
	DECLARE_COMPLETION_ONSTACK(completion);
	struct pm8001_ioctl_payload payload;

	pm8001_ha->nvmd_completion = &completion;
	payload.minor_function = 0;
	payload.offset = SERIALNO_OFFSET; 
	payload.length = 11;
	payload.func_specific = kzalloc(11, GFP_KERNEL);
	PM8001_CHIP_DISP->get_nvmd_req(pm8001_ha, &payload);
	wait_for_completion(&completion);
	virt_addr = pm8001_ha->memoryMap.region[NVMD].virt_ptr;
        for(Index=0;Index<11;Index++)
        {
		str += sprintf(str,"%c",*((u8 *)(virt_addr+Index)));
        }
	return str - buf;
}
static DEVICE_ATTR(ctrl_serialno,S_IRUGO,pm8001_ctl_ctrl_serialno_show,NULL);
/**
 * pm8001_ctl_seeprom_read - Read seeprom file
 * @cdev:pointer to embedded class device
 * @buf:the buffer returned
 * A sysfs 'read-only' shost attribute.
 */
static ssize_t pm8001_ctl_seeprom_read(struct device *cdev,
	struct device_attribute *attr,char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
	struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;
	char *str = buf;
	void *virt_addr=NULL;
	int ret=0;
	DECLARE_COMPLETION_ONSTACK(completion);
	struct pm8001_ioctl_payload payload;

	pm8001_ha->nvmd_completion = &completion;
	payload.minor_function = C_SEEPROM;
	payload.offset = 0; 
	payload.length = 4096;
	payload.func_specific = kzalloc(4096, GFP_KERNEL);
	ret = PM8001_CHIP_DISP->get_nvmd_req(pm8001_ha, &payload);
	wait_for_completion(&completion);
	virt_addr = pm8001_ha->memoryMap.region[NVMD].virt_ptr;

	memcpy( (u8*)str,(u8 *)virt_addr, 4096);
	return 4095;	        

}
static DEVICE_ATTR(seeprom_read,S_IRUGO,pm8001_ctl_seeprom_read,NULL);

/**
 * pm8001_ctl_edit_wwn - write seeprom_file
 * @cdev:pointer to embedded class device
 * @buf:the buffer returned
 */
static ssize_t pm8001_ctl_edit_wwn(struct device *cdev,
        struct device_attribute *attr,char *buf, size_t count)
{
        struct Scsi_Host *shost = class_to_shost(cdev);
        struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
        struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;
        int length,i,j;
        u8 *ioctlbuffer=NULL;
        DECLARE_COMPLETION_ONSTACK(completion);
        struct pm8001_ioctl_payload *payload;
	unsigned char *sas_addr;
	unsigned char sas_addr_char[2] = { '\0', '\0' };
	u8 addr[8];

	sas_addr = buf;
 
	for (i = 0, j = 0; i < 16; i += 2, j++) {
		sas_addr_char[0] = sas_addr[i];
		sas_addr[i] = (u8) simple_strtoul(sas_addr_char, 0, 16);
		sas_addr_char[0] = (u8) sas_addr[i + 1];
		sas_addr[i + 1] = simple_strtoul(sas_addr_char, 0, 16);
		addr[j] = sas_addr[i] << 4 | sas_addr[i + 1];
        }
	sas_addr = addr;
 
	length = 8 + sizeof(*payload);
        ioctlbuffer = kzalloc(length, GFP_KERNEL);
        if (!ioctlbuffer)
                return -ENOMEM;
        payload = (struct pm8001_ioctl_payload *)ioctlbuffer;
        pm8001_ha->nvmd_completion = &completion;
        payload->minor_function = 0;
        payload->offset = WWNOFFSET; 
        payload->length = 8;
        memcpy((u8*)&payload->func_specific, (u8*)sas_addr, 8);

	PM8001_CHIP_DISP->set_nvmd_req(pm8001_ha, payload);
        wait_for_completion(&completion);
	if (ioctlbuffer != NULL)
		kfree(ioctlbuffer);
        return count;
}

static DEVICE_ATTR(edit_wwn,S_IRUGO|S_IWUSR|S_IWGRP,NULL,pm8001_ctl_edit_wwn);

/**
 * pm8001_ctl_edit_serialno - write seeprom file
 * @cdev:pointer to embedded class device
 * @buf:the buffer returned
 */
static ssize_t pm8001_ctl_edit_serialno(struct device *cdev,
        struct device_attribute *attr,char *buf, size_t count)
{
        struct Scsi_Host *shost = class_to_shost(cdev);
        struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
        struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;
        u8 *ioctlbuffer=NULL, *data = NULL;
	int i,length;
	char *str=buf;
        DECLARE_COMPLETION_ONSTACK(completion);
        struct pm8001_ioctl_payload *payload;

	data = (u8*)kmalloc(11,GFP_KERNEL);
	memset(data, 0, 11);
	for(i=0;i<count-1;i++)
	{
		sprintf(data+i,"%c",(u8)str[i]);
	}
	length = 11 + sizeof(*payload);
        ioctlbuffer = kzalloc(length, GFP_KERNEL);
        if (!ioctlbuffer)
                return -ENOMEM;
        payload = (struct pm8001_ioctl_payload *)ioctlbuffer;
        pm8001_ha->nvmd_completion = &completion;
        payload->minor_function = 0;
        payload->offset = SERIALNO_OFFSET; 
        payload->length = 11;
        memcpy((u8*)&payload->func_specific, (u8*)data, 11);

	PM8001_CHIP_DISP->set_nvmd_req(pm8001_ha, payload);
        wait_for_completion(&completion);
	if(data!=NULL)
		kfree(data);
	if (ioctlbuffer != NULL)
		kfree(ioctlbuffer);
        return count;
}

static DEVICE_ATTR(edit_serialno ,S_IRUGO|S_IWUSR|S_IWGRP,NULL,pm8001_ctl_edit_serialno);


/**
 * pm8001_ctl_ib_queue_log_show - Out bound Queue log
 * @cdev:pointer to embedded class device
 * @buf: the buffer returned
 *
 * A sysfs 'read-only' shost attribute.
 */

static ssize_t pm8001_ctl_ib_queue_log_show(struct device *cdev,
        struct device_attribute *attr, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
	struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;
	int offset;
	char *str = buf;
	int start=0;
 
	if(pm8001_ha->chip_id!=chip_8001){ //spcv
#define IB_MEMMAP(c) \
	(*(u32 *)((u8*)pm8001_ha->memoryMap.region[IB].virt_ptr + pm8001_ha->evtlog_ib_offset + (c)))
	}
	else { //spc
#define IB_MEMMAP_SPC(c) \
	(*(u32 *)((u8*)pm8001_ha->memoryMap.region[IB].virt_ptr  + pm8001_ha->evtlog_ib_offset + (c)))
	}
	for(offset=0;offset<IB_OB_READ_TIMES;offset++)
	{
	  if(pm8001_ha->chip_id!=chip_8001){ //spcv
		str +=sprintf(str,"0x%08x\n",IB_MEMMAP(start));
          }
          else{
                str +=sprintf(str,"0x%08x\n",IB_MEMMAP_SPC(start));
          }
	  start = start + 4;
	}
        pm8001_ha->evtlog_ib_offset += SYSFS_OFFSET;
	if((((pm8001_ha->evtlog_ib_offset)%(PM80XX_IB_OB_QUEUE_SIZE))==0)&&(pm8001_ha->chip_id!=chip_8001)) {
	 pm8001_ha->evtlog_ib_offset = 0;
	}
	if((((pm8001_ha->evtlog_ib_offset)%(PM8001_IB_OB_QUEUE_SIZE))==0)&&(pm8001_ha->chip_id==chip_8001)) {
	 pm8001_ha->evtlog_ib_offset = 0;
	}

	return str - buf;
}
static DEVICE_ATTR(ib_log, S_IRUGO, pm8001_ctl_ib_queue_log_show, NULL);
/**
 * pm8001_ctl_ob_queue_log_show - Out bound Queue log
 * @cdev:pointer to embedded class device
 * @buf: the buffer returned
 *
 * A sysfs 'read-only' shost attribute.
 */

static ssize_t pm8001_ctl_ob_queue_log_show(struct device *cdev,
        struct device_attribute *attr, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
	struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;
	int offset;
	char *str = buf;
	int start=0;
	

	if(pm8001_ha->chip_id!=chip_8001){ //spcv
#define OB_MEMMAP(c) \
	(*(u32 *)((u8*)pm8001_ha->memoryMap.region[OB].virt_ptr + pm8001_ha->evtlog_ob_offset + (c)))
	}
	else { //spc
#define OB_MEMMAP_SPC(c) \
	(*(u32 *)((u8*)pm8001_ha->memoryMap.region[OB].virt_ptr  +pm8001_ha->evtlog_ob_offset + (c)))
	}
	for(offset=0;offset<IB_OB_READ_TIMES;offset++)
	{
             if(pm8001_ha->chip_id!=chip_8001){ //spcv
		str +=sprintf(str,"0x%08x\n",OB_MEMMAP(start));
             }
             else{
               str +=sprintf(str,"0x%08x\n",OB_MEMMAP_SPC(start));
             }
	     start =start + 4;
	}
        pm8001_ha->evtlog_ob_offset += SYSFS_OFFSET;
	if((((pm8001_ha->evtlog_ob_offset)%(PM80XX_IB_OB_QUEUE_SIZE))==0)&&(pm8001_ha->chip_id!=chip_8001)) {
	 pm8001_ha->evtlog_ob_offset = 0;
	}
	if((((pm8001_ha->evtlog_ob_offset)%(PM8001_IB_OB_QUEUE_SIZE))==0)&&(pm8001_ha->chip_id==chip_8001)) {
	 pm8001_ha->evtlog_ob_offset = 0;
	}

	return str - buf;
}
static DEVICE_ATTR(ob_log, S_IRUGO, pm8001_ctl_ob_queue_log_show, NULL);
/**
 * pm8001_ctl_aap_log_show - IOP event log
 * @cdev: pointer to embedded class device
 * @buf: the buffer returned
 *
 * A sysfs 'read-only' shost attribute.
 */
static ssize_t pm8001_ctl_iop_log_show(struct device *cdev,
	struct device_attribute *attr, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
	struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;
	char *str = buf;
#if 0 /* original implementation */
#define IOP_MEMMAP(r, c) \
	(*(u32 *)((u8*)pm8001_ha->memoryMap.region[IOP].virt_ptr + (r) * 32 \
	+ (c)))
	int i;
	char *str = buf;
	int max = 2;
	for (i = 0; i < max; i++) {
		str += sprintf(str, "0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x"
			       "0x%08x 0x%08x\n",
			       IOP_MEMMAP(i, 0),
			       IOP_MEMMAP(i, 4),
			       IOP_MEMMAP(i, 8),
			       IOP_MEMMAP(i, 12),
			       IOP_MEMMAP(i, 16),
			       IOP_MEMMAP(i, 20),
			       IOP_MEMMAP(i, 24),
			       IOP_MEMMAP(i, 28));
	}
#endif

#if 1	
/* event logging to 128k - starts */
{
	static u32 start, end, count;
	u32 maxreadtimes = 32;		//change this value based on one time read size
	// 256Bytes = 4bytes * 64 times
	// 1K = 4bytes * 256 times
	// 4K = 4bytes * 1024 times
	// 8K = 4bytes * 2048 times
	// 16K = 4bytes * 4096 times
	// 32K = 4bytes * 8192 times
	// 128K = 4bytes x 32768 times
	// 4 bytes bcoz of u32 for temp
        u32 maxcount = (128 * 1024) / (maxreadtimes * 4);
	u32 *temp = (u32 *)pm8001_ha->memoryMap.region[IOP].virt_ptr;
	
	if((count%maxcount) == 0) {
		start=0;
		end=maxreadtimes;
		count=0;
	}
	else {
		start=end;
		end=end+maxreadtimes;
	}
	PM8001_EVENT_DBG(pm8001_ha, printk("\n[read from %d to %d loop count %d]\n\n", start, end, count));	
	
	for(; start < end; start++) 	
	{
		str += sprintf(str, "%08x ",*(temp+start));
	}
	count++;
	PM8001_EVENT_DBG(pm8001_ha, printk("\n\n"));
}
/* event logging - ends */
#endif
	return str - buf;
}
static DEVICE_ATTR(iop_log, S_IRUGO, pm8001_ctl_iop_log_show, NULL);

/**
 ** pm8001_ctl_fatal_log_show - fatal error logging
 ** @cdev:pointer to embedded class device
 ** @buf: the buffer returned
 **
 ** A sysfs 'read-only' shost attribute.
 **/

static ssize_t pm8001_ctl_fatal_log_show(struct device *cdev,
	struct device_attribute *attr, char *buf)
{
    u32 count;

    count = pm80xx_get_fatal_dump(cdev, attr, buf);
    return count;
}

static DEVICE_ATTR(fatal_log, S_IRUGO, pm8001_ctl_fatal_log_show, NULL);


/**
 ** pm8001_ctl_gsm_log_show - gsm dump collection
 ** @cdev:pointer to embedded class device
 ** @buf: the buffer returned
 **A sysfs 'read-only' shost attribute.
 **/
static ssize_t pm8001_ctl_gsm_log_show(struct device *cdev,
	struct device_attribute *attr, char *buf)
{
    u32 count;

    count = pm8001_get_gsm_dump(cdev, SYSFS_OFFSET, buf);
    return count;
}

static DEVICE_ATTR(gsm_log, S_IRUGO, pm8001_ctl_gsm_log_show, NULL);

#define FLASH_CMD_NONE      0x00
#define FLASH_CMD_UPDATE    0x01
#define FLASH_CMD_SET_NVMD    0x02

struct flash_command {
     u8      command[8];
     int     code;
};

static struct flash_command flash_command_table[] =
{
     {"set_nvmd",    FLASH_CMD_SET_NVMD},
     {"update",      FLASH_CMD_UPDATE},
     {"",            FLASH_CMD_NONE} /* Last entry should be NULL. */
};

struct error_fw {
     char    *reason;
     int     err_code;
};

static struct error_fw flash_error_table[] =
{
     {"Failed to open fw image file",	FAIL_OPEN_BIOS_FILE},
     {"image header mismatch",		FLASH_UPDATE_HDR_ERR},
     {"image offset mismatch",		FLASH_UPDATE_OFFSET_ERR},
     {"image CRC Error",		FLASH_UPDATE_CRC_ERR},
     {"image length Error.",		FLASH_UPDATE_LENGTH_ERR},
     {"Failed to program flash chip",	FLASH_UPDATE_HW_ERR},
     {"Flash chip not supported.",	FLASH_UPDATE_DNLD_NOT_SUPPORTED},
     {"Flash update disabled.",		FLASH_UPDATE_DISABLED},
     {"Flash in progress",		FLASH_IN_PROGRESS},
     {"Image file size Error",		FAIL_FILE_SIZE},
     {"Input parameter error",		FAIL_PARAMETERS},
     {"Out of memory",			FAIL_OUT_MEMORY},
     {"OK", 0}	/* Last entry err_code = 0. */
};

static int pm8001_set_nvmd(struct pm8001_hba_info *pm8001_ha)
{
	struct pm8001_ioctl_payload	*payload;
	DECLARE_COMPLETION_ONSTACK(completion);
	u8		*ioctlbuffer = NULL;
	u32		length = 0;
	u32		ret = 0;

	length = 1024 * 5 + sizeof(*payload) - 1;
	ioctlbuffer = kzalloc(length, GFP_KERNEL);
	if (!ioctlbuffer)
		return -ENOMEM;
	if ((pm8001_ha->fw_image->size <= 0) ||
	    (pm8001_ha->fw_image->size > 4096)) {
		ret = FAIL_FILE_SIZE;
		goto out;
	}
	payload = (struct pm8001_ioctl_payload *)ioctlbuffer;
	memcpy((u8 *)&payload->func_specific, (u8 *)pm8001_ha->fw_image->data,
				pm8001_ha->fw_image->size);		/* updated addr of func_specific */
	payload->length = pm8001_ha->fw_image->size;
	payload->id = 0;
	payload->minor_function = 0x1;
	pm8001_ha->nvmd_completion = &completion;

	scsi_block_requests(pm8001_ha->shost);
	ret = PM8001_CHIP_DISP->set_nvmd_req(pm8001_ha, payload);
	wait_for_completion(&completion);
	scsi_unblock_requests(pm8001_ha->shost);
out:
	kfree(ioctlbuffer);
	return ret;
}

static int pm8001_update_flash(struct pm8001_hba_info *pm8001_ha)
{
	struct pm8001_ioctl_payload	*payload;
	DECLARE_COMPLETION_ONSTACK(completion);
	u8		*ioctlbuffer = NULL;
	u32		length = 0;
	struct fw_control_info	*fwControl;
	u32		loopNumber, loopcount = 0;
	u32		sizeRead = 0;
	u32		partitionSize, partitionSizeTmp;
	u32		ret = 0;
	u32		partitionNumber = 0;
	struct pm8001_fw_image_header *image_hdr;

	length = 1024 * 16 + sizeof(*payload) - 1;
	ioctlbuffer = kzalloc(length, GFP_KERNEL);
	image_hdr = (struct pm8001_fw_image_header *)pm8001_ha->fw_image->data;
	if (!ioctlbuffer)
		return -ENOMEM;
	if (pm8001_ha->fw_image->size < 28) {
		ret = FAIL_FILE_SIZE;
		goto out;
	}

	while (sizeRead < pm8001_ha->fw_image->size) {
		partitionSizeTmp =
			*(u32 *)((u8 *)&image_hdr->image_length + sizeRead);
		partitionSize = be32_to_cpu(partitionSizeTmp);
		loopcount = (partitionSize + HEADER_LEN)/IOCTL_BUF_SIZE;
		if (loopcount % IOCTL_BUF_SIZE)
			loopcount++;
		if (loopcount == 0)
			loopcount++;
		for (loopNumber = 0; loopNumber < loopcount; loopNumber++) {
			payload = (struct pm8001_ioctl_payload *)ioctlbuffer;
			payload->length = 1024*16;
			payload->id = 0;
			fwControl =
			      (struct fw_control_info *)&payload->func_specific;
			fwControl->len = IOCTL_BUF_SIZE;   /* IN */
			fwControl->size = partitionSize + HEADER_LEN;/* IN */
			fwControl->retcode = 0;/* OUT */
			fwControl->offset = loopNumber * IOCTL_BUF_SIZE;/*OUT */

		/* for the last chunk of data in case file size is not even with
		4k, load only the rest*/
		if (((loopcount-loopNumber) == 1) &&
			((partitionSize + HEADER_LEN) % IOCTL_BUF_SIZE)) {
			fwControl->len =
				(partitionSize + HEADER_LEN) % IOCTL_BUF_SIZE;
			memcpy((u8 *)fwControl->buffer,
				(u8 *)pm8001_ha->fw_image->data + sizeRead,
				(partitionSize + HEADER_LEN) % IOCTL_BUF_SIZE);
			sizeRead +=
				(partitionSize + HEADER_LEN) % IOCTL_BUF_SIZE;
		} else {
			memcpy((u8 *)fwControl->buffer,
				(u8 *)pm8001_ha->fw_image->data + sizeRead,
				IOCTL_BUF_SIZE);
			sizeRead += IOCTL_BUF_SIZE;
		}
		
		pm8001_ha->nvmd_completion = &completion;
		scsi_block_requests(pm8001_ha->shost);
		ret = PM8001_CHIP_DISP->fw_flash_update_req(pm8001_ha, payload);
		wait_for_completion(&completion);
           scsi_unblock_requests(pm8001_ha->shost);
		if (ret || (fwControl->retcode > FLASH_UPDATE_IN_PROGRESS)) {
			ret = fwControl->retcode;
			kfree(ioctlbuffer);
			ioctlbuffer = NULL;
			break;
		}
	}
	if (ret)
		break;
	partitionNumber++;
}
out:
	kfree(ioctlbuffer);
	return ret;
}
static ssize_t pm8001_store_update_fw(struct device *cdev,
				      struct device_attribute *attr,
				      const char *buf, size_t count)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
	struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;
	char *cmd_ptr, *filename_ptr;
	int res, i;
	int flash_command = FLASH_CMD_NONE;
	int err = 0;
	if (!capable(CAP_SYS_ADMIN))
		return -EACCES;

	cmd_ptr = kzalloc(count*2, GFP_KERNEL);

	if (!cmd_ptr) {
		err = FAIL_OUT_MEMORY;
		goto out;
	}

	filename_ptr = cmd_ptr + count;
	res = sscanf(buf, "%s %s", cmd_ptr, filename_ptr);
	if (res != 2) {
		err = FAIL_PARAMETERS;
		goto out1;
	}

	for (i = 0; flash_command_table[i].code != FLASH_CMD_NONE; i++) {
		if (!memcmp(flash_command_table[i].command,
				 cmd_ptr, strlen(cmd_ptr))) {
			flash_command = flash_command_table[i].code;
			break;
		}
	}
	if (flash_command == FLASH_CMD_NONE) {
		err = FAIL_PARAMETERS;
		goto out1;
	}

	if (pm8001_ha->fw_status == FLASH_IN_PROGRESS) {
		err = FLASH_IN_PROGRESS;
		goto out1;
	}
	err = request_firmware(&pm8001_ha->fw_image,
			       filename_ptr,
			       pm8001_ha->dev);

	if (err) {
		PM8001_FAIL_DBG(pm8001_ha,
			pm8001_printk("Failed to load firmware image file %s,"
			" error %d\n", filename_ptr, err));
		err = FAIL_OPEN_BIOS_FILE;
		goto out1;
	}

	switch (flash_command) {
	case FLASH_CMD_UPDATE:
		pm8001_ha->fw_status = FLASH_IN_PROGRESS;
		err = pm8001_update_flash(pm8001_ha);
		break;
	case FLASH_CMD_SET_NVMD:
		pm8001_ha->fw_status = FLASH_IN_PROGRESS;
		err = pm8001_set_nvmd(pm8001_ha);
		break;
	default:
		pm8001_ha->fw_status = FAIL_PARAMETERS;
		err = FAIL_PARAMETERS;
		break;
	}
	release_firmware(pm8001_ha->fw_image);
out1:
	kfree(cmd_ptr);
out:
	pm8001_ha->fw_status = err;

	if (!err)
		return count;
	else
		return -err;
}

static ssize_t pm8001_show_update_fw(struct device *cdev,
				     struct device_attribute *attr, char *buf)
{
	int i;
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
	struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;

	for (i = 0; flash_error_table[i].err_code != 0; i++) {
		if (flash_error_table[i].err_code == pm8001_ha->fw_status)
			break;
	}
	if (pm8001_ha->fw_status != FLASH_IN_PROGRESS)
		pm8001_ha->fw_status = FLASH_OK;

	return snprintf(buf, PAGE_SIZE, "status=%x %s\n",
			flash_error_table[i].err_code,
			flash_error_table[i].reason);
}
static DEVICE_ATTR(update_fw, S_IRUGO|S_IWUSR|S_IWGRP,
	pm8001_show_update_fw, pm8001_store_update_fw);

/**
* pm8001_ctl_num_phys_show - Number of phys
* @cdev:pointer to embedded class device
* @buf:the buffer returned
* A sysfs 'read-only' shost attribute.
*/
static ssize_t pm8001_ctl_num_phys_show(struct device *cdev,
        struct device_attribute *attr,char *buf)
{
	int ret;
	struct Scsi_Host *shost = class_to_shost(cdev);
	struct sas_ha_struct *sha = SHOST_TO_SAS_HA(shost);
	struct pm8001_hba_info *pm8001_ha = sha->lldd_ha;
	ret = sprintf(buf,"%d",pm8001_ha->chip->n_phy);
	return ret;
}
static DEVICE_ATTR(num_phys,S_IRUGO,pm8001_ctl_num_phys_show,NULL);

struct device_attribute *pm8001_host_attrs[] = {
	&dev_attr_interface_rev,
	&dev_attr_fw_version,
	&dev_attr_update_fw,
	&dev_attr_aap_log,
	&dev_attr_iop_log,
	&dev_attr_fatal_log,
	&dev_attr_gsm_log,
	&dev_attr_max_out_io,
	&dev_attr_max_devices,
	&dev_attr_max_sg_list,
	&dev_attr_sas_spec_support,
	&dev_attr_logging_level,
	&dev_attr_host_sas_address,
	&dev_attr_ib_log,
	&dev_attr_ob_log,
	&dev_attr_bios_version,
	&dev_attr_num_phys,
	&dev_attr_seeprom_read,
	&dev_attr_edit_wwn,
	&dev_attr_edit_serialno,
	&dev_attr_ctrl_wwn,
	&dev_attr_ctrl_serialno,
	NULL,
};

/**
 *	pm8001_open - open the configuration file
 *	@inode: inode being opened
 *	@file: file handle attached
 *
 *	Called when the configuration device is opened. Does the needed
 *	set up on the handle and then returns
 *
 */

static int pm8001_open(struct inode *inode, struct file *file)
{
	struct pm8001_hba_info *pm8001_ha;
	unsigned minor_number = iminor(inode);
	int err = -ENODEV;

	list_for_each_entry(pm8001_ha, &hba_list, list) {
        /* pm8001_ha->id  is the major number argument when register Beechnut as a MISC driver.
         * When Beechnut driver is registered as MISC driver,
         * the input of major number argument is zero.
         * Then Linux kernel will assign MISC driver
         * a fixed major number in 10 and a dynamic minor number, for now set to 254.*/
		PM8001_MSG_DBG(pm8001_ha,printk(KERN_ERR "reached: pm8001_open:id=%d\n", pm8001_ha->id)); 
		if  ((pm8001_ha->id == minor_number) || (pm8001_ha->id == 0))  {

			file->private_data = pm8001_ha;
			err = 0;
			break;
		}
	}

	return err;
}

/**
 *      pm8001_close - close the configuration file
 *      @inode: inode being opened
 *      @file: file handle attached
 *
 *      Called when the configuration device is closed. Does the needed
 *      set up on the handle and then returns
 *
 */

static int pm8001_close(struct inode *inode, struct file *file)
{
	return 0;
}

int pm8001_phy_rescan(struct pm8001_hba_info *pm8001_ha)
{
        int i;
        DECLARE_COMPLETION(comp);
        pm8001_ha->scan_completion = &comp;
        pm8001_ha->rescan_host.scan_status = 1;
        pm8001_ha->rescan_host.phy_start_index = 0;
        pm8001_ha->rescan_host.phys_count = pm8001_ha->chip->n_phy;

        for (i = pm8001_ha->rescan_host.phy_start_index; i < (pm8001_ha->rescan_host.phy_start_index+pm8001_ha->rescan_host.phys_count); ++i)
                PM8001_CHIP_DISP->phy_start_req(pm8001_ha, i);
        wait_for_completion(&comp);
        PM8001_INIT_DBG(pm8001_ha,pm8001_printk("PHY rescan completed for %d PHYs starting from phy %d\n", pm8001_ha->rescan_host.phys_count, pm8001_ha->rescan_host.phy_start_index));
        memset(&pm8001_ha->rescan_host, 0, sizeof(struct ioctl_scan_host));
        return 0;
}

int pm8001_ioctl_soft_reset(struct pm8001_hba_info *pm8001_ha, unsigned long arg)
{
	int i;
#ifndef DIAG_MODE
	int type;
	struct domain_device *dev,*parent_dev;
	struct pm8001_device *pm8001_dev;
	u32 device_id;
	pm8001_ha->reset_in_progress = 1;
    /*Block the request to this host */
	scsi_block_requests(pm8001_ha->shost);
	if(pm8001_ha->chip_id == chip_8001){
		PM8001_CHIP_DISP->chip_rst(pm8001_ha);
		msleep(5000);	
    		pm8001_ha->reset_in_progress = 0;
    		/* Reset finished, Unblock requests */			
    		scsi_unblock_requests(pm8001_ha->shost);
		return 0;
	}

	/* Deregsiter all the device ids  */
	for(i=0;i<PM8001_MAX_DEVICES;i++) {
		#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0) )
		if(pm8001_ha->devices[i].dev_type != NO_DEVICE) {
		#else
		if(pm8001_ha->devices[i].dev_type != SAS_PHY_UNUSED) {	
		#endif	
			dev = pm8001_ha->devices[i].sas_device;
			pm8001_dev = dev->lldd_dev;
			device_id = pm8001_dev->device_id;
			if(device_id)
				PM8001_CHIP_DISP->dereg_dev_req(pm8001_ha, device_id);
            }       
    }
	msleep(3000);
#else
	pm8001_ha->reset_in_progress = 1;
#endif
	/*Invoke controller reset */ 
	PM8001_CHIP_DISP->chip_soft_rst(pm8001_ha);
	/*MPI Initialize */ 
	PM8001_CHIP_DISP->chip_init(pm8001_ha);
	mdelay(1000);
	/* Enable interrupts */
	PM8001_CHIP_DISP->interrupt_enable(pm8001_ha, 0);
#ifdef PM8001_USE_MSIX
	for (i = 1;i < pm8001_ha->number_of_intr;i++)
		PM8001_CHIP_DISP->interrupt_enable(pm8001_ha, i);
#endif
	mdelay(1000);
	pm8001_ha->reset_in_progress = 0;
	/* Start phys */
	pm8001_phy_rescan(pm8001_ha);
	mdelay(2000);
#ifndef DIAG_MODE
	/* Re register devices */
	for(i=0;i<PM8001_MAX_DEVICES;i++) {
	#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0) )
		if(pm8001_ha->devices[i].dev_type != NO_DEVICE) {
	#else
		if(pm8001_ha->devices[i].dev_type != SAS_PHY_UNUSED) {
	#endif
			dev = pm8001_ha->devices[i].sas_device;
			pm8001_dev = dev->lldd_dev;
			dev = pm8001_ha->devices[i].sas_device;
				parent_dev = dev->parent;
			if(parent_dev && DEV_IS_EXPANDER(parent_dev->dev_type)) {
				type = 0;   
			}
			else{
				#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0) )
					type = (dev->dev_type  == SATA_DEV) ? 1 : 0;  //Direct SATA
				#else
					type = (dev->dev_type  == SAS_SATA_DEV) ? 1 : 0;  //Direct SATA
				#endif
			} 
			device_id = pm8001_dev->device_id;
			PM8001_CHIP_DISP->reg_dev_req(pm8001_ha, pm8001_dev,type);
			msleep(1000);
		}
	}
		
	msleep(9000);
	pm8001_ha->reset_in_progress = 0;
	/* Reset finished, Unblock requests */			
	scsi_unblock_requests(pm8001_ha->shost);
#endif
	return 0;
}

static int pm8001_ioctl_get_phy_profile(struct pm8001_hba_info *pm8001_ha,
                                                unsigned long arg)
{
	struct phy_profile phy_prof[16];
	int nphys;
	DECLARE_COMPLETION_ONSTACK(completion);
        unsigned long timeout = 2000;
        u32 ret = 0, i;
	int page_code = SAS_PHY_GENERAL_STATUS_PAGE;

	if(pm8001_ha->pdev->device == 0x8001 || pm8001_ha->pdev->device == 0x8081){
		return ADPT_IOCTL_CALL_INVALID_DEVICE;
	}
	
        mutex_lock(&pm8001_ha->ioctl_mutex);

	mod_timer( &pm8001_ha->ioctl_timer, jiffies + msecs_to_jiffies(timeout) );
	
	if (copy_from_user(&phy_prof[0], (struct phy_profile *)arg,
                sizeof(struct phy_profile)))
        {
                return ADPT_IOCTL_CALL_FAILED;
        }
	
	nphys = phy_prof[0].phy_id;
	if (nphys == -1) {
		for(i = 0;i < pm8001_ha->chip->n_phy;i++)
		{
        		pm8001_ha->ioctl_completion = &completion;
			ret = PM8001_CHIP_DISP->get_phy_profile_req(pm8001_ha,
								i, page_code);
			if(ret != 0) {
				ret = ADPT_IOCTL_CALL_FAILED;
			}
			wait_for_completion(&completion);
			memcpy((void *)&phy_prof[i], (void *)&pm8001_ha->phy_profile_resp,
						sizeof(struct phy_profile));
		}
			
		if (copy_to_user((void *)arg, (void *)&phy_prof,
				sizeof(struct phy_profile) * (i+1) )) {
                	ret = ADPT_IOCTL_CALL_FAILED;
        	}
	}
	else {
        	pm8001_ha->ioctl_completion = &completion;
		ret = PM8001_CHIP_DISP->get_phy_profile_req(pm8001_ha, 
							nphys, page_code);
		if(ret != 0) {
			ret = ADPT_IOCTL_CALL_FAILED;
			goto exit;
		}
		wait_for_completion(&completion);
		if (copy_to_user((void *)arg, (void *)&pm8001_ha->phy_profile_resp,
				sizeof(struct phy_profile))) {
                	ret = ADPT_IOCTL_CALL_FAILED;
        	}
	}
		
exit:
	mutex_unlock(&pm8001_ha->ioctl_mutex);
	return ret;
	
}

static int pm8001_ioctl_get_phy_err(struct pm8001_hba_info *pm8001_ha,
                                                unsigned long arg)
{
	struct phy_errcnt phy_err[16];
	DECLARE_COMPLETION_ONSTACK(completion);
        unsigned long timeout = 2000;
        u32 ret = 0, i;
	int page_code = SAS_PHY_ERR_COUNTERS_PAGE;
	/*6H card does not support phyerr*/	
	if(pm8001_ha->pdev->device == 0x8001 || pm8001_ha->pdev->device == 0x8081){
		return ADPT_IOCTL_CALL_INVALID_DEVICE;
	}

        mutex_lock(&pm8001_ha->ioctl_mutex);

	mod_timer( &pm8001_ha->ioctl_timer, jiffies + msecs_to_jiffies(timeout) );
	

	
		for(i = 0;i < pm8001_ha->chip->n_phy;i++)
		{
        		pm8001_ha->ioctl_completion = &completion;
			ret = PM8001_CHIP_DISP->get_phy_profile_req(pm8001_ha,
								i, page_code);
			if(ret != 0) {
				ret = ADPT_IOCTL_CALL_FAILED;
			}
			wait_for_completion(&completion);
			memcpy((void *)&phy_err[i], (void *)&pm8001_ha->phy_profile_resp,
						sizeof(struct phy_errcnt));
		}
			
		if (copy_to_user((void *)arg, (void *)&phy_err,
				sizeof(struct phy_errcnt) * (i+1) )) {
                	ret = ADPT_IOCTL_CALL_FAILED;
        	}
	
	mutex_unlock(&pm8001_ha->ioctl_mutex);
	return ret;
	
}

static int pm8001_ioctl_list_devices(struct pm8001_hba_info *pm8001_ha,
                                                        unsigned long arg)
{
	u32 ret = 0,i = 0,offset =4,count =0,size =0;
	struct device_info devinfo;
	struct domain_device *dev;
	void *tmp;
	u8 cdb[16], *resp;
	u32 resp_len = 16, lun;

	size= PM8001_MAX_DEVICES * sizeof(struct device_info);
	tmp = kmalloc(size, GFP_KERNEL);
	memset(tmp, 0, size);
	memset(cdb, 0x00, 16);
	resp = kmalloc(16,GFP_KERNEL);
	memset(resp, 0x00, resp_len);

	for(i=0;i<PM8001_MAX_DEVICES;i++) {
		#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0) )
			if(pm8001_ha->devices[i].dev_type != NO_DEVICE) {
		#else
			if(pm8001_ha->devices[i].dev_type != SAS_PHY_UNUSED)  {
		#endif
			dev = pm8001_ha->devices[i].sas_device;	
			devinfo.iproto = dev->iproto;
			devinfo.tproto = dev->tproto;
			devinfo.luns = 0;
			/* To get number of luns from RBOD */
			if(dev->tproto == SAS_PROTOCOL_SSP) { 
				cdb[0] = SCSI_REPORT_LUN;
				cdb[8] = 0x10;/*resp_len*/
				PM8001_CHIP_DISP->ssp_util_req(pm8001_ha, dev, resp, (u8 *)&cdb, resp_len);
				lun = le32_to_cpu((resp[0] << 24) | (resp[1] << 16) | 
					(resp[2] << 8) | (resp[3]));
				lun = lun/8;
				devinfo.luns = lun;
			}
			if((dev->tproto == SAS_PROTOCOL_STP) ||
		       		(dev->tproto == SAS_PROTOCOL_SATA) ||
				   (dev->tproto == (SAS_PROTOCOL_STP | SAS_PROTOCOL_SATA)) ) {
				devinfo.luns = 1;/*SATA devices*/
			}
			devinfo.linkrate = dev->linkrate;
			devinfo.phy_id = sas_find_local_phy_id(dev);
			memcpy(devinfo.sas_address,dev->sas_addr,8);	
			memcpy(tmp+offset,(void *)&devinfo,sizeof(devinfo));
			offset += sizeof(devinfo);
			count ++;
		}
	}
		
	*((u32 *)tmp) = count ; //4 byte device count at offset 0
	if (copy_to_user((void *)arg, (void *)tmp, offset+1)) {
                ret = ADPT_IOCTL_CALL_FAILED;
        }

	kfree(tmp);
	kfree(resp);

	return ret;
}

static long pm8001_info_ioctl(struct pm8001_hba_info *pm8001_ha, unsigned long arg)
{
	u32 ret = 0;
	struct ioctl_info_buffer info_buf;

	strcpy(info_buf.information.sz_name, DRV_NAME);
					
	info_buf.information.usmajor_revision = DRV_MAJOR;
	info_buf.information.usminor_revision = DRV_MINOR;
	info_buf.information.usbuild_revision = DRV_BUILD;
	if (pm8001_ha->chip_id == chip_8001) {
		info_buf.information.maxoutstandingIO = 
			pm8001_ha->main_cfg_tbl.pm8001_tbl.max_out_io;
		info_buf.information.maxdevices =
			(pm8001_ha->main_cfg_tbl.pm8001_tbl.max_sgl >> 16) & 0xFFFF;
	}
	else {
		info_buf.information.maxoutstandingIO =
			pm8001_ha->main_cfg_tbl.pm80xx_tbl.max_out_io;
		info_buf.information.maxdevices =
			(pm8001_ha->main_cfg_tbl.pm80xx_tbl.max_sgl >> 16) & 0xFFFF;
	}
	info_buf.header.return_code = ADPT_IOCTL_CALL_SUCCESS;

	if (copy_to_user((void *)arg, (void *)&info_buf, sizeof(struct ioctl_info_buffer)))
	{
		ret = ADPT_IOCTL_CALL_FAILED;
	}
	return ret;
}

static long pm8001_gpio_ioctl(struct pm8001_hba_info *pm8001_ha, unsigned long arg)
{
	struct gpio_buffer buffer;
	struct pm8001_gpio *payload;
	struct gpio_ioctl_resp *gpio_resp;
	DECLARE_COMPLETION_ONSTACK(completion);
	unsigned long timeout;
	u32 ret = 0, operation;

	if(pm8001_ha->pdev->subsystem_vendor == PCI_VENDOR_ID_ADAPTEC2)
		return ADPT_IOCTL_CALL_INVALID_DEVICE;

	mutex_lock(&pm8001_ha->ioctl_mutex);

	if (copy_from_user(&buffer, (struct gpio_buffer *)arg,
		sizeof(struct gpio_buffer)))
	{
		ret = ADPT_IOCTL_CALL_FAILED;
		goto exit;
	}
	pm8001_ha->ioctl_completion = &completion;
	payload = &buffer.gpio_payload;
	operation = payload->operation;
	ret = PM8001_CHIP_DISP->gpio_req(pm8001_ha, payload);
	if(ret != 0) {
		ret = ADPT_IOCTL_CALL_FAILED;
		goto exit;
	}

	timeout = (unsigned long)buffer.header.timeout * 1000;
	if(timeout < 2000)
		timeout = 2000;

	ret = mod_timer( &pm8001_ha->ioctl_timer, jiffies + msecs_to_jiffies(timeout) );
	if (ret)
		PM8001_MSG_DBG(pm8001_ha, pm8001_printk("Error in mod_timer \n"));

	wait_for_completion(&completion);

	if(pm8001_ha->ioctl_timer_expired)
	{
		ret = ADPT_IOCTL_CALL_TIMEOUT;
		goto exit;
	}
	gpio_resp = &pm8001_ha->gpio_resp;

	buffer.header.return_code = ADPT_IOCTL_CALL_SUCCESS;

	if(operation == GPIO_READ)
	{
		payload->rd_wr_val		= gpio_resp->gpio_rd_val;
		payload->input_enable		= gpio_resp->gpio_in_enabled;
		payload->pinsetup1		= gpio_resp->gpio_pinsetup1;
		payload->pinsetup2		= gpio_resp->gpio_pinsetup2;
		payload->event_level		= gpio_resp->gpio_evt_change;
		payload->event_rising_edge	= gpio_resp->gpio_evt_rise;
		payload->event_falling_edge	= gpio_resp->gpio_evt_fall;

		if (copy_to_user((void *)arg, (void *)&buffer, sizeof(struct gpio_buffer)))
		{
			ret = ADPT_IOCTL_CALL_FAILED;
		}
	} 
	else
	{
		if (copy_to_user((void *)arg, (void *)&buffer.header, sizeof(struct ioctl_header)))
		{
			ret = ADPT_IOCTL_CALL_FAILED;
		}
	}

exit:
	pm8001_ha->ioctl_timer_expired = 0;
	mutex_unlock(&pm8001_ha->ioctl_mutex);
	return ret;
}

/**
 *	pm8001_sgpio_ioctl - pm8001 SGPIO request
 *	@pm8001_ha: HBA info structure
 *	@arg: argument from user
 *
 *	Handles SGPIO request from user.
 *
 */

static long pm8001_sgpio_ioctl(struct pm8001_hba_info *pm8001_ha, 
							unsigned long arg)
{
	struct sgpio_buffer buffer;
	struct read_write_req_resp *req = &buffer.sgpio_req;
	struct sgpio_req payload; 
	struct sgpio_ioctl_resp *sgpio_resp;
	DECLARE_COMPLETION_ONSTACK(completion);
	unsigned long timeout;
	u32 ret = 0, i;

	if (copy_from_user(&buffer, (struct sgpio_buffer *)arg,
		sizeof(struct sgpio_buffer)))
	{
		return ADPT_IOCTL_CALL_FAILED;
	}
	mutex_lock(&pm8001_ha->ioctl_mutex);
	pm8001_ha->ioctl_completion = &completion;

	payload.func_reg_index = cpu_to_le32((req->register_index << 24) |
		 (req->register_type << 16) | (req->function << 8) | SMP_FRAME_REQ);
	payload.count = req->register_count;

	if(req->function == WRITE_SGPIO_REGISTER)
	{
		if(req->register_count > MAX_SGPIO_REQ_PAYLOAD)
		{
			ret = ADPT_IOCTL_CALL_FAILED;
			goto exit;
		}
		for(i=0; i<req->register_count; i++)
			payload.value[i] = req->read_write_data[i];
	}

	ret = PM8001_CHIP_DISP->sgpio_req(pm8001_ha, &payload);
	if(ret != 0) {
		ret = ADPT_IOCTL_CALL_FAILED;
		goto exit;
	}
	if(timeout < 2000)
		timeout = 2000;

	mod_timer( &pm8001_ha->ioctl_timer, jiffies + msecs_to_jiffies(timeout) );

	wait_for_completion(&completion);
	pm8001_ha->ioctl_completion = NULL;

	if(pm8001_ha->ioctl_timer_expired)
	{
		ret = ADPT_IOCTL_CALL_TIMEOUT;
		goto exit;
	}

	sgpio_resp = &pm8001_ha->sgpio_resp;
	req->frame_type		= sgpio_resp->func_result & 0xff;
	req->function		= (sgpio_resp->func_result >> 8) & 0xff;
	req->function_result	= (sgpio_resp->func_result >> 16) & 0xff;
	if(req->function == READ_SGPIO_REGISTER)
	{
		for(i=0; i<req->register_count; i++)
		{
			req->read_write_data[i] = sgpio_resp->value[i];
		}
	}
	ret = ADPT_IOCTL_CALL_SUCCESS;	
exit:
	buffer.header.return_code = ret;
	if (copy_to_user((void *)arg, (void *)&buffer, sizeof(struct sgpio_buffer)))
	{
		ret = ADPT_IOCTL_CALL_FAILED;
	}
	
	mutex_unlock(&pm8001_ha->ioctl_mutex);
	
	return ret;
}

/**
 *	pm8001_ioctl - pm8001 configuration request
 *	@inode: inode of device
 *	@file: file handle
 *	@cmd: ioctl command code
 *	@arg: argument
 *
 *	Handles a configuration ioctl.
 *
 */

static long pm8001_ioctl(struct file *file,
		unsigned int cmd, unsigned long arg)
{
	u32 ret = -EACCES;
	struct pm8001_hba_info *pm8001_ha;
	struct ioctl_header header;
	u8* sas_addr;
	u8 temp;

	pm8001_ha = file->private_data;

	switch(cmd) {
		case ADPT_IOCTL_GPIO:
			ret = pm8001_gpio_ioctl(pm8001_ha, arg);
			break;

		case ADPT_IOCTL_INFO:
			ret = pm8001_info_ioctl(pm8001_ha, arg);
			break;

		case ADPT_IOCTL_SGPIO:
			ret = pm8001_sgpio_ioctl(pm8001_ha, arg);
			break;

		case ADPT_IOCTL_LIST_DEVICES:
			ret = pm8001_ioctl_list_devices(pm8001_ha, arg);
			break;
		
		case ADPT_IOCTL_GET_PHY_PROFILE:
			ret = pm8001_ioctl_get_phy_profile(pm8001_ha, arg);
			return ret;
	
		case ADPT_IOCTL_SOFT_RESET:
			ret = pm8001_ioctl_soft_reset(pm8001_ha, arg);
            /* Initialize tags */
            pm8001_tag_init(pm8001_ha);
			return ret;

		case ADPT_IOCTL_GET_PHY_ERR_CNT:
			ret = pm8001_ioctl_get_phy_err(pm8001_ha, arg);
			break;
 
	        case ADPT_COMMAND:
	        {
		        struct adpt_ioctl_command __user *uarg = (void __user *) arg;
                        struct adpt_ioctl_command  *karg;
                        pSmpPassthroughRequest_t smpReq;
                        int i;
                        u8* buffer;

                       buffer = kzalloc((sizeof(struct adpt_ioctl_command) + \
                                sizeof(MSG_SMP_PASSTHROUGH_REQUEST)), GFP_KERNEL);
                       karg = (struct adpt_ioctl_command*)buffer;

                       if (copy_from_user(karg, uarg, (sizeof(struct adpt_ioctl_command) \
                          + sizeof(MSG_SMP_PASSTHROUGH_REQUEST) - 1))) 
                       {
		         PM8001_FAIL_DBG(pm8001_ha, pm8001_printk("copy_to_user failed \n"));
                         return -1;
                       }

                       smpReq = (pSmpPassthroughRequest_t)karg->MF; 
                       sas_addr = (u8*)&smpReq->SASAddress;

                       if(sas_addr == NULL)
                       {
                         return -1;
                       }
                       for(i=0; i<(SAS_ADDR_SIZE/2); i++)
                       {
                         temp = sas_addr[i];
                         sas_addr[i] = sas_addr[SAS_ADDR_SIZE-i-1];
                         sas_addr[SAS_ADDR_SIZE-i-1] = temp;
                       }
                           
                       ret = PM8001_CHIP_DISP->smp_util_req(pm8001_ha, karg->dataOutSize + 4, karg->dataInSize - 4, \
		       karg->dataOutBufPtr, karg->dataInBufPtr, (u8*)&smpReq->SASAddress);
	               if(ret != 0)
		       {
		         return -1;
		       }
                       if (copy_to_user((void *)uarg->dataInBufPtr, (void *)karg->dataInBufPtr, karg->dataInSize))
		       {
		         return -1;
		       }
		       kfree(buffer);

                       ret=0;
		       break;
	  	}

		case IOCTL_SMP_CMND_PASSTHROUGH:
		{
			PM8001_IOCTL_DBG(pm8001_ha, printk(KERN_ERR "reached: IOCTL_SMP_CMND_PASSTHROUGH\n"));
			ret = pm8001_ctl_ioctl_smp(file, cmd, (void __user *)arg);
			break;
		}

		case ADP80XX_SMP_COMMAND:
		{
			struct smp_pass_through_req *uarg = (void *)arg;
			struct smp_pass_through_req *smp_req_resp;
			//printk(KERN_INFO "reached\n");


			smp_req_resp = (struct smp_pass_through_req *)kzalloc(sizeof(struct smp_pass_through_req), GFP_KERNEL);

                       	if (copy_from_user(smp_req_resp, uarg, (sizeof(struct smp_pass_through_req)))) 
                       	{
		        	PM8001_FAIL_DBG(pm8001_ha, pm8001_printk("copy_to_user failed \n"));
                        	return -1;
			}

			PM8001_IOCTL_DBG(pm8001_ha, printk(KERN_ERR "reqLen = %d respLen = %d\n", smp_req_resp->smp_req_len, smp_req_resp->smp_resp_len));
			for(ret=0;ret < smp_req_resp->smp_req_len; ret++)
			  PM8001_IOCTL_DBG(pm8001_ha, printk(KERN_ERR "%x ", smp_req_resp->smp_req[ret]));


			

                       	ret = PM8001_CHIP_DISP->smp_util_req(pm8001_ha, smp_req_resp->smp_req_len, smp_req_resp->smp_resp_len + 4, \
		       	smp_req_resp->smp_req, smp_req_resp->smp_resp, smp_req_resp->exp_sas_addr);

	               	if (ret != 0)
		       	{
		        	return -1;
		       	}
                       	if (copy_to_user((void *)uarg->smp_resp, (void *)smp_req_resp->smp_resp, smp_req_resp->smp_resp_len))
		       	{
		        	return -1;
		       	}
			kfree(smp_req_resp);
                       	ret=0;
		       	break;
			

			
		}
		case ADPT_IOCTL_SCAN:
                        ret = pm8001_phy_rescan(pm8001_ha);
                        break;

        	case ADPT_IOCTL_RESET_LUN:
		{
			struct tmf_ioctl *dev_info;
			struct tmf_ioctl *dev_info2 = (void *) arg;  
			PM8001_INIT_DBG(pm8001_ha, pm8001_printk("IOCTL lun reset request\n"));
			dev_info = kzalloc(sizeof(struct tmf_ioctl),GFP_KERNEL);
			if (copy_from_user(dev_info,dev_info2,sizeof(struct tmf_ioctl))) {
				PM8001_FAIL_DBG(pm8001_ha,
						pm8001_printk("copy from user failed\n"));
			} else {
				PM8001_INIT_DBG(pm8001_ha, pm8001_printk(" Lun "
						"number is %d  and sas addres "
						"is %llx\n",dev_info->lun_number,
						dev_info->sas_address));
				sas_addr = (u8 *)&dev_info->sas_address;
				sas_address_conversion(sas_addr);
				PM8001_INIT_DBG(pm8001_ha, 
					pm8001_printk("Converted sas address is %016llx\n",
					SAS_ADDR(sas_addr)));
			}
			pm8001_ha->tmf_resp.status = 0;
			pm8001_ha->tmf_resp.resp_code = 0;
			ret = ioctl_lu_reset(pm8001_ha,sas_addr,dev_info->lun_number);
			if (ret == -1) {
				return ret;
			}
			if ((pm8001_ha->tmf_resp.status) || (pm8001_ha->tmf_resp.resp_code)) {
				dev_info->iomb_status = pm8001_ha->tmf_resp.status;
				dev_info->resp_code = pm8001_ha->tmf_resp.resp_code;
				ret = -1;
            		}
			else {
                         return 0;
            		}			
			if (copy_to_user(dev_info2,dev_info,sizeof(struct tmf_ioctl))) {
				PM8001_FAIL_DBG(pm8001_ha,
						pm8001_printk("copy to user failed\n"));
				return -1;
			}
			
                        break;
		}
        	case ADPT_IOCTL_RESET_DEVICE:
			PM8001_INIT_DBG(pm8001_ha, pm8001_printk("IOCTL device reset request\n"));
			sas_addr = kzalloc(8, GFP_KERNEL);
			if (copy_from_user(sas_addr,(void*)arg,(SAS_ADDR_SIZE))) {
				PM8001_FAIL_DBG(pm8001_ha,
                                                pm8001_printk("copy from user failed\n"));
			} else {
				sas_address_conversion(sas_addr);
				PM8001_INIT_DBG(pm8001_ha,
					pm8001_printk("Converted sas address is %016llx\n",
					SAS_ADDR(sas_addr)))
			}
			ret =  ioctl_device_reset(pm8001_ha,sas_addr);
                        break;
        case ADPT_IOCTL_GET_VPD:
            ret = pm8001_ioctl_get_vpd(pm8001_ha, arg);
            return ret;


                case ADPT_IOCTL_ADAPTER_REVISION:
                        ret = pm8001_ioctl_set_adapter_revision(pm8001_ha, arg);
                        return ret;
        	case ADPT_IOCTL_DIAG_LOOPBACK:
        	case ADPT_IOCTL_DIAG_INTERNAL_LOOPBACK:
        	case ADPT_IOCTL_DIAG_EXTERNAL_LOOPBACK:
        	case ADPT_IOCTL_SAS_CONNECTION_STATE:
		case ADPT_IOCTL_DIAG_ENABLE:
		case ADPT_IOCTL_DIAG_DISABLE:
		case ADPT_IOCTL_START_STOP_PHY_DIAG:
		case ADPT_IOCTL_START_PHY_NORMAL:
		case ADPT_IOCTL_STOP_PHY_NORMAL:
			ret = pm8001_ctl_ioctl_diag(pm8001_ha, cmd, (void *)arg);
			return ret;
		default:
			ret = ADPT_IOCTL_CALL_INVALID_CODE;
	}
	if(ret == 0)
		return ret;
	header.return_code = ret;
	ret = -EACCES;
	if(copy_to_user((void *)arg, (void *)&header, sizeof(struct ioctl_header)))
	{
		PM8001_FAIL_DBG(pm8001_ha, pm8001_printk("copy_to_user failed \n"));
	}
	PM8001_FAIL_DBG(pm8001_ha, pm8001_printk("IOCTL failed \n"));

	return ret;
}
void sas_address_conversion(u8* sas_addr)
{
	int i;
	char temp;

	for(i = 0 ; i < (SAS_ADDR_SIZE / 2) ; i++) {
		temp = sas_addr[i];
		sas_addr[i] = sas_addr[SAS_ADDR_SIZE-i-1];
		sas_addr[SAS_ADDR_SIZE-i-1] = temp;
	}
}

int ioctl_device_reset(struct pm8001_hba_info *pm8001_ha, u8* sas_addr)
{

	int ret;
	struct sas_phy *phy;
	struct domain_device *dev;
	struct pm8001_device *pm8001_dev;
	DECLARE_COMPLETION_ONSTACK(completion_setstate);


	pm8001_dev = pm8001_find_dev_by_sas_address(pm8001_ha, sas_addr);
        if(!pm8001_dev) {
		PM8001_INIT_DBG(pm8001_ha,
			pm8001_printk("Given sas address not present"))
		return -1;
	}
	dev = pm8001_dev->sas_device;
	
	if (DEV_IS_EXPANDER(pm8001_dev->dev_type)) {
		return DEV_RESET_EXPANDER;
	}

	if(dev_is_sata(dev) && !(dev->parent)) {

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3, 4, 1) && \
                        !defined (PM8001_RHEL63)) && LINUX_VERSION_CODE != KERNEL_VERSION(3, 0, 76)
                                phy = sas_find_local_phy(dev);
#else
                                phy = sas_get_local_phy(dev);
#endif
                                ret = sas_phy_reset(phy, 1);
                                msleep(2000);
                                ret= pm8001_exec_internal_task_abort(pm8001_ha, pm8001_dev ,
                                                                        dev, 1, 0);
                                pm8001_dev->setds_completion = &completion_setstate;
                                ret = PM8001_CHIP_DISP->set_dev_state_req(pm8001_ha,
                                                                pm8001_dev, 0x01);
                                wait_for_completion(&completion_setstate);
                        }
                        else
                        ret = pm8001_I_T_nexus_reset(dev);
	return ret;

}
int ioctl_lu_reset(struct pm8001_hba_info *pm8001_ha, u8* sas_address, int lun_number)
{

	struct pm8001_device *pm8001_dev;
	struct scsi_lun lun;
	int rc ;
	struct domain_device *dev;
	DECLARE_COMPLETION(comp);
	DECLARE_COMPLETION_ONSTACK(completion_setstate);

	pm8001_dev = pm8001_find_dev_by_sas_address(pm8001_ha, sas_address);
	if(!pm8001_dev) {
		PM8001_FAIL_DBG(pm8001_ha,
			pm8001_printk("sas address of the device  not found\n"));
		return -1;
	} else {
		PM8001_INIT_DBG(pm8001_ha,
			pm8001_printk("sas address of the device found \n"));
	}
	dev = pm8001_dev->sas_device;
	if(DEV_IS_EXPANDER(pm8001_dev->dev_type)) {
		return -1;
	}

	#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0) )
	if (pm8001_dev->dev_type == SATA_DEV) { 
	#else
	if (pm8001_dev->dev_type == SAS_SATA_DEV) { 
	#endif
		if(!lun_number) {		
			return pm8001_lu_reset(dev,sas_address);
		}
		else {
			return -1;
		}
	}
	int_to_scsilun(lun_number, &lun);
	rc = pm8001_lun_reset_req(pm8001_ha,dev,lun.scsi_lun);
	pm8001_ha->ioctl_completion = &comp;
	wait_for_completion(&comp);
	if (pm8001_ha->tmf_resp.status != 0)
                return -1;
	pm8001_dev->setds_completion = &completion_setstate;
	PM8001_CHIP_DISP->set_dev_state_req(pm8001_ha,
					pm8001_dev, 0x01);
	wait_for_completion(&completion_setstate);
	return rc;
}

/**
 *	pm8001_poll - pm8001 poll request function
 *	@file: file handle
 *	@wait: poll table to wait
 *
 *	Handles a poll request.
 *
 */
unsigned int pm8001_poll (struct file *file, poll_table *wait)
{
	struct pm8001_hba_info *pm8001_ha;
	unsigned int mask = 0;

	pm8001_ha = file->private_data;

	poll_wait(file, &pm8001_ha->pollq,  wait);

	if(pm8001_ha->gpio_event_occured == 1)
	{
		pm8001_ha->gpio_event_occured = 0;
		mask |= POLLIN | POLLRDNORM;
	}
	
	return mask;
}

static const struct file_operations pm8001_fops = {
	.owner		= THIS_MODULE,
	.open		= pm8001_open,
	.release	= pm8001_close,
	.unlocked_ioctl	= pm8001_ioctl,
	.poll		= pm8001_poll,
};

static struct miscdevice ctl_dev = {
	.minor  = PM8001_MINOR,
	.name   = PM8001_DEV_NAME,
	.fops   = &pm8001_fops,
};


/*
 * pm8001_ctl_init - main entry point for ctl.
 */
void
pm8001_ctl_init(void)
{
	async_queue = NULL;
	if (misc_register(&ctl_dev) < 0)
		printk(KERN_ERR "%s can't register misc device [minor=%d]\n",
			PM8001_DEV_NAME, PM8001_MINOR);
	printk(KERN_ERR "reached: pm8001_ctl_init:%s:%d:%p\n", ctl_dev.name, ctl_dev.minor, ctl_dev.fops );
}

/*
 * pm8001_ctl_exit - exit point for ctl
 */
void
pm8001_ctl_exit(void)
{
	misc_deregister(&ctl_dev);
}





/**
 * pm8001_setup_chrdev - registers a char device
 *
 * Return value
 *	0 in case of success, otherwise non-zero
 */
int pm8001_setup_chrdev(void)
{
	pm8001_major = register_chrdev( 0, DRV_NAME, &pm8001_fops);
	if (pm8001_major < 0) {
		printk(KERN_WARNING
			"pm8001: unable to register \"%s\" device.\n", DRV_NAME);
		return pm8001_major;
	}
	return 0;
}

/**
 * pm8001_release_chrdev - unregisters per-adapter management interface
 *
 * Return value
 *  none
 */
void pm8001_release_chrdev(void)
{
	if (pm8001_major > -1)
	{
		unregister_chrdev(pm8001_major, DRV_NAME);
		pm8001_major = -1;
	}
}
